;|================================================================================================================|
;|====================================------_____________________________-----====================================|
;|########88888888XXXX>>>>>>>>////////-----/                             |----\\\\\\\\<<<<<<<<XXXX88888888########|
;|########88888888XXXX>>>>>>>>////====----/            PureSVG           |>---====\\\\<<<<<<<<XXXX88888888########|
;|########88888888XXXX>>>>>>>>========---/                               |>>--========<<<<<<<<XXXX88888888########|
;|########88888888XXXX>>>>>>>>========--|        A PureBasic library     |>>--========<<<<<<<<XXXX88888888########|
;|########88888888XXXX>>>>>>>>========--|        c/o Seymour Clufley     |>>--========<<<<<<<<XXXX88888888########|
;|########88888888XXXX>>>>>>>>========--|                                |>>--========<<<<<<<<XXXX88888888########|
;|########88888888XXXX>>>>>>>>========---\      Released under LGPL      |>>--========<<<<<<<<XXXX88888888########|
;|########88888888XXXX>>>>>>>>\\\\====----\   work begun 2nd Jan 2010    |>---====////<<<<<<<<XXXX88888888########|
;|########88888888XXXX>>>>>>>>\\\\\\\\-----\_____________________________|----////////<<<<<<<<XXXX88888888########|
;|====================================----------------------------------------====================================|
;|================================================================================================================|


Global ybx.i, sybx.i


;- DEPENDENCIES -------------------------------


#SVG_Base64 = "!BASE64"




;-
;- ELLIPTICAL ARC STUFF
; courtesy of Ollivier!

Structure EllipticalArc
	pA.PointD
	pB.PointD
	x_radius.d
	y_radius.d
	x_rotation.d
	large.b
	sweep.b
	
	c1.PointD
	c2.PointD
	startangle.d
	endangle.d
	bound.RectD
EndStructure

Procedure.F EArc_GetOrientation(xCenter.F, yCenter.F, xSource.F, ySource.F)
	Protected.F xDelta, yDelta
	Protected.F Rho, Alpha
	xDelta = xSource - xCenter
	yDelta = ySource - yCenter
	Rho = Sqr((xDelta * xDelta) + (yDelta * yDelta) )
	Alpha = ACos(xDelta / Rho)
	If yDelta > 0.
		Alpha = - Alpha
	EndIf
	ProcedureReturn Alpha
EndProcedure


Procedure EArc_RotatePoint(xCenter.d, yCenter.d, xSource.d, ySource.d, Beta.d, *xResult, *yResult)
	xDelta1.d = xSource - xCenter
	yDelta1.d = ySource - yCenter
	xRef.d = Cos(Beta)
	yRef.d = Sin(Beta)
	PokeD(*xResult, xCenter + xRef * xDelta1 + yRef * yDelta1)
	PokeD(*yResult, yCenter - yRef * xDelta1 + xRef * yDelta1)
EndProcedure

Procedure EArc_ScalePoint(xCenter.d, yCenter.d, xScale.d, yScale.d, xSource.d, ySource.d, *xResult, *yResult)
	PokeD(*xResult, (xSource - xCenter) * xScale + xCenter)
	PokeD(*yResult, (ySource - yCenter) * yScale + yCenter)
EndProcedure

Procedure.b EArc_CircleEquation(x1.d,y1.d,x2.d,y2.d,R0.d,*px0,*py0,*pxE,*pyE)
	Protected.d x1P2, y1P2, x2P2, y2P2, xDelta, yDelta, Dir
	Protected.d N, A, B, C, Base, Root
	Protected.d x0, y0, xE, yE
	Protected.b FlipFlag
	
	If y1 = y2
		Swap x1, y1
		Swap x2, y2
		FlipFlag = #True
	EndIf
	xDelta = x1 - x2
	yDelta = y1 - y2
	Dir = xDelta / yDelta
	x1P2 = x1 * x1
	y1P2 = y1 * y1
	x2P2 = x2 * x2
	y2P2 = y2 * y2
	N = (x1P2 - x2P2 + y1P2 - y2P2) / (2 * yDelta)
	A = (Dir * Dir) + 1
	B = 2 * ((y1 * Dir) - (N * Dir) - x1)
	C = x1P2 + y1P2 + (N + R0) * (N - R0) - (2 * Y1 * N)
	Delta.d = (B * B) - (4 * A * C)
	Debug xDelta
	Debug yDelta
	Debug Dir
	Debug N
	Debug A
	Debug B
	Debug C
	;Delta = Abs(Delta)
	Debug Delta
	If Delta => -0.00000001
		Base = - B / (2 * A)
		Root = Sqr(Delta) / (2 * A)
		x0 = Base - Root
		y0 = N - (x0 * Dir)
		xE = Base + Root
		yE = N - (xE * Dir)
		If FlipFlag
			Swap x0, y0
			Swap xE, yE
		EndIf
		PokeD(*px0, x0)
		PokeD(*py0, y0)
		PokeD(*pxE, xE)
		PokeD(*pyE, yE)
		ProcedureReturn #True
	Else
		R("CircleEquation fails."+c13+"Delta: "+StrD(Delta))
		ProcedureReturn #False
	EndIf
EndProcedure

Procedure.b EArc_GetCharacteristics(*ea.EllipticalArc)
	Protected.d xTemp, yTemp, xTmp, yTmp, xTmp0, yTmp0
	Protected.d xA, yA, xB, yB, A1, A2, A3
	With *ea
		xStart.d = \pA\x
		yStart.d = \pA\y
		xEnd.d = \pB\x
		yEnd.d = \pB\y
		
		xradius.d = \y_radius
		yradius.d = \x_radius
		
		EArc_RotatePoint(xStart, yStart, xEnd, yEnd, -\x_rotation, @xTemp, @yTemp)
		EArc_ScalePoint(xStart, yStart, 1.0, yRadius / xRadius, xTemp, yTemp, @xTemp, @yTemp)
		If EArc_CircleEquation(xStart, yStart, xTemp, yTemp, yRadius, @xTmp, @yTmp, @xTmp0, @yTmp0)
			StartArc.d = EArc_GetOrientation(xTmp, yTmp, xStart, yStart)
			EndArc.d = EArc_GetOrientation(xTmp, yTmp, xTemp, yTemp)
			EArc_ScalePoint(xStart, yStart, 1.0, xRadius / yRadius, xTmp, yTmp, @xTmp, @yTmp)
			EArc_ScalePoint(xStart, yStart, 1.0, xRadius / yRadius, xTmp0, yTmp0, @xTmp0, @yTmp0)
			EArc_RotatePoint(xStart, yStart, xTmp, yTmp, \x_rotation, @xA, @yA)
			EArc_RotatePoint(xStart, yStart, xTmp0, yTmp0, \x_rotation, @xB, @yB)
			A1 = EArc_GetOrientation(xStart, yStart, xA, yA) + #PI
			A2 = EArc_GetOrientation(xStart, yStart, xEnd, yEnd) + #PI
			A3 = EArc_GetOrientation(xStart, yStart, xB, yB) + #PI
			If ((A1 > A2) And (A2 > A3) ) Or ((A1 < A2) And (A2 > A3) And (A1 < A3) ) Or ((A1 > A2) And (A2 < A3) And (A1 < A3) )
				Swap xA, xB
				Swap yA, yB
				Swap StartArc, EndArc
				StartArc + #PI
				EndArc + #PI
			EndIf
			If \large = #False
				Swap xA, xB
				Swap yA, yB
				Swap StartArc, EndArc
				StartArc + #PI
				EndArc + #PI
			EndIf
			If \sweep
				Swap xA, xB
				Swap yA, yB
				StartArc + #PI
				EndArc + #PI
			EndIf
			\startangle = StartArc
			\endangle = EndArc
			\c1\x = xA
			\c1\y = yA
			\c2\x = xB
			\c2\y = yB
			ProcedureReturn #True
		Else
			ProcedureReturn #False
		EndIf
	EndWith
EndProcedure



Procedure EArc_GetBounds(*ea.EllipticalArc,*rect.RectF)
	
	Protected.F AngleStep, Angle, xEllipse, yEllipse, xDraw, yDraw, xPlot, yPlot
	With *ea
		If Not EArc_GetCharacteristics(*ea)
			Debug("COULD NOT GetArcBounds BECAUSE EArc_GetArcCharacteristics FAILS.")
		EndIf
		EndArc.f = \endangle
		StartArc.f = \startangle
		AngleStep = #PI / (2 * (\x_radius + \y_radius) )
		If EndArc < StartArc
			EndArc + (2 * #PI)
		EndIf
		Angle = StartArc
		*rect\x = 9999999.
		*rect\y = 9999999.
		*rect\w = -9999999.
		*rect\h = -9999999.
		Repeat
			xEllipse = Cos(Angle) * \x_radius
			yEllipse = Sin(Angle) * \y_radius
			xDraw = Cos(\x_rotation) * xEllipse - Sin(\x_rotation) * yEllipse
			yDraw = - Sin(\x_rotation) * xEllipse - Cos(\x_rotation) * yEllipse
			xPlot = \c1\x + xDraw
			yPlot = \c1\y + yDraw
			If xPlot < *rect\x
				*rect\x = xPlot
			EndIf
			If xPlot > *rect\w
				*rect\w = xPlot
			EndIf
			If yPlot < *rect\y
				*rect\y = yPlot
			EndIf
			If yPlot > *rect\h
				*rect\h = yPlot
			EndIf
			Angle + AngleStep
		Until Angle > EndArc
		*rect\w-*rect\x
		*rect\h-*rect\y
	EndWith
EndProcedure










;-
;------------------------------------------------------------------------------
;- LIBRARY INTERNAL CODE -------------------------------
;------------------------------------------------------------------------------






Procedure.b ReportRectF(*r.RectF,rnd=0,title.s="")
	
	If Nots title : title="ReportRectF" : EndIf
	t.s = "X = "+StrF(*r\x,rnd) +c13+ "Y = "+StrF(*r\y,rnd) +c13+ "W = "+StrF(*r\w,rnd) +c13+ "H = "+StrF(*r\h,rnd)
	R(title+c13+c13+t)
	
EndProcedure

Procedure.b DebugRectF(*r.RectF,rnd=0,title.s="")
	size = 4
	If title : title=RSet(title,6,c32) : EndIf
	Debug title+RSet(StrF(*r\x,rnd),size)+","+RSet(StrF(*r\y,rnd),size)+","+RSet(StrF(*r\w,rnd),size)+","+RSet(StrF(*r\h,rnd),size)
EndProcedure

Macro ReportPointF(p,rnd=2)
	R(StrF(p\x,rnd)+","+StrF(p\y,rnd))
EndMacro
Macro Report2PointF(a,b,rnd=2)
	R(StrF(a\x,rnd)+","+StrF(a\y,rnd)+" -> "+StrF(b\x,rnd)+","+StrF(b\y,rnd))
EndMacro



Procedure CopyRectF(*a.RectF,*b.RectF)
	*b\x = *a\x
	*b\y = *a\y
	*b\w = *a\w
	*b\h = *a\h
EndProcedure

Procedure.s RemoveTrailingZeroes(t.s,bad.s)
	bl = Len(bad)
	While Right(t,bl)=bad
		t = Left(t,Len(t)-bl)
	Wend
	If Right(t,1)="."
		t = Left(t,Len(t)-1)
	EndIf
	ProcedureReturn t
EndProcedure

#SVG_FloatAccuracy = 6

Macro SF(n,zoom=1)
	RemoveTrailingZeroes(StrF(n*zoom,#SVG_FloatAccuracy),"0")
EndMacro

Macro SFO(n,zoom=1)
	RemoveTrailingZeroes(StrF(n*svgmem(s)\output_zoom,#SVG_FloatAccuracy),"0")
EndMacro

Global previewfol.s = "C:\PureSVG_testing\"


Global tooltip_js.s
;.s = FileToString("P:\Documents\PB includes and libs\svg lib\tooltip js.txt")
tooltip_js + "function ShowTooltip(evt,tt) {"
tooltip_js + "if (navigator.appName=='Microsoft Internet Explorer') { return; }"
tooltip_js + "var tooltipbgid = 'inbuilt_tooltipbg_'+tt;"
tooltip_js + "var tooltiptxid = 'inbuilt_tooltiptx_'+tt;"
tooltip_js + "var x = evt.x+10;"
tooltip_js + "var y = evt.y-10;"
tooltip_js + "var txmargin = 3;"
tooltip_js + "var tel = document.getElementById(tooltiptxid);"
tooltip_js + "var w = tel.getComputedTextLength()+(2*txmargin);"
tooltip_js + "var gel = document.getElementById(tooltipbgid);"
tooltip_js + "gel.style.visibility = 'visible';"
tooltip_js + "gel.setAttribute('x',x);"
tooltip_js + "gel.setAttribute('y',y);"
tooltip_js + "gel.setAttribute('width',w);"
tooltip_js + "tel.style.visibility = 'visible';"
tooltip_js + "tel.setAttribute('x',x+txmargin);"
tooltip_js + "tel.setAttribute('y',y+12);"
tooltip_js + "}"
tooltip_js + "function ClearTooltip(tt) {"
tooltip_js + "var tooltipbgid = 'inbuilt_tooltipbg_'+tt;"
tooltip_js + "var tooltiptxid = 'inbuilt_tooltiptx_'+tt;"
tooltip_js + "document.getElementById(tooltipbgid).style.visibility = 'hidden';"
tooltip_js + "document.getElementById(tooltiptxid).style.visibility = 'hidden';"
tooltip_js + "}"



#SVG_DocType = "<!DOCTYPE svg PUBLIC "+Chr(34)+"-//W3C//DTD SVG 1.1//EN"+Chr(34)+" "+Chr(34)+"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"+Chr(34)+">"
#MaxSVGs = 10
;#MaxSVGFilterStages = 10
#SVG_ArrayJump = 10



Enumeration 1 ; SVG elements
	#SVG_Element_Box
	#SVG_Element_Circle
	#SVG_Element_Text
	#SVG_Element_WrappedText
	#SVG_Element_TextPath
	#SVG_Element_Line
	#SVG_Element_Ellipse
	#SVG_Element_Polyline
	#SVG_Element_Polygon
	#SVG_Element_Path
	#SVG_Element_Image
	#SVG_Element_PBImage
	#SVG_Element_Base64Image
	#SVG_Element_ForeignObject
	#SVG_Element_Group
	#SVG_Element_GroupCloser
	#SVG_Element_Comment
	#SVG_Element_CustomCode
EndEnumeration


#SVG_Action_OnClick = "onclick"
#SVG_Action_OnMouseOver = "onmouseover"
#SVG_Action_OnMouseOut = "onmouseout"
#SVG_Action_OnMouseMove = "onmousemove"
#SVG_Action_OnMouseDown = "onmousedown"
#SVG_Action_OnMouseUp = "onmouseup"
#SVG_Action_OnLoad = "onload"
#SVG_Action_OnActivate = "onactivate"
#SVG_Action_OnFocusIn = "onfocusin"
#SVG_Action_OnFocusOut = "onfocusout"



Enumeration 1 ; stroke types
	#SVG_Stroke_Miter
	#SVG_Stroke_Round
	#SVG_Stroke_Bevel
	#SVG_Stroke_Inherit
	#SVG_Stroke_Butt
	#SVG_Stroke_Square
EndEnumeration

Enumeration 0 ; element fill modes
	#SVG_FillMode_None
	#SVG_FillMode_Color
	#SVG_FillMode_Gradient
	#SVG_FillMode_Pattern
EndEnumeration

Enumeration -20 ; cursors
	#SVG_Cursor_CrossHair
	#SVG_Cursor_Pointer
	#SVG_Cursor_Move
	#SVG_Cursor_Help
	#SVG_Cursor_Text
	#SVG_Cursor_Wait
	#SVG_Cursor_EResize
	#SVG_Cursor_NEResize
	#SVG_Cursor_NWResize
	#SVG_Cursor_NResize
	#SVG_Cursor_SEResize
	#SVG_Cursor_SWResize
	#SVG_Cursor_SResize
	#SVG_Cursor_WResize
	
	#SVG_Cursor_URI = -1
	
	#SVG_Cursor_Default = 0
EndEnumeration

Enumeration -2 ; path movement modes
	#SVG_Path_SegmentStart
	#SVG_Path_SegmentEnd
	
	#SVG_Path_Straight
	#SVG_Path_CubicBezier
	#SVG_Path_SmoothCubicBezier
	#SVG_Path_QuadraticBezier
	#SVG_Path_SmoothQuadraticBezier
	
	#SVG_Path_EllipticalArc
EndEnumeration

Enumeration 1 ; hyperlink targets
	#SVG_Target_Self
	#SVG_Target_Blank
	#SVG_Target_Top
	#SVG_Target_Parent
	#TotalSVGTargets = #PB_Compiler_EnumerationValue-1
EndEnumeration
	

Procedure.s PathMovementModeName(m)
	Select m
		Case #SVG_Path_SegmentStart
			mn.s = "start"
		Case #SVG_Path_SegmentEnd
			mn = "end"
		Case #SVG_Path_Straight
			mn = "straight"
		Case #SVG_Path_CubicBezier
			mn = "cubic bezier"
		Case #SVG_Path_SmoothCubicBezier
			mn = "smooth cubic bezier"
		Case #SVG_Path_QuadraticBezier
			mn = "quadratic bezier"
		Case #SVG_Path_SmoothQuadraticBezier
			mn = "smooth quadratic bezier"
		Default
			mn = "UNKNOWN"
	EndSelect
	ProcedureReturn mn
EndProcedure


;-
;- STRUCTURES

Structure SVG_PathPoint_Structure
	x.f
	y.f
	mode.b
	intermediary_points.i
	Array intermediary_point.PointF(1)
	
	x_rotation.f
	rx.f
	ry.f
	large_arc.b
	sweep.b
EndStructure

Structure SVG_Path_Structure
	pathpoints.i
	Array pathpoint.SVG_PathPoint_Structure(0)
EndStructure

Structure SVG_MotionKeyStructure
	keytime.s
	keypoint.f
EndStructure

Structure SVG_Animator_Structure
	in_use.b
	type.b
	
	begin.s
	;trigger.b
	
	attr_name.s
	from_val.s
	to_val.s
	duration.s
	restart.b
	repeat_count.i
	repeat_duration.i
	
	fill.b
	
	path.SVG_Path_Structure
	calc_mode.b
	List key.SVG_MotionKeyStructure()
	rotate.b
	
	transform_type.b
	
EndStructure

Structure SVG_Element_Structure
	in_use.b
	type.b
	id.s
	
	x.f
	y.f
	w.f
	h.f
	cx.f
	cy.f
	rx.f
	ry.f
	
	bound.RectF
	bounds_known.b
	
	tooltip.s
	content.s
	class.s
	styling.s
	visibility.b
	Map attribute.s()
	strokecolor.i
	strokewidth.f
	strokeopacity.i
	strokelinejoin.b
	strokelinecap.b
	opacity.f
	filter.i
	
	ismask.b
	mask.i
	
	cursor.b
	cursor_url.s
	
	fillmode.b
	fillvalue.i
	
	fontname.s
	fontsize.i
	
	href.s
	href_target.b
	
	path.SVG_Path_Structure
	
	imagecodec.i
	imagenumber.i
	filename.s
	store_internally.b
	
	base64data.s
	
	animators.i
	Array animator.SVG_Animator_Structure(0)
	
	;action.s[#TotalSVGActions+1]
	Map action.s()
EndStructure


Enumeration 1 ; SVG gradient types
	#SVG_Gradient_Linear
	#SVG_Gradient_Radial
EndEnumeration

Structure SVG_GradientPoint_Structure
	position.f
	color.i
	opacity.f
EndStructure

Structure SVG_Gradient_Structure
	in_use.b
	id.s
	type.b
	x.f
	y.f
	w.f
	h.f
	List point.SVG_GradientPoint_Structure()
	
	employed.b
EndStructure


Enumeration 1 ; pattern modes
	#SVG_Pattern_SVG
	#SVG_Pattern_DirectCode
EndEnumeration

Structure SVG_Pattern_Structure
	in_use.b
	id.s
	w.f
	h.f
	mode.b
	svg.i
	code.s
	
	employed.b
EndStructure


#SVG_Composite_Over = "over" ; default
#SVG_Composite_In = "in"
#SVG_Composite_Out = "out"
#SVG_Composite_Atop = "atop"
#SVG_Composite_Xor = "xor"
#SVG_Composite_Arithmetic = "arithmetic"

#SVG_Blend_Normal = "normal" ; default
#SVG_Blend_Multiply = "multiply"
#SVG_Blend_Screen = "screen"
#SVG_Blend_Darken = "darken"
#SVG_Blend_Lighten = "lighten"

#SVG_FilterInput_SourceGraphic = "SourceGraphic" ; default
#SVG_FilterInput_SourceAlpha = "SourceAlpha"
#SVG_FilterInput_BackgroundImage = "BackgroundImage"
#SVG_FilterInput_BackgroundAlpha = "BackgroundAlpha"
#SVG_FilterInput_FillPaint = "FillPaint"
#SVG_FilterInput_StrokePaint = "StrokePaint"

#SVG_Morphology_Erode = "erode" ; default
#SVG_Morphology_Dilate = "dilate"

#SVG_PreserveAspectRatio_Default = "xMidYMid meet"
#SVG_PreserveAspectRatio_ = ""
#SVG_PreserveAspectRatio_ = ""

Enumeration 1 ; filter primitives
	#SVG_FP_Blend
	#SVG_FP_ColorMatrix
	#SVG_FP_ComponentTransfer
	#SVG_FP_Composite
	#SVG_FP_ConvolveMatrix
	#SVG_FP_DiffuseLighting
	#SVG_FP_DisplacementMap
	#SVG_FP_DistantLight
	#SVG_FP_Flood
	#SVG_FP_GaussianBlur
	#SVG_FP_Image
	#SVG_FP_Merge
	#SVG_FP_Morphology
	#SVG_FP_Offset
	#SVG_FP_PointLight
	#SVG_FP_SpecularLighting
	#SVG_FP_SpotLight
	#SVG_FP_Tile
	#SVG_FP_Turbulence
	#SVG_TotalFilterPrimitives = #PB_Compiler_EnumerationValue-1
EndEnumeration


Structure PrimitiveRequirements
	tag.s
	inputs.i
	uses_region.b
EndStructure
Global Dim SVGFilterPrimitive.PrimitiveRequirements(#SVG_TotalFilterPrimitives)
SVGFilterPrimitive(#SVG_FP_ColorMatrix)\inputs = 1
SVGFilterPrimitive(#SVG_FP_ComponentTransfer)\inputs = 1
SVGFilterPrimitive(#SVG_FP_ConvolveMatrix)\inputs = 1
SVGFilterPrimitive(#SVG_FP_DiffuseLighting)\inputs = 1
SVGFilterPrimitive(#SVG_FP_GaussianBlur)\inputs = 1
SVGFilterPrimitive(#SVG_FP_Merge)\inputs = 1
SVGFilterPrimitive(#SVG_FP_Offset)\inputs = 1
SVGFilterPrimitive(#SVG_FP_Morphology)\inputs = 1
SVGFilterPrimitive(#SVG_FP_SpecularLighting)\inputs = 1
SVGFilterPrimitive(#SVG_FP_Tile)\inputs = 1
SVGFilterPrimitive(#SVG_FP_Blend)\inputs = 2
SVGFilterPrimitive(#SVG_FP_Composite)\inputs = 2
SVGFilterPrimitive(#SVG_FP_DisplacementMap)\inputs = 2
For p = 0 To #SVG_TotalFilterPrimitives
	SVGFilterPrimitive(p)\uses_region = #True
Next p
SVGFilterPrimitive(#SVG_FP_Blend)\uses_region = #False

SVGFilterPrimitive(#SVG_FP_Blend)\tag = "Blend"
SVGFilterPrimitive(#SVG_FP_ColorMatrix)\tag = "ColorMatrix"
SVGFilterPrimitive(#SVG_FP_ComponentTransfer)\tag = "ComponentTransfer"
SVGFilterPrimitive(#SVG_FP_Composite)\tag = "Composite"
SVGFilterPrimitive(#SVG_FP_ConvolveMatrix)\tag = "ConvolveMatrix"
SVGFilterPrimitive(#SVG_FP_DiffuseLighting)\tag = "DiffuseLighting"
SVGFilterPrimitive(#SVG_FP_DisplacementMap)\tag = "DisplacementMap"
SVGFilterPrimitive(#SVG_FP_DistantLight)\tag = "DistantLight"
SVGFilterPrimitive(#SVG_FP_Flood)\tag = "Flood"
SVGFilterPrimitive(#SVG_FP_GaussianBlur)\tag = "GaussianBlur"
SVGFilterPrimitive(#SVG_FP_Image)\tag = "Image"
SVGFilterPrimitive(#SVG_FP_Merge)\tag = "Merge"
SVGFilterPrimitive(#SVG_FP_Morphology)\tag = "Morphology"
SVGFilterPrimitive(#SVG_FP_Offset)\tag = "Offset"
SVGFilterPrimitive(#SVG_FP_PointLight)\tag = "PointLight"
SVGFilterPrimitive(#SVG_FP_SpecularLighting)\tag = "SpecularLighting"
SVGFilterPrimitive(#SVG_FP_SpotLight)\tag = "SpotLight"
SVGFilterPrimitive(#SVG_FP_Tile)\tag = "Tile"
SVGFilterPrimitive(#SVG_FP_Turbulence)\tag = "Turbulence"


Structure SVG_LightStructure
	id.s
	type.b
	
	azimuth.f
	elevation.f
	
	source_x.f
	source_y.f
	source_z.f
	
	target_x.f
	target_y.f
	target_z.f
	
	specular_exponent.f
	limiting_cone_angle.f
EndStructure

Structure SVG_FilterPrimitive_Structure
	
	id.s
	type.b
	input1.s
	input2.s
	
	region.RectF
	
	; blend
	blend_mode.s
	
	; composite
	Array k.f(0)
	
	; convolve matrix
	order.s
	kernel_matrix.s
	divisor.f
	bias.f
	target_x.i
	target_y.i
	edge_mode.s
	kernel_unit_length.s
	preserve_alpha.b
	
	; diffuse lighting
	List light.SVG_LightStructure()
	light_color.i
	diffuse_constant.f
	
	; flood
	flood_color.i
	flood_opacity.f
	
	; gaussian blur
	blur_amount.f
	
	; image
	preserveAspectRatio.s
	
	; morphology
	radius_x.f
	radius_y.f
	operator.s
	
	; offset
	dx.f
	dy.f
	
	; turbulence
	baseXFrequency.f
	baseYFrequency.f
	octaves.i
	seed.f
	stitch.s
	turbulence_type.s
	
	; specular lighting
	surface_scale.f
	specular_constant.f
	specular_exponent.f
	;kernel_unit_length.s ; already with ConvolveMatrix
	
	; displacement map
	scale.f
	xChannelSelector.s
	yChannelSelector.s
	
EndStructure

Structure SVG_Filter_Structure
	in_use.b
	id.s
	region.RectF
	
	List primitive.SVG_FilterPrimitive_Structure()
	
	employed.b
EndStructure




Enumeration 1 ; metadata fields
	#SVG_Metadata_Title
	#SVG_Metadata_Description
	#SVG_Metadata_Creator
	#SVG_Metadata_Publisher
	#SVG_Metadata_Rightsholder
	#SVG_Metadata_Source
	#SVG_Metadata_Coverage
	#SVG_Metadata_Contributors
	#SVG_Metadata_Language
	#SVG_Metadata_Date
	#SVG_Metadata_Keywords
	
	#TotalSVGMetadataFields
EndEnumeration




Structure SVG_Structure
	
	in_use.b
	
	md.s[#TotalSVGMetadataFields+1]
	
	setw.f
	seth.f
	w.f
	h.f
	
	bound.RectF
	bounds_known.b
	visible.RectF
	
	output_zoom.f
	output_w.f
	output_h.f
	
	selviz.RectF
	
	
	bg_mode.b
	bg_value.i
	
	id.s
	elements.i
	Array element.SVG_Element_Structure(0)
	javascript.s
	List external_js_file.s()
	List external_css_file.s()
	classinfo.s
	gradients.i
	Array gradient.SVG_Gradient_Structure(0)
	patterns.i
	Array pattern.SVG_Pattern_Structure(0)
	filters.i
	Array filter.SVG_Filter_Structure(0)
	
	Map namespace.s()
	
	Map attribute.s()
	
	; during outputting...
	tooltiph.s
	
EndStructure




Global Dim svgmem.SVG_Structure(#MaxSVGs)
Structure SVGDataStructure
	s.i ; current svg
	e.i ; current element
	g.i ; current gradient
	p.i ; current pattern
	f.i ; current filter
	a.i ; current animator
	FrontColor.i
	BackColor.i
	GroupLevel.i
	FontName.s
	FontSize.i
	
	arrow_stalkw.f
	arrow_stalkl.f
	arrow_headw.f
	arrow_headl.f
	
EndStructure
Global svgdata.SVGDataStructure
svgdata\arrow_stalkw = 6
svgdata\arrow_stalkl = 40
svgdata\arrow_headw = 16
svgdata\arrow_headl = 10



Declare.b IsSVG(s.i)
Declare.b SVG_EnsureAllVisible()
;Declare.b CalculateTotalSVGElementBounds(s.i)
Declare.i SVG_Path_Line(dx.f,dy.f,mode.b=#SVG_Path_Straight,ix1.f=-#PB_Ignore,iy1.f=#PB_Ignore,ix2.f=#PB_Ignore,iy2.f=#PB_Ignore,ix3.f=#PB_Ignore,iy3.f=#PB_Ignore,ix4.f=#PB_Ignore,iy4.f=#PB_Ignore)
Declare.i SVG_Path_StartSegment(x.f,y.f)
Declare.i SVG_Path_EndSegment()
Declare.i SVG_Path_EllipticalArc(dx.f,dy.f,rx.f,ry.f,xrot.f,large_arc.b,sweep.b)
Declare.s RandomID(s.i)
Declare.i CloserForSVGGroup(s,g)
Declare.b AddToSVGPath(arr.s,*e.SVG_Path_Structure)
Declare.i Mark(*c.PointF,clr)
Declare.b ReportElementBounds(s.i,e.i)
Declare.b SVGInternal_RotateGroupElement(s.i,g.i,deg.f,cx.f=#PB_Ignore,cy.f=#PB_Ignore)
Declare.b PolyRect(*r.RectF,clr=#Green)


Procedure.s StringMapToXMLAttributes(Map mp.s())
	
	attrarr.s = ""
	ForEach mp()
		attrarr + MapKey(mp())+"='"+mp()+"' "
	Next
	ProcedureReturn attrarr
	
EndProcedure


Macro EnsureEnoughPathPoints(pathstruc,pnts)
	pathpointarrsize = ArraySize(pathstruc\pathpoint(),1)
	If pathpointarrsize<pnts
		If pathpointarrsize=-1
			;Dim svgmem(s)\element(e)\pathpoint(pnts+#SVG_ArrayJump)
			Dim pathstruc\pathpoint(pnts+#SVG_ArrayJump)
		Else
			ReDim pathstruc\pathpoint(pnts+#SVG_ArrayJump)
		EndIf
	EndIf
EndMacro

Macro EnsureEnoughPathPointsPointered(elptr,pnts)
	pathpointarrsize = ArraySize(elptr\pathpoint(),1)
	;R("PNTS: "+Str(pnts))
	If pathpointarrsize < pnts
		If pathpointarrsize < 1
			Dim elptr\pathpoint.SVG_PathPoint_Structure(pnts+#SVG_ArrayJump)
			;R("DIM")
		Else
			ReDim elptr\pathpoint(pnts+#SVG_ArrayJump)
			;R("REDIM")
		EndIf
	EndIf
	;R("PT: "+Str(elptr)+c13+c13+"OLD PATH POINT ARR SIZE: "+Str(pathpointarrsize)+c13+c13+"NEW Array SIZE: "+Str(ArraySize(elptr\pathpoint(),1)))
EndMacro



Procedure.s SVGElementTypeName(t.b)
	
	n.s
	
	Select t
		Case #SVG_Element_Box
			n = "Rect"
		Case #SVG_Element_Circle
			n = "Circle"
		Case #SVG_Element_Text
			n = "Text"
		Case #SVG_Element_WrappedText
			n = "WrappedText"
		Case #SVG_Element_TextPath
			n = "TextPath"
		Case #SVG_Element_Line
			n = "Line"
		Case #SVG_Element_Ellipse
			n = "Ellipse"
		Case #SVG_Element_Polyline
			n = "Polyline"
		Case #SVG_Element_Polygon
			n = "Polygon"
		Case #SVG_Element_Path
			n = "Path"
		Case #SVG_Element_Image
			n = "Image"
		Case #SVG_Element_Base64Image
			n = "Base64Image"
		Case #SVG_Element_PBImage
			n = "PBImage"
		Case #SVG_Element_Group
			n = "Group"
		Case #SVG_Element_GroupCloser
			n = "GroupCloser"
		Case #SVG_Element_CustomCode
			n = "CustomCode"
		Case #SVG_Element_Comment
			n = "Comment"
	EndSelect
	
	ProcedureReturn n
	
EndProcedure

Procedure.s SVGElementTypeTag(t.b)
	
	n.s
	
	Select t
		Case #SVG_Element_Box
			n = "rect"
		Case #SVG_Element_Circle
			n = "circle"
		Case #SVG_Element_Text
			n = "text"
		Case #SVG_Element_TextPath
			n = "textpath"
		Case #SVG_Element_Line
			n = "line"
		Case #SVG_Element_Ellipse
			n = "ellipse"
		Case #SVG_Element_Polyline
			n = "polyline"
		Case #SVG_Element_Polygon
			n = "polygon"
		Case #SVG_Element_Path
			n = "path"
		Case #SVG_Element_Image, #SVG_Element_PBImage, #SVG_Element_Base64Image
			n = "image"
		Case #SVG_Element_Group
			n = "g"
	EndSelect
	
	ProcedureReturn n
	
EndProcedure


Macro SVGInternal_ForEachElement(targetsvg,svg_ge)
	For svg_ge = 1 To svgmem(targetsvg)\elements
		If Not svgmem(targetsvg)\element(svg_ge)\in_use : Continue : EndIf
		Select svgmem(targetsvg)\element(svg_ge)\type
			Case #SVG_Element_GroupCloser
				Continue
		EndSelect
		
	EndMacro
	
	Macro SVGInternal_ForEachElementInSVGGroup(s,g,svg_ge)
		;svg_lastge = CloserForSVGGroup(s,g)
		For svg_ge = g+1 To CloserForSVGGroup(s,g)-1;svg_lastge
			If Not svgmem(s)\element(svg_ge)\in_use : Continue : EndIf
			If svgmem(s)\element(svg_ge)\type=#SVG_Element_GroupCloser : Continue : EndIf
		EndMacro
		
		
		Procedure.b SVGGroupContainsElements(s.i,e.i)
			
			If svgmem(s)\element(e)\type=#SVG_Element_Group
				SVGInternal_ForEachElementInSVGGroup(s,e,a)
				If svgmem(s)\element(a)\in_use
					ProcedureReturn #True
				EndIf
			Next a
		EndIf
		
		ProcedureReturn #False
		
	EndProcedure
	
	
	
	Procedure.s HyperlinkTargetName(trg)
		
		Select trg
			Case #SVG_Target_Self
				ProcedureReturn "_self"
			Case #SVG_Target_Blank
				ProcedureReturn "_blank"
			Case #SVG_Target_Top
				ProcedureReturn "_top"
			Case #SVG_Target_Parent
				ProcedureReturn "_parent"
		EndSelect
		
	EndProcedure
	
	Procedure.b HyperlinkTargetNumber(nm.s)
		nm = LCase(nm)
		For a = 1 To #TotalSVGTargets
			If nm=HyperlinkTargetName(a)
				ProcedureReturn a
			EndIf
		Next a
		
		ProcedureReturn #SVG_Target_Top
		
	EndProcedure
	
	
	
	
	Procedure.i FirstAvailableSVGSlot()
		
		For s = 1 To #MaxSVGs
			If Not svgmem(s)\in_use
				ProcedureReturn s
			EndIf
		Next s
		ProcedureReturn 0
		
	EndProcedure
	
	Macro ValidSVGNumber(s)
		(s>0 And s<=#MaxSVGs And s<>#PB_Any)
	EndMacro
	
	Macro InitiateSVG(s)
		
		ClearStructure(@svgmem(s),SVG_Structure)
		InitializeStructure(@svgmem(s),SVG_Structure)
		Dim svgmem(s)\element(#SVG_ArrayJump)
		
		;With svgmem(s)
		svgmem(s)\in_use = #True
		svgmem(s)\output_zoom = 1
		For e = 0 To #SVG_ArrayJump;ArraySize(svgmem(s)\element(),1)
			;InitializeStructure(@svgmem(s)\element(e),SVG_Element_Structure)
			svgmem(s)\element(e)\fillmode=-1
			svgmem(s)\element(e)\fillvalue=-1
			svgmem(s)\element(e)\strokecolor=-1
		Next e
		
		os = svgdata\s
		svgdata\s = s
		SVG_DefaultNameSpaces()
		svgdata\s = os
		
		;EndWith
		
	EndMacro
	
	
	Macro ResetSVGElementBounds(e)
		svgmem(svgdata\s)\element(e)\bound\x = 0
		svgmem(svgdata\s)\element(e)\bound\y = 0
		svgmem(svgdata\s)\element(e)\bound\w = 0
		svgmem(svgdata\s)\element(e)\bound\h = 0
	EndMacro
	
	
	
	
	Procedure.b SizeToFitInside(iw.f,ih.f,fw.f,fh.f,*dw,*dh)
		
		dw.f
		dh.f
		
		If fw=fh
			;Debug "FRAME IS SQUARE"
			If iw=ih
				;Debug "square"
				; square
				dw=fw
				dh=fh
			EndIf
			If iw>ih
				;Debug "landscape"
				; landscape
				dh = ih/iw*fw
				dw = fw
			EndIf
			If iw<ih
				;Debug "portrait"
				; portrait
				dw = iw/ih*fh
				dh = fh
			EndIf
		Else
			If fw>fh
				;Debug "FRAME IS LANDSCAPE"
				If iw=ih
					;Debug "square image"
					; square
					dw = Defeat(fw,fh)
					dh = dw
				EndIf
				If iw>ih
					;Debug "landscape image"
					; landscape
					dw = fw
					dh = ih/iw*dw
				EndIf
				If iw<ih
					;Debug "portrait image"
					;R("YEAH")
					; portrait
					dh = fh
					dw = iw/ih*dh
				EndIf
			Else
				;Debug "FRAME IS PORTRAIT"
				If iw = ih
					;Debug "square image"
					; square
					dw = Defeat(fw,fh)
					dh = dw
				EndIf
				If iw>ih
					;Debug "landscape image"
					; landscape
					dh = ih/iw*fw
					dw = fw
				EndIf
				If iw<ih
					;Debug "portrait image"
					; portrait
					dw = fw
					dh = fh/iw*fw
				EndIf
			EndIf
		EndIf
		
		;If donotenlarge
		;    If dh>ih Or dw>iw
		;        dh = ih
		;        dw = iw
		;    EndIf
		;EndIf
		
		;R("IMAGE: "+StrF(iw,0)+","+StrF(ih,0)+c13+"FRAME: "+StrF(fw,0)+","+StrF(fh,0)+c13+"    D: "+StrF(dw,0)+","+StrF(dh,0))
		
		PokeF(*dw,dw)
		PokeF(*dh,dh)
		
		ProcedureReturn #True
		
	EndProcedure
	
	
	
	
	Procedure.f SVG_TextWidth(text.s,FontID.i)
		
		TEST=CreateImage(#PB_Any,500,250)
		StartDrawing(ImageOutput(TEST))
		DrawingFont(FontID)
		tw.f = TextWidth(text)
		StopDrawing()
		FreeImage(TEST)
		
		;R(text)
		;R(StrF(tw))
		ProcedureReturn tw/1.38443935;/1.55833;0.725;/72.46
		
	EndProcedure
	
	Procedure.f SVG_TextHeight(text.s,FontID.i)
		
		TEST=CreateImage(#PB_Any,250,500)
		StartDrawing(ImageOutput(TEST))
		DrawingFont(FontID)
		th.f = TextHeight(text)
		StopDrawing()
		FreeImage(TEST)
		
		ProcedureReturn th
		
	EndProcedure
	
	
	
	
	Procedure.b CopyArcDetails(s.i,e.i,p.i,*ea.EllipticalArc)
		With svgmem(s)\element(e)\path\pathpoint(p)
			*ea\pA\x = svgmem(s)\element(e)\path\pathpoint(p-1)\x
			*ea\pA\y = svgmem(s)\element(e)\path\pathpoint(p-1)\y
			*ea\pB\x = \x
			*ea\pB\y = \y
			*ea\large = \large_arc
			*ea\sweep = \sweep
			*ea\x_radius = \rx
			*ea\y_radius = \ry
			*ea\x_rotation = Degree(\x_rotation)
		EndWith
	EndProcedure
	
	
	Procedure.b GetEllipticalArcBounds(s.i,e.i,p.i,*bb.RectF)
		
		If p<1 : ProcedureReturn #False : EndIf
		If p=1
			R("An elliptical arc cannot be the first point in a path. You need to start with a MoveTo (the SVG_Path_StartSegment command)"+c13+c13+"ELEMENT ID: "+svgmem(s)\element(e)\id)
			ProcedureReturn #False
		EndIf
		
		With svgmem(s)\element(e)\path\pathpoint(p)
			If \mode<>#SVG_Path_EllipticalArc
				R("This is not an elliptical arc.")
				ProcedureReturn #False
			EndIf
			
			ea.EllipticalArc
			CopyArcDetails(s,e,p,@ea)
			
			;EArc_GetCharacteristics(@ea)
			
			EArc_GetBounds(@ea,*bb)
			
		EndWith
		
	EndProcedure
	
	Procedure.b CalculateSVGPathBounds(s.i,e.i)
		
		px.f = 9999999
		pw.f = -9999999
		py.f = 9999999
		ph.f = -9999999
		
		;R("CalculatePathBounds"+c13+"Path points: "+Str(svgmem(s)\element(e)\pathpoints))
		
		If svgmem(s)\element(e)\path\pathpoints>0
			For p = 1 To svgmem(s)\element(e)\path\pathpoints
				EnsureEnoughPathPoints(svgmem(s)\element(e)\path,svgmem(s)\element(e)\path\pathpoints)
				With svgmem(s)\element(e)\path\pathpoint(p)
					If \mode = #SVG_Path_SegmentEnd
						Continue
					EndIf
					
					;R("MODE: "+Str(\mode)+c13+RSet(StrF(\x,0),3,c32)+","+RSet(StrF(\y,0),3,c32))
					
					DefeatThis(px,\x)
					BeatThis(pw,\x)
					DefeatThis(py,\y)
					BeatThis(ph,\y)
					
					Select \mode
						Case #SVG_Path_EllipticalArc
							GetEllipticalArcBounds(s,e,p,@bb.RectF)
							DefeatThis(px,bb\x)
							DefeatThis(py,bb\y)
							BeatThis(pw,bb\x+bb\w)
							BeatThis(ph,bb\y+bb\h)
						Case #SVG_Path_SmoothCubicBezier, #SVG_Path_QuadraticBezier, #SVG_Path_SmoothQuadraticBezier
							
							If \intermediary_points
								For i = 1 To \intermediary_points
									DefeatThis(px,\intermediary_point(i)\x)
									BeatThis(pw,\intermediary_point(i)\x)
									DefeatThis(py,\intermediary_point(i)\y)
									BeatThis(ph,\intermediary_point(i)\y)
								Next i
							EndIf
					EndSelect
					
				EndWith
			Next p
		EndIf
		
		;R("PATHPOINTS: "+Str(svgmem(s)\element(e)\pathpoints)+c13+c13+"LOW: "+StrF(px,0)+","+StrF(py,0)+c13+"HIGH: "+StrF(pw,0)+","+StrF(ph,0))
		
		With svgmem(s)\element(e)
			\bound\x = px
			\bound\y = py
			\bound\w = pw-px
			\bound\h = ph-py
			\bounds_known = #True
		EndWith
		
		ProcedureReturn #True
		
	EndProcedure
	
	
	
	
	Procedure.b CalculateSVGElementBounds(s.i,e.i,rep.b=#False)
		
		With svgmem(s)\element(e)
			;If \bounds_known : ProcedureReturn #True : EndIf
			\bound\x = 9999999
			\bound\y = 9999999
			\bound\w = -9999999
			\bound\h = -9999999
			Select \type
				Case #SVG_Element_GroupCloser, #SVG_Element_CustomCode, #SVG_Element_Comment
					ProcedureReturn #False
				Case #SVG_Element_Box, #SVG_Element_Image, #SVG_Element_PBImage, #SVG_Element_Base64Image, #SVG_Element_WrappedText
					\bound\x = \x
					\bound\y = \y
					\bound\w = \w ; -\x
					\bound\h = \h ; -\y
				Case #SVG_Element_Circle, #SVG_Element_Ellipse
					\bound\x = \cx-\rx
					\bound\y = \cy-\ry
					\bound\w = \rx*2
					\bound\h = \ry*2
				Case #SVG_Element_Line, #SVG_Element_Path, #SVG_Element_TextPath, #SVG_Element_Polygon, #SVG_Element_Polyline
					;R("ELEMENT IS PATH/ABLE")
					CalculateSVGPathBounds(s,e)
				Case #SVG_Element_Text ;`
					\bound\x = \x
					\bound\y = \y
					fnt = LoadFont(#PB_Any,\fontname,\fontsize)
					\bound\w = SVG_TextWidth(\content,FontID(fnt))
					\bound\h = SVG_TextHeight(\content,FontID(fnt))
					FreeFont(fnt)
					\bound\y-\bound\h
					;ReportRectF(@\bound)
				Case #SVG_Element_Group
					;If rep : R(Str(e+1)+" TO "+Str(gc-1)) : EndIf
					;rep=#True
					If SVGGroupContainsElements(s,e)
						SVGInternal_ForEachElementInSVGGroup(s,e,ge)
						;KnowBounds(s,ge)
						CalculateSVGElementBounds(s,ge)
						\bound\x = Defeat(\bound\x,svgmem(s)\element(ge)\bound\x)
						\bound\y = Defeat(\bound\y,svgmem(s)\element(ge)\bound\y)
						\bound\w = Beat(\bound\w,svgmem(s)\element(ge)\bound\x+svgmem(s)\element(ge)\bound\w)
						\bound\h = Beat(\bound\h,svgmem(s)\element(ge)\bound\y+svgmem(s)\element(ge)\bound\h)
						;\bound\w = Beat(\bound\w,svgmem(s)\element(ge)\bound\w)
						;\bound\h = Beat(\bound\h,svgmem(s)\element(ge)\bound\h)
						
						If rep : DebugRectF(@svgmem(s)\element(ge)\bound,0,svgmem(s)\element(ge)\id) : EndIf
					Next ge
					\bound\w - \bound\x
					\bound\h - \bound\y
				Else
					\bound\x = 0
					\bound\y = 0
					\bound\w = 0
					\bound\h = 0
				EndIf
			Default
				\bound\x = \x
				\bound\y = \y
				\bound\w = \w ; -\x
				\bound\h = \h ; -\y
		EndSelect
		
		\bounds_known = #True
		
		If rep : ReportRectF(@\bound,0,"ID: "+\id) : EndIf
		;\bound\x = x
		;\bound\y = y
		;\bound\w = w
		;\bound\h = h
	EndWith
	
EndProcedure

Macro KnowBounds(s,e,rep=#False)
	If Not svgmem(s)\element(e)\bounds_known
		CalculateSVGElementBounds(s,e,rep)
		;If rep=#True
		;    DebugRectF(@svgmem(s)\element(e)\bound)
		;EndIf
	EndIf
EndMacro

Procedure.b GetSVGElementBoundingBox(s.i,e.i,*b.RectF,includestroke.b=#False)
	
	With svgmem(s)\element(e)
		KnowBounds(s,e)
		*b\x = \bound\x
		*b\y = \bound\y
		*b\w = \bound\w
		*b\h = \bound\h
		;If *b\w<0 : *b\w=0-*b\w : EndIf
		;If *b\h<0 : *b\h=0-*b\h : EndIf
		
		If includestroke
			If \strokewidth>0
				*b\x - (\strokewidth/2)
				*b\y - (\strokewidth/2)
				*b\w + (\strokewidth)
				*b\h + (\strokewidth)
			EndIf
		EndIf
	EndWith
	
EndProcedure

Procedure.b GetSVGElementCenter(s.i,e.i,*c.PointF)
	With svgmem(s)\element(e)
		Select \type
			Case #SVG_Element_GroupCloser, #SVG_Element_CustomCode, #SVG_Element_Comment
				ProcedureReturn #False
			Case #SVG_Element_Circle, #SVG_Element_Ellipse
				*c\x = \cx
				*c\y = \cy
				ProcedureReturn #True
		EndSelect
		KnowBounds(s,e)
		;ReportRectF(@\bound)
		*c\x = \bound\x+(\bound\w/2)
		*c\y = \bound\y+(\bound\h/2)
		;ReportPointF(*c)
		ProcedureReturn #True
	EndWith
EndProcedure


Macro MasterForAllElements
	;R("ALL ELEMENT MASTER")
	master.RectF
	master\x = 999999
	master\y = 999999
	master\w = -999999
	master\h = -999999
	SVGInternal_ForEachElement(s,e)
	If Not svgmem(s)\element(e)\in_use : Continue : EndIf
	Select svgmem(s)\element(e)\type
		Case #SVG_Element_Group, #SVG_Element_GroupCloser, #SVG_Element_CustomCode, #SVG_Element_Comment
			Continue
	EndSelect
	
	GetSVGElementBoundingBox(s,e,@bb.RectF,#True)
	If bb\x>0
		DefeatThis(master\x,bb\x)
		BeatThis(master\w,bb\x+bb\w)
	Else
		DefeatThis(master\x,bb\x)
		BeatThis(master\w,bb\w+bb\x)
	EndIf
	If bb\y>0
		DefeatThis(master\y,bb\y)
		BeatThis(master\h,bb\y+bb\h)
	Else
		DefeatThis(master\y,bb\y)
		BeatThis(master\h,bb\h+bb\y)
	EndIf
	;ReportRectF(@bb,0,"Element #"+Str(e))
	;PolyRect(@bb,#Yellow);,svgmem(s)\element(e)\fillvalue)
Next e
;If master\x>0
master\w-master\x
;EndIf
;If master\y>0
master\h-master\y
;EndIf
;ReportRectF(@master)
EndMacro

Macro SafeMasterForAllElements
	;R("ALL ELEMENT MASTER")
	master.RectF
	master\x = 999999
	master\y = 999999
	master\w = -999999
	master\h = -999999
	SVGInternal_ForEachElement(s,e)
	If Not svgmem(s)\element(e)\in_use : Continue : EndIf
	Select svgmem(s)\element(e)\type
		Case #SVG_Element_Group, #SVG_Element_GroupCloser, #SVG_Element_CustomCode, #SVG_Element_Comment
			Continue
	EndSelect
	
	GetSVGElementBoundingBox(s,e,@bb.RectF,#True)
	If bb\x>0
		DefeatThis(master\x,bb\x)
		BeatThis(master\w,bb\x+bb\w)
	Else
		DefeatThis(master\x,bb\x)
		BeatThis(master\w,bb\w+bb\x)
	EndIf
	If bb\y>0
		DefeatThis(master\y,bb\y)
		BeatThis(master\h,bb\y+bb\h)
	Else
		DefeatThis(master\y,bb\y)
		BeatThis(master\h,bb\h+bb\y)
	EndIf
	;ReportRectF(@bb,0,"Element #"+Str(e))
	;ReportRectF(@master,0,"Master")
	;PolyRect(@bb,#Yellow);,svgmem(s)\element(e)\fillvalue)
Next e
If master\x>0
	;master\w-master\x
EndIf
If master\y>0
	;master\h-master\y
EndIf
master\x = 0
master\y = 0
;ReportRectF(@master)
EndMacro


Procedure.b SVGElementTypeIsFontable(t)
	Select t
		Case #SVG_Element_Text, #SVG_Element_WrappedText, #SVG_Element_TextPath, #SVG_Element_Path, #SVG_Element_Group
			ProcedureReturn #True
		Default
			ProcedureReturn #False
	EndSelect
EndProcedure


Procedure.i FirstAvailableElementSlot(s.i,forcenew.b=#False)
	
	With svgmem(s)
	
	If Not forcenew
		For e = 1 To \elements
			If Not \element(e)\in_use
				; check we're not inside a group
				grouplevel=0
				For a = 1 To e-1
					If \element(a)\type=#SVG_Element_Group
						grouplevel+1
					EndIf
					If \element(a)\type=#SVG_Element_GroupCloser
						grouplevel-1
					EndIf
				Next a
				If grouplevel<1
					ProcedureReturn e
				EndIf
			EndIf
		Next e
	EndIf
	
	arrsize = ArraySize(\element(),1)
	needs = \elements+1
	If arrsize<needs
		;R("CURRENT ARRSIZE: "+Str(arrsize)+c13+"NEED: "+Str(needs)+c13+"RE-DIMMING TO: "+Str(arrsize+#SVG_ArrayJump))
		;ReDim \element(arrsize+#SVG_ArrayJump)
		If \elements<1 Or arrsize<1
			Dim svgmem(s)\element(needs+2)
		Else
			ReDim svgmem(s)\element(needs+2)
		EndIf
		If ArraySize(\element(),1) = -1
			R("SVG: ERROR REDIMMING ELEMENT ARRAY.")
			ProcedureReturn 0
		EndIf
	EndIf
	
	\elements = needs
	ProcedureReturn \elements
	
	EndWith
	
EndProcedure


Procedure.i AddSVGElement(s.i,type.b,id.s,x.f,y.f,w.f,h.f,color.i=-#PB_Ignore,content.s="",imagenumber.i=0,imagecodec.i=0,filename.s="",linethickness.f=1)
	If Not svgdata\s : R("CAN'T CREATE ELEMENT: NO CURRENT SVG.") : EndIf
	If Not IsSVG(s) : R("IS NOT SVG") : ProcedureReturn #False : EndIf
	
	If type=#SVG_Element_GroupCloser
		e = FirstAvailableElementSlot(s,#True)
	Else
		e = FirstAvailableElementSlot(s)
	EndIf
	If ArraySize(svgmem(s)\element(),1)<e : ReDim svgmem(s)\element(e) : EndIf
	If Not e : R("NO AVAILABLE ELEMENT SLOT") : ProcedureReturn #False : EndIf
	;InitializeStructure(@svgmem(s)\element(e),SVG_Element_Structure)
	
	If type<>#SVG_Element_GroupCloser
		svgmem(s)\bounds_known = #False
	EndIf
	svgdata\e = e
	With svgmem(s)\element(e)
		\in_use = #True
		\type = type
		\id = id
		\filename = filename
		Select \type
			Case #SVG_Element_Circle
				\cx = x
				\cy = y
				\rx = w
				\ry = w
				
			Case #SVG_Element_Ellipse
				\cx = x
				\cy = y
				\rx = w
				\ry = h
				
			Case #SVG_Element_Base64Image
				\x = x
				\y = y
				\w = w
				\h = h
				\base64data = content
			Case #SVG_Element_PBImage
				\x = x
				\y = y
				\w = w
				\h = h
				If \w<1
					\w = ImageWidth(imagenumber)
				EndIf
				If \h<1
					\h = ImageHeight(imagenumber)
				EndIf
				\imagenumber = imagenumber
				\imagecodec = imagecodec
				If Nots \filename
					\filename = Str((Random(33)+1)*ElapsedMilliseconds())
				Else
					If \filename=#SVG_Base64
						\store_internally = #True
					Else
						\store_internally = #False
					EndIf
				EndIf
				
			Case #SVG_Element_Path
				If content
					AddToSVGPath(content,@svgmem(s)\element(e)\path)
					content = ""
				EndIf
				
			Case #SVG_Element_Line
				\path\pathpoints=2
				
				*e.SVG_Element_Structure = @svgmem(s)\element(e)
				Dim *e\path\pathpoint(2)
				
				\path\pathpoint(1)\mode = #SVG_Path_SegmentStart
				\path\pathpoint(1)\x = x
				\path\pathpoint(1)\y = y
				\path\pathpoint(1)\intermediary_points = 0
				\path\pathpoint(2)\mode = #SVG_Path_Straight
				\path\pathpoint(2)\x = x+w
				\path\pathpoint(2)\y = y+h
				\path\pathpoint(2)\intermediary_points = 0
				
			Default
				\x = x
				\y = y
				\w = w
				\h = h
				
		EndSelect
		
		If \type = #SVG_Element_Line Or \type=#SVG_Element_Polyline
			\strokewidth = Beat(1,linethickness)
			\strokecolor = color
			If color=-1
				strokecolor = svgdata\FrontColor
			EndIf
		EndIf
		
		\visibility = #True
		\opacity = 255
		\fillmode = #SVG_FillMode_Color
		If color=#PB_Ignore
			\fillvalue = svgdata\FrontColor
		Else
			\fillvalue = color
		EndIf
		;R("COLOUR: "+Str(\color))
		;If strokecolor>-1
		;    \strokecolor = strokecolor
		;Else
		;    \strokecolor = svgdata\BackColor
		;EndIf
		\content = content
		
		
		If SVGElementTypeIsFontable(\type)
			\fontname = svgdata\FontName
			\fontsize = svgdata\FontSize
		EndIf
		
		
	EndWith
	ProcedureReturn e
EndProcedure


Procedure.i FirstAvailableGradientSlot(s.i)
	
	With svgmem(s)
		If \gradients
			For a = 1 To \gradients
				If Not \gradient(a)\in_use
					R("USING OLD GRADIENT: "+Str(a))
					ProcedureReturn a
				EndIf
			Next a
		EndIf
		
		arrsize = ArraySize(\gradient(),1)
		needs = \gradients+1
		If arrsize<needs
			;R("CURRENT ARRSIZE: "+Str(arrsize))
			If \gradients<1 Or arrsize<1
				Dim \gradient(needs+2)
			Else
				ReDim \gradient(needs+2)
			EndIf
			If ArraySize(\gradient(),1) = -1
				ProcedureReturn 0
			EndIf
		EndIf
		
		\gradients = needs
		ProcedureReturn \gradients
		
	EndWith
	
EndProcedure

Procedure.i AddSVGGradient(s.i,type,id.s,x.f,y.f,w.f,h.f)
	
	ng = FirstAvailableGradientSlot(s)
	If Not ng
		ProcedureReturn 0
	EndIf
	With svgmem(s)\gradient(ng)
		\in_use = #True
		\type = type
		If id
			\id = id
		Else
			\id = RandomID(s)
		EndIf
		\x = x
		\y = y
		\w = w
		\h = h
		ResetList(\Point())
	EndWith
	
	svgdata\g = ng
	ProcedureReturn ng
	
EndProcedure

Procedure.i FirstAvailablePatternSlot(s.i)
	
	With svgmem(s)
	
	If \patterns
		For a = 1 To \patterns
			If Not \pattern(a)\in_use
				ProcedureReturn a
			EndIf
		Next a
	EndIf
	
	arrsize = ArraySize(\pattern(),1)
	needs = \patterns+1
	If arrsize<needs
		R("CURRENT ARRSIZE: "+Str(arrsize))
		If \patterns<1 Or arrsize<1
			Dim \pattern(needs+2)
		Else
			ReDim \pattern(needs+2)
		EndIf
		If ArraySize(\pattern(),1) = -1
			ProcedureReturn 0
		EndIf
	EndIf
	
	\patterns = needs
	ProcedureReturn \patterns
EndWith

EndProcedure

Procedure.i AddSVGPattern(s.i,w.f,h.f,mode,id.s)
	
	np = FirstAvailablePatternSlot(s)
	If Not np
		ProcedureReturn 0
	EndIf
	With svgmem(s)\pattern(np)
		\mode = mode
		If id
			\id = id
		Else
			\id = RandomID(s)
		EndIf
		\w = w
		\h = h
	EndWith
	
	svgdata\p = np
	ProcedureReturn np
	
EndProcedure




Procedure.b ReportSVGDimensions(s.i)
	
	If Not IsSVG(s) : ProcedureReturn #False : EndIf
	
	t.s
	With svgmem(s)
		t = "         X BOUND: "+Str(\bound\x)+c13
		t + "         Y BOUND: "+Str(\bound\y)+c13
		t + "HORIZONTAL BOUND: "+Str(\bound\w)+c13
		t + "  VERTICAL BOUND: "+Str(\bound\h)+c13
		t + c13
		
		t + "     ARBITRARY CROP X: "+Str(\visible\x)+c13
		t + "     ARBITRARY CROP Y: "+Str(\visible\y)+c13
		t + " ARBITRARY CROP WIDTH: "+Str(\visible\w)+c13
		t + "ARBITRARY CROP HEIGHT: "+Str(\visible\h)+c13
	EndWith
	
	R(t)
	
EndProcedure



CompilerIf Defined(MonthName,#PB_Procedure)=0
	Procedure.s MonthName(num.b)
		
		Select num
			Case 1
				ProcedureReturn "January"
			Case 2
				ProcedureReturn "February"
			Case 3
				ProcedureReturn "March"
			Case 4
				ProcedureReturn "April"
			Case 5
				ProcedureReturn "May"
			Case 6
				ProcedureReturn "June"
			Case 7
				ProcedureReturn "July"
			Case 8
				ProcedureReturn "August"
			Case 9
				ProcedureReturn "September"
			Case 10
				ProcedureReturn "October"
			Case 11
				ProcedureReturn "November"
			Case 12
				ProcedureReturn "December"
		EndSelect
		
	EndProcedure
CompilerEndIf
Procedure.s SVGToCoolDate(dl.i)
	
	d = Day(dl)
	m = Month(dl)
	y = Year(dl)
	mn.s = Left(MonthName(m),3)
	
	ProcedureReturn RSet(Str(d),2,c32)+"/"+mn+"/"+Str(y)
	
EndProcedure



Declare.b SVG_IDIsUsed(id.s,s=0)

Prototype.s SVGPassString(str.s)

Procedure.s GetAutomaticCopyName(s.i,id.s)
	If Nots id
		id = "unnamed"
	EndIf
	Repeat
		copy+1
		nid.s = id+"_"+Str(copy)
		If Not SVG_IDIsUsed(nid,s)
			Break
		EndIf
	ForEver
	
EndProcedure

Procedure.s RandomID(s.i)
	Repeat
		id.s = LCase(RandomLetters(8))
	Until Not SVG_IDIsUsed(id)
	ProcedureReturn id
EndProcedure


Procedure.i CloserForSVGGroup(s.i,g.i)
	
	If svgmem(s)\element(g)\type=#SVG_Element_Group
		grouplevel=1
	Else
		ProcedureReturn 0    
	EndIf
	
	;R("S: "+Str(s)+c13+"GROUP: "+Str(g)+c13+"ELEMENTS: "+Str(svgmem(s)\elements))
	For e = g+1 To svgmem(s)\elements
		If svgmem(s)\element(e)\type=#SVG_Element_Group
			grouplevel+1
		EndIf
		If svgmem(s)\element(e)\type=#SVG_Element_GroupCloser
			grouplevel-1
			If grouplevel=0
				;R("FOUND CLOSER: "+Str(e))
				ProcedureReturn e
			EndIf
		EndIf
	Next e
	R("S: "+Str(s)+c13+"GROUP: "+Str(g)+c13+"ELEMENTS: "+Str(svgmem(s)\elements))
	R("DID NOT FIND CLOSER: "+Str(e))
	
EndProcedure

Procedure.i OpenerForSVGGroup(s.i,g.i)
	
	grouplevel=1
	;R("S: "+Str(s)+c13+"ELEMENTS: "+Str(svgmem(s)\elements))
	For e = g-1 To 1 Step -1
		If svgmem(s)\element(e)\type=#SVG_Element_GroupCloser
			grouplevel+1
		EndIf
		If svgmem(s)\element(e)\type=#SVG_Element_Group
			grouplevel-1
			If grouplevel=0
				ProcedureReturn e
			EndIf
		EndIf
	Next e
	
EndProcedure




Macro RYBX
	R("SYBX: "+Str(sybx)+c13+"YBX: "+Str(ybx)+c13+"ID: "+svgmem(sybx)\element(ybx)\id+c13+"ARRAY SIZE: "+Str(ArraySize(svgmem(sybx)\element(ybx)\path\pathpoint())))
	;R(Str(svgmem(sybx)\element(ybx)\pathpoint(1)\x)+","+Str(svgmem(sybx)\element(ybx)\pathpoint(1)\y))
EndMacro

Declare.i SVGInternal_CopyGroupElement(s.i,g1.i,*namechanger.SVGPassString=0)


Procedure.b CopySVGElementStructure(*e1.SVG_Element_Structure,*e2.SVG_Element_Structure)
	;/ BUG FIX BECAUSE COPYSTRUCTURE CORRUPTS ARRAYS INSIDE THE ORIGINAL STRUCTURE
		cs = *e1\path\pathpoints
		Dim inter_choice.SVG_PathPoint_Structure(cs)
		CopyArray(*e1\path\pathpoint(),inter_choice())
		; information is now in the "inter" (dummy) array
		
		;R("el 1 id: "+svgmem(s)\element(e1)\id)
		;R("el 2 id: "+svgmem(s)\element(e2)\id)
		CopyStructure(*e1,*e2,SVG_Element_Structure)
		
		; now copy array information into slave and back into master
		Dim *e1\path\pathpoint(cs)
		CopyArray(inter_choice(),*e1\path\pathpoint())
		Dim *e2\path\pathpoint(cs)
		CopyArray(inter_choice(),*e2\path\pathpoint())
		
		FreeArray(inter_choice())
	;/
EndProcedure

Procedure.i SVGInternal_CopyElement(s.i,e1.i,*namechanger.SVGPassString=0)
	
	;RYBX
	If svgmem(s)\element(e1)\type = #SVG_Element_Group
		;R("COPYING A GROUP ELEMENT")
		ProcedureReturn SVGInternal_CopyGroupElement(s,e1,*namechanger)
		;e2 = FirstAvailableElementSlot(s,#True)
	Else
		;e2 = FirstAvailableElementSlot(s)
		;R("GOING TO ADD A DUMMY BOX")
		e2 = AddSVGElement(s,#SVG_Element_Box,"",0,0,0,0)
		;R("E2: "+Str(e2))
		RYBX
	EndIf
	
	;R("COPYING ELEMENT."+c13+"SLOT "+Str(e1)+" TO SLOT "+Str(e2)+c13+"TYPE: "+SVGElementTypeName(svgmem(s)\element(e1)\type)+c13+c13+"ELEMENT 1 PATHPOINT ARRAY SIZE: "+Str(ArraySize(svgmem(s)\element(e1)\pathpoint(),1))+c13+c13+"SVG ELEMENT ARRAY SIZE: "+Str(ArraySize(svgmem(s)\element(),1)))
	
	
	If e2>ArraySize(svgmem(s)\element(),1)
		ReDim svgmem(s)\element(e2)
	EndIf
	
	
	
	
	;/ BUG FIX BECAUSE COPYSTRUCTURE CORRUPTS ARRAYS INSIDE THE ORIGINAL STRUCTURE
	CopySVGElementStructure(@svgmem(s)\element(e1),@svgmem(s)\element(e2))
	;/
	
	
	
	
	;svgmem(s)\element(e2)\in_use = #True
	
	;RYBX
	If svgmem(s)\element(e1)\id
		If *namechanger
			svgmem(s)\element(e2)\id = *namechanger(svgmem(s)\element(e1)\id)
		Else
			svgmem(s)\element(e2)\id = GetAutomaticCopyName(s,svgmem(s)\element(e1)\id)
		EndIf
	EndIf
	
	ProcedureReturn e2
	
EndProcedure

Procedure.i SVGInternal_CopyGroupElement(s.i,g1.i,*namechanger.SVGPassString=0)
	
	If g1<1
		ProcedureReturn #False
	EndIf
	If svgmem(s)\element(g1)\type <> #SVG_Element_Group
		ProcedureReturn #False
	EndIf
	
	;R("CopyElement does not currently work with group elements.") : ProcedureReturn #False
	
	gc = CloserForSVGGroup(s,g1)
	If Not gc
		R("CopyGroupElement."+c13+c13+"ERROR: THIS GROUP HAS NOT BEEN CLOSED!")
		ProcedureReturn #False
	EndIf
	
	g2 = AddSVGElement(s,#SVG_Element_Group,"",0,0,0,0)
	CopySVGElementStructure(@svgmem(s)\element(g1),@svgmem(s)\element(g2))
	If svgmem(s)\element(g1)\id
		If *namechanger
			svgmem(s)\element(g2)\id = *namechanger(svgmem(s)\element(g1)\id)
		Else
			svgmem(s)\element(g2)\id = GetAutomaticCopyName(s,svgmem(s)\element(g1)\id)
		EndIf
	EndIf
	
	For e = g1+1 To gc
		If svgmem(s)\element(e)\type = #SVG_Element_Group
			SVGInternal_CopyGroupElement(s,e,*namechanger)
		Else
		e2 = AddSVGElement(s,#SVG_Element_Box,"",0,0,0,0)
		CopySVGElementStructure(@svgmem(s)\element(e),@svgmem(s)\element(e2))
		;SVGInternal_CopyElement(s,e,*namechanger)
		If svgmem(s)\element(e)\id
			If *namechanger
				svgmem(s)\element(e2)\id = *namechanger(svgmem(s)\element(e)\id)
			Else
				svgmem(s)\element(e2)\id = GetAutomaticCopyName(s,svgmem(s)\element(e)\id)
			EndIf
		EndIf
		EndIf
	Next e
	
	ProcedureReturn g2
	
EndProcedure




Procedure.b SVGInternal_RemoveElement(s.i,e.i)
	
	If svgmem(s)\element(e)\type = #SVG_Element_Group
		SVGInternal_ForEachElementInSVGGroup(s,e,a)
			SVGInternal_RemoveElement(s,a)
		Next a
		SVGInternal_RemoveElement(s,CloserForSVGGroup(s,e))
	EndIf
	
	ClearStructure(@svgmem(s)\element(e),SVG_Element_Structure)
	
	ProcedureReturn #True
	
EndProcedure




Procedure.i SVGInternal_SeparatePathParts_V01(arr.s,Array item.s(1))
	
	;arrsize = ArraySize(item(),1)
	Trim(arr)
	EnsureThisNotStart(arr,"d=")
	EnsureThisNotStart(arr,"D=")
	
	L = Len(arr)
	build.s
	mode.s
	spacespan = 0
	
	Repeat
		a+1
		t.s = Mid(arr,a,1)
		Select LCase(t)
			Case ","
				build+t
				space+1
			Case c32
				space+1
				If space=>spacespan And build
					items+1
					;If items=#MaxSVGPathPoints : Break : EndIf
					arrsize=ArraySize(item(),1) : If arrsize<items : ReDim item(arrsize+10) : EndIf
					item(items) = mode+c32+build
					build=""
					space=0
				Else
					build+t
				EndIf
			Case "m", "l", "h", "v", "c", "s", "q", "t", "a", "z", "d"
				; end previous
				If build And mode
					items+1
					;If items=#MaxSVGPathPoints : Break : EndIf
					arrsize=ArraySize(item(),1) : If arrsize<items : ReDim item(arrsize+10) : EndIf
					item(items) = mode+c32+build
				EndIf
				; start new
				If LCase(t)="z"
					items+1
					arrsize=ArraySize(item(),1) : If arrsize<items : ReDim item(arrsize+10) : EndIf
					item(items) = t
					;Break ; could be multiple segments
				Else
					build=""
					space=0
					mode = t
					
					Select LCase(t)
						Case "m"
							spacespan = 2
						Case "l"
							spacespan = 2
						Case "h"
							spacespan = 2
						Case "v"
							spacespan = 2
						Case "c"
							spacespan = 6
						Case "s"
							spacespan = 4
						Case "q"
							spacespan = 4
						Case "t"
							spacespan = 4
						Case "a"
							spacespan = 7
					EndSelect
				EndIf
			Default ; number, hyphen
				build+t
		EndSelect
	Until a=>L
	
	ProcedureReturn items
	
EndProcedure

Macro NewPathItem_V01
					items+1
					arrsize=ArraySize(item(),1) : If arrsize<items : ReDim item(arrsize+10) : EndIf
EndMacro

Procedure.b CharacterIsNumeric(t.s)
	For a = 0 To 9
		If t=Str(a)
			ProcedureReturn #True
		EndIf
	Next a
	If t="-" Or t="."
		ProcedureReturn #True
	EndIf
	ProcedureReturn #False
EndProcedure

Procedure.i SVGInternal_SeparatePathParts_V02(arr.s,Array item.s(1))
	
	;arrsize = ArraySize(item(),1)
	Trim(arr)
	EnsureThisNotStart(arr,"d=")
	EnsureThisNotStart(arr,"D=")
	
	L = Len(arr)
	build.s = ""
	mode.s = ""
	spacespan = 0
	lastmode.s = ""
	lastt.s = ""
	
	Repeat
		a+1
		t.s = Mid(arr,a,1)
		Select LCase(t)
			Case ","
				build+t
				space+1
			Case c32
				space+1
				If space=>spacespan And build
					NewPathItem_V01
					item(items) = mode+c32+build
					build = ""
					space = 0
				Else
					build+t
				EndIf
			Default;Case "m", "l", "h", "v", "c", "s", "q", "t", "a", "z", "d"
				numeric.b = CharacterIsNumeric(t)
				;R(t+c13+"numeric: "+Str(numeric))
				
				; end previous
				If build
					If mode And Not numeric
						NewPathItem_V01
						item(items) = mode+c32+build
					EndIf
				Else
					If numeric
						mode = lastmode
						t = mode
					EndIf
				EndIf
				
				; start new
				If LCase(t)="z"
					NewPathItem_V01
					item(items) = t
					;Break ; could be multiple segments
				Else
					build = ""
					space = 0
					mode = t
					lastmode = mode
					
					Select LCase(t)
						Case "m"
							spacespan = 2
						Case "l"
							spacespan = 2
						Case "h"
							spacespan = 2
						Case "v"
							spacespan = 2
						Case "c"
							spacespan = 6
						Case "s"
							spacespan = 4
						Case "q"
							spacespan = 4
						Case "t"
							spacespan = 4
						Case "a"
							spacespan = 7
					EndSelect
				EndIf
			;Default ; number, hyphen
			;	R("DEF: "+t)
			;	build+t
		EndSelect
		lastt = t
	Until a=>L
	
	ProcedureReturn items
	
EndProcedure


Procedure.s EmulateCase(basis.s,newletter.s)
	If LCase(basis)=basis
		ProcedureReturn LCase(newletter)
	Else
		ProcedureReturn UCase(newletter)
	EndIf
EndProcedure

Structure PathParseStructure
	mode.s
	info.s
	relative.b
	definite.b
EndStructure

Macro NewPathItem(m,i,changed)
	AddElement(item())
	item()\mode = m
	item()\info = i
	item()\definite = changed
	If LCase(m) = m
		item()\relative = #True
	EndIf
EndMacro

Procedure.i SVGInternal_SeparatePathParts_V03(arr.s,List item.PathParseStructure())
	
	ClearList(item())
	
	NewMap length.i()
	length("m") = 2
	length("l") = 2
	length("h") = 1
	length("v") = 1
	length("c") = 6
	length("s") = 4
	length("q") = 4
	length("t") = 4
	length("a") = 7
	length("z") = 0
	
	
	arr = Trim(arr)
	EnsureThisNotStart(arr,"d=")
	EnsureThisNotStart(arr,"D=")
	arr = RemoveString(arr,c34)
	arr = RemoveString(arr,c39)
	arr = Trim(arr)
	arr = ReplaceString(arr,#cma,c32)
	arr = RecursiveReplaceString(arr,"  "," ")
	
	;Debug "         1         2         3         4         5         6         7         8         9         10"
	;Debug "123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 1234567890"
	;Debug arr
	;Debug ""
	
	L = Len(arr)
	t.s = ""
	
	mode.s = "m"
	lmode.s = LCase(mode)
	
	a = 0
	Repeat
		a+1
		changemode.b = #False
		t = Mid(arr,a,1)
		;Debug "T: "+t
		lt.s = LCase(t)
		Select lt
			Case "m", "l", "h", "v", "c", "s", "q", "t", "a", "z"
				R("MODE CHANGE: "+t)
				mode = t
				lmode = LCase(mode)
				changemode = #True
				a+2
		EndSelect
		examine.s = Mid(arr,a,L)
		
		jump = 0
		While Mid(examine,jump,1) = " "
			jump+1
		Wend
		;Debug "JUMP: "+Str(jump)
		If jump
			;jump-1
			a+jump
			examine = Mid(examine,jump,Len(examine)-jump)
			;Debug "EXATRIM: *"+examine+"*"
		EndIf
		Debug "EXAMINE: *"+examine+"*"
		EnsureThisEnd(examine," ")
		
		build.s = ""
		For bit = 1 To length(lmode)
			build + StringField(examine,bit," ") + " "
		Next bit
		;Debug "BUILD: *"+build+"*"
		NewPathItem(mode,build,#False)
		a + Len(build)-1
		Debug "A: "+Str(a)
	Until a=>L
	
	ProcedureReturn items
	
EndProcedure


Procedure.b SetPathItem(*t.PathParseStructure,mode.s,info.s,definite.b=#False)
	*t\mode = mode
	*t\info = info
	*t\definite = definite
	If LCase(mode) = mode
		*t\relative = #True
	Else
		Debug "ABSOLUTE"
		*t\relative = #False
	EndIf
EndProcedure

Procedure.i SVGInternal_SeparatePathParts(arr.s,List item.PathParseStructure())
	
	ClearList(item())
	
	NewMap length.i()
	length("m") = 2
	length("l") = 2
	length("h") = 1
	length("v") = 1
	length("c") = 6
	length("s") = 4
	length("q") = 4
	length("t") = 4
	length("a") = 7
	length("z") = 0
	
	
	arr = Trim(arr)
	EnsureThisNotStart(arr,"d=")
	EnsureThisNotStart(arr,"D=")
	arr = RemoveString(arr,c34)
	arr = RemoveString(arr,c39)
	arr = ReplaceString(arr,#cma,c32)
	ForEach length()
		letter.s = MapKey(length())
		arr = ReplaceString(arr,letter," "+letter+" ")
		letter = UCase(letter)
		arr = ReplaceString(arr,letter," "+letter+" ")
	Next
	arr = RecursiveReplaceString(arr,"  "," ")
	arr = Trim(arr)
	
	
	;Debug "         1         2         3         4         5         6         7         8         9         10"
	;Debug "123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 1234567890"
	;Debug arr
	;Debug ""
	
	L = Len(arr)
	t.s = ""
	
	mode.s = "m"
	lmode.s = LCase(mode)
	
	a = 0
	Repeat
		a+1
		changemode.b = #False
		t = Mid(arr,a,1)
		;Debug "T: "+t
		lt.s = LCase(t)
		Select lt
			Case "m", "l", "h", "v", "c", "s", "q", "t", "a", "z"
				;R("MODE CHANGE: "+t)
				mode = t
				lmode = LCase(mode)
				changemode = #True
				a+2
		EndSelect
		examine.s = Mid(arr,a,L)
		
		jump = 0
		While Mid(examine,jump,1) = " "
			jump+1
		Wend
		;Debug "JUMP: "+Str(jump)
		If jump
			;jump-1
			a+jump
			examine = Mid(examine,jump,Len(examine)-jump)
			;Debug "EXATRIM: *"+examine+"*"
		EndIf
		;Debug "EXAMINE: *"+examine+"*"
		EnsureThisEnd(examine," ")
		
		build.s = ""
		For bit = 1 To length(lmode)
			build + StringField(examine,bit," ") + " "
		Next bit
		;Debug "BUILD: *"+build+"*"
		NewPathItem(mode,build,changemode)
		a + Len(build)-1
		;Debug "A: "+Str(a)
	Until a=>L
	
	Repeat
		changed.b = #False
		ResetList(item())
		lastmode.s = ""
		While NextElement(item())
			mode.s = item()\mode
			If item()\definite=#False And LCase(lastmode)="m" And LCase(mode)="m"
				changed = #True
				newmode.s = EmulateCase(lastmode,"l")
				Debug "NEWMODE: "+lastmode+" "+newmode
				PreviousElement(item())
				While NextElement(item())
					If LCase(item()\mode)<>"m" Or item()\definite
						Break 2
					EndIf
					SetPathItem(@item(),newmode,item()\info)
				Wend
				lastmode = mode
				Break
			EndIf
			lastmode = mode
		Wend
	Until Not changed
	
	ProcedureReturn items
	
EndProcedure


Procedure.b AddToSVGPath(arr.s,*e.SVG_Path_Structure)
	
	NewList item.PathParseStructure()
	;R("ARR"+c13+c13+arr)
	SVGInternal_SeparatePathParts(arr,item())
	items = ListSize(item())
	
	cur = *e\pathpoints
	*e\pathpoints + items
	ReDim *e\pathpoint(*e\pathpoints)
	
	
	curx.f = 0
	cury.f = 0
	cm.s = ","
	a=0
	ForEach item()
		a+1
		With *e\pathpoint(a+cur)
			letter.s = item()\mode
			thisone.s = item()\info
			;R(thisone)
			
			
			fx.f = ValF(StringField(thisone,1,c32))
			fy.f = ValF(StringField(thisone,2,c32))
			\x = fx
			\y = fy
			;R(StrF(\x,0)+cm+StrF(\y,0))
			
			Select letter
					;   "m", "l", "h", "v", "c", "s", "q", "t", "a"
				Case "m", "M"
					\mode = #SVG_Path_SegmentStart
				Case "l", "L" ; straight line to
					\mode = #SVG_Path_Straight
				Case "h", "H" ; horizontal straight line to
					\mode = #SVG_Path_Straight
					\x = fx
					\y = cury
				Case "v", "V" ; vertical straight line to
					\mode = #SVG_Path_Straight
					\x = curx
					\y = fx ; first number given, since only one will be given
				Case "c", "C"
					\mode = #SVG_Path_CubicBezier
				Case "s", "S"
					\mode = #SVG_Path_SmoothCubicBezier
				Case "q", "Q"
					\mode = #SVG_Path_QuadraticBezier
				Case "t", "T"
					\mode = #SVG_Path_SmoothQuadraticBezier
				Case "a", "A"
					\mode = #SVG_Path_EllipticalArc
				Case "z", "Z"
					\mode = #SVG_Path_SegmentEnd
				Default
					;\mode = #SVG_Path_Straight
					;letter = "l"
					R("Don't know letter: *"+letter+"*")
			EndSelect
			
			
			; get intermediary points
			Select letter
				Case "c", "C", "s", "S", "q", "Q"
					;thisone = RemoveFieldsFromStart(thisone,c32,2) ; get rid of main (destination) coordinates
					more = CountString(thisone,c32)
					more/2
					\intermediary_points = 0
					Dim \intermediary_point(more)
					For i = 1 To more-1
						\intermediary_points+1
						inp.s = StringField(thisone,1,c32)+c32+StringField(thisone,2,c32)+c32
						\intermediary_point(\intermediary_points)\x = ValF(StringField(inp+c32,1,c32))
						\intermediary_point(\intermediary_points)\y = ValF(StringField(inp+c32,2,c32))
						;R("INP: "+inp)
						
						EnsureThisNotStart(thisone,inp)
					Next i
					\x = ValF(StringField(thisone,1,c32))
					\y = ValF(StringField(thisone,2,c32))
				Case "a", "A"
					\rx = ValF(StringField(thisone,1,c32))
					\ry = ValF(StringField(thisone,2,c32))
					\x_rotation = ValF(StringField(thisone,3,c32))
					\large_arc = Val(StringField(thisone,4,c32))
					\sweep = Val(StringField(thisone,5,c32))
					\x = ValF(StringField(thisone,6,c32))
					\y = ValF(StringField(thisone,7,c32))
					;thisone = RemoveFieldsFromStart(thisone,c32,2) ; get rid of main (destination) coordinates
					;R(thisone)
				Default
					\intermediary_points = 0
			EndSelect
			
			
			; relative coordinates...
			If a>1
				Select letter
					Case "m"
						\x = curx+fx
						\y = cury+fy
					Case "l"
						\x = curx+fx
						\y = cury+fy
					Case "h"
						\x = curx+fx
						\y = cury
					Case "v"
						\x = curx
						\y = cury+fy
					Case "c", "s", "q", "t"
						\x = curx+\x
						\y = cury+\y
						If \intermediary_points
							For i = 1 To \intermediary_points
								\intermediary_point(i)\x = curx+\intermediary_point(i)\x
								\intermediary_point(i)\y = cury+\intermediary_point(i)\y
							Next i
						EndIf
					Case "a"
						\x = curx+\x
						\y = cury+\y
						;Default
						;    R("NOT RELATIVE")
				EndSelect
			EndIf
			
			;R(StrF(\x,0)+cm+StrF(\y,0))
			curx = \x
			cury = \y
		EndWith
	Next
	
	
	;If forceend
	;    *e\pathpoint(items)\mode = #SVG_Path_SegmentEnd
	;Else
	If *e\pathpoint(items)\mode <> #SVG_Path_SegmentEnd
		items+1
		
		arrsize=ArraySize(*e\pathpoint(),1) : If arrsize<items : ReDim *e\pathpoint(arrsize+5) : EndIf
		
		*e\pathpoint(items)\mode = #SVG_Path_SegmentEnd
	EndIf
	;EndIf
	
	ProcedureReturn #True
	
EndProcedure

Macro SVGInternal_CalibratePoint(n,zoom,offset)
	If zoom<>1 : n=n*zoom : EndIf
	If offset : n-offset : EndIf
EndMacro

Procedure.s SVGInternal_CalibratePath(s,*path.SVG_Path_Structure)
	
	;If Not svgmem(s)\element(e)\path\pathpoints : ProcedureReturn "" : EndIf
	rebuild.s
	
	;R("CalibrateSVGPath"+c13+"Output zoom: "+StrF(svgmem(s)\output_zoom))
	zoom.f = svgmem(s)\output_zoom
	;zoom=1
	xoffset.f = svgmem(s)\selviz\x
	yoffset.f = svgmem(s)\selviz\y
	
	EnsureEnoughPathPoints(*path,*path\pathpoints)
	For p = 1 To *path\pathpoints
		With *path\pathpoint(p)
			If \mode=#SVG_Path_SegmentEnd
				rebuild+"z"
				Continue
			EndIf
			x.f = \x
			y.f = \y
			SVGInternal_CalibratePoint(x,zoom,xoffset)
			SVGInternal_CalibratePoint(y,zoom,yoffset)
			;R(StrF(\x,0)+","+StrF(\y,0)+c13+c13+"ZOOM: "+StrF(zoom,0)+c13+"OFFSET: "+StrF(xoffset,0)+","+StrF(yoffset,0)+c13+c13+StrF(x,0)+","+StrF(y,0))
			Select \mode
				Case #SVG_Path_SegmentStart
					rebuild+"M "+StrF(x,#SVG_FloatAccuracy)+","+StrF(y,#SVG_FloatAccuracy)+" "
					Continue
				Case #SVG_Path_Straight
					rebuild+"L "+StrF(x,#SVG_FloatAccuracy)+","+StrF(y,#SVG_FloatAccuracy)+" "
					Continue
				Case #SVG_Path_EllipticalArc
					rebuild+"A "+StrF(\rx*zoom,#SVG_FloatAccuracy)+" "+StrF(\ry*zoom,#SVG_FloatAccuracy)+" "+StrF(\x_rotation,#SVG_FloatAccuracy)+" "+Str(\large_arc)+","+Str(\sweep)+" "+StrF(x,#SVG_FloatAccuracy)+","+StrF(y,#SVG_FloatAccuracy)+" "
					Continue
			EndSelect
			
			Select \mode
				Case #SVG_Path_CubicBezier
					rebuild + "C "
				Case #SVG_Path_SmoothCubicBezier
					rebuild + "S "
				Case #SVG_Path_QuadraticBezier
					rebuild + "Q "
				Case #SVG_Path_SmoothQuadraticBezier
					rebuild + "T "
			EndSelect
			dx.f = x
			dy.f = y
			If \intermediary_points
				For m = 1 To \intermediary_points
					x = \intermediary_point(m)\x
					y = \intermediary_point(m)\y
					SVGInternal_CalibratePoint(x,zoom,xoffset)
					SVGInternal_CalibratePoint(y,zoom,yoffset)
					rebuild+StrF(x,#SVG_FloatAccuracy)+","+StrF(y,#SVG_FloatAccuracy)+" "
				Next m
			EndIf
			rebuild + StrF(dx,#SVG_FloatAccuracy)+","+StrF(dy,#SVG_FloatAccuracy)+" "
		Next p
	EndWith
	
	ProcedureReturn rebuild
	
EndProcedure




Procedure.b PatternRecursivityOK(*pat.SVG_Pattern_Structure,bs.i)
	With *pat
		Select \mode
			Case #SVG_Pattern_DirectCode
				ProcedureReturn #True
			Case #SVG_Pattern_SVG
				If \svg = bs
					ProcedureReturn #False
				EndIf
				If svgmem(\svg)\patterns<1 : ProcedureReturn #True : EndIf
				For p = 1 To svgmem(\svg)\patterns
					If Not PatternRecursivityOK(@svgmem(\svg)\pattern(p),bs)
						ProcedureReturn #False
					EndIf
				Next p
				ProcedureReturn #True
		EndSelect
	EndWith
	ProcedureReturn #True
EndProcedure




Procedure.s CalibrateSVGPoly(s.i,e.i)
	
	With svgmem(s)\element(e)\path
		If \pathpoints<2
			ProcedureReturn ""
		EndIf
		
		zoom.f = svgmem(s)\output_zoom
		xoffset.f = svgmem(s)\selviz\x
		yoffset.f = svgmem(s)\selviz\y
		
		arr.s
		For p = 1 To \pathpoints
			x.f = \pathpoint(p)\x
			y.f = \pathpoint(p)\y
			SVGInternal_CalibratePoint(x,zoom,xoffset)
			SVGInternal_CalibratePoint(y,zoom,yoffset)
			
			arr+StrF(x,#SVG_FloatAccuracy)+","+StrF(y,#SVG_FloatAccuracy)+" "
		Next p
		EnsureThisNotEnd(arr," ")
		ProcedureReturn arr
	EndWith
	
EndProcedure

;"x1="+c34+SF(ex)+c34+" y1="+c34+SF(ey)+c34+" x2="+c34+SF(ew)+c34+" y2="+c34+SF(eh)+c34+" "+


Procedure.s CalibrateSVGLine(s.i,e.i)
	
	With svgmem(s)\element(e)\path
		If \pathpoints<2
			ProcedureReturn ""
		EndIf
		
		zoom.f = svgmem(s)\output_zoom
		xoffset.f = svgmem(s)\selviz\x
		yoffset.f = svgmem(s)\selviz\y
		
		arr.s
		For p = 1 To 2
			x.f = \pathpoint(p)\x
			y.f = \pathpoint(p)\y
			SVGInternal_CalibratePoint(x,zoom,xoffset)
			SVGInternal_CalibratePoint(y,zoom,yoffset)
			
			arr+"x"+Str(p)+"='"+StrF(x,#SVG_FloatAccuracy)+"' y"+Str(p)+"='"+StrF(y,#SVG_FloatAccuracy)+"' "
		Next p
		EnsureThisNotEnd(arr," ")
		ProcedureReturn arr
	EndWith
	
EndProcedure




Procedure.s ExtractXMLSegment(xml.s,n.s)
	xml = c32+xml+c32
	need.s = c32+n+"="+c34
	detect = FindString(xml,need,1)
	;R(xml+c13+c13+need+c13+Str(detect))
	If detect
		detectend = FindString(xml,c34,detect+Len(need))
		If detectend
			ex.s = Mid(xml,detect+1,detectend-detect)
			;R("."+ex+".")
			EnsureThisNotStart(ex,n+"=")
			ProcedureReturn Trim(ex,c34)
		EndIf
	EndIf
	
EndProcedure




Procedure.b SVGHasVisibleQuadrant(s.i)
	With svgmem(s)\visible
		If \x Or \y Or \w Or \h
			ProcedureReturn #True
		EndIf
		ProcedureReturn #False
	EndWith
EndProcedure


Procedure.b CalibrateSVGForOutputting(s.i)
	
	With svgmem(s)
		
		;\xml_defs = ""
		;\xml_els = ""
		\tooltiph = ""
		
		
		\w = \setw
		\h = \seth
		If \setw<1 Or \seth<1
			SafeMasterForAllElements
			If \setw<1
				\w = master\w
			EndIf
			If \seth<1
				\h = master\h
			EndIf
		EndIf
		
		If \output_zoom<0.000001 : \output_zoom=1 : EndIf
		
		\selviz\x = 0
		\selviz\y = 0
		\selviz\w = \w
		\selviz\h = \h
		If SVGHasVisibleQuadrant(s)
			;R("VIZ USED")
			CopyRectF(@\visible,@\selviz)
		EndIf
		;ReportSVGDimensions(s)
		;zoom = \output_zoom
		\output_w = svgmem(s)\output_zoom * (\selviz\w)
		\output_h = svgmem(s)\output_zoom * (\selviz\h)
		;R("S = "+Str(s)+c13+"OUTPUTW: "+StrF(\output_w)+c13+"OUTPUTH: "+StrF(\output_h))
	EndWith
	
EndProcedure








;-
;-
;- ANIMATORS

Enumeration 1 ; animator types
	#SVG_Animator_Attribute
	#SVG_Animator_Motion
	#SVG_Animator_Transform
EndEnumeration

Enumeration 1 ; transform types
	#SVG_Transform_Translate
	#SVG_Transform_Scale
	#SVG_Transform_Rotate
	#SVG_Transform_SkewX
	#SVG_Transform_SkewY
EndEnumeration

Enumeration ; restart modes
	#SMIL_Restart_Always
	#SMIL_Restart_WhenNotActive
	#SMIL_Restart_Never
EndEnumeration

#SMIL_Repeat_Indefinite = 0

Enumeration ; fill modes
	#SMIL_Fill_Remove
	#SMIL_Fill_Freeze
EndEnumeration

Enumeration ; rotation modes
	#SMIL_Rotate_AutoReverse = -1
	#SMIL_Rotate_Auto = 0
EndEnumeration

Enumeration ; path interpolation modes
	#SMIL_Path_Paced
	#SMIL_Path_Discrete
	#SMIL_Path_Linear
	#SMIL_Path_Spline
EndEnumeration

Macro AnimatorCheck
	s = svgdata\s
	e = svgdata\e
	If Not s Or Not e
		ProcedureReturn #False
	EndIf
EndMacro

Macro SVGInternal_AddAnimator
	AnimatorCheck
	svgmem(s)\element(e)\animators+1
	a = svgmem(s)\element(e)\animators
	If ArraySize(svgmem(s)\element(e)\animator())<a
		ReDim svgmem(s)\element(e)\animator(a+2)
	EndIf
EndMacro


Procedure.i SVG_AnimateAttribute(attrname.s,begin.s,duration.s,from_val.s,to_val.s,repeatcount.b=#SMIL_Repeat_Indefinite,repeatduration.b=#SMIL_Repeat_Indefinite)
	
	SVGInternal_AddAnimator
	
	With svgmem(s)\element(e)\animator(a)
		\in_use = #True
		\type = #SVG_Animator_Attribute
		\attr_name = attrname
		\from_val = from_val
		\to_val = to_val
		\begin = begin
		\duration = duration
		\repeat_count = repeatcount
		\repeat_duration = repeatduration
	EndWith
	
	ProcedureReturn a
	
EndProcedure

Procedure.i SVG_AnimateMotion(pathdata.s,begin.s,duration.s,repeatcount.b=#SMIL_Repeat_Indefinite,rotatemode.b=#SMIL_Rotate_Auto,calcmode.b=#SMIL_Path_Paced,keytimes.s="",keypoints.s="")
	;R("S: "+Str(svgdata\s)+c13+"E: "+Str(svgdata\e))
	
	SVGInternal_AddAnimator
	
	With svgmem(s)\element(e)\animator(a)
		\in_use = #True
		\type = #SVG_Animator_Motion
		\begin = begin
		\duration = duration
		\path\pathpoints = 0
		\rotate = rotatemode
		\repeat_count = repeatcount
		\repeat_duration = repeatduration
		\calc_mode = calcmode
		
		If keypoints And keytimes
			d1.s = "|"
			keys = Defeat(CountString(keypoints,d1),CountString(keytimes,d1))
			For k = 1 To keys
				AddElement(\key())
				\key()\keypoint = ValF(StringField(keypoints,k,d1))
				\key()\keytime = StringField(keytimes,k,d1)
			Next k
		EndIf
		
		If pathdata
			AddToSVGPath(pathdata,@\path)
			;ReportPath(@\path,"BUM",pathdata)
		EndIf
		
	EndWith
	
	ProcedureReturn a
	
EndProcedure

Procedure.s SVGInternal_AnimatorToString(s,e,a)
	
	xm.s = ""
	
	With svgmem(s)\element(e)\animator(a)
		
		If Not \type
			ProcedureReturn ""
		EndIf
		
		If \repeat_count = #SMIL_Repeat_Indefinite
			repcon.s = "indefinite"
		Else
			repcon = Str(\repeat_count)
		EndIf
		
		Select \fill
			Case #SMIL_Fill_Remove
				filcon.s = "remove"
			Case #SMIL_Fill_Freeze
				filcon = "freeze"
		EndSelect
		
		Select \rotate
			Case #SMIL_Rotate_Auto
				rotcon.s = "auto"
			Case #SMIL_Rotate_AutoReverse
				rotcon.s = "auto-reverse"
		EndSelect
		
		Select \calc_mode
			Case #SMIL_Path_Discrete
				calcon.s = "discrete"
			Case #SMIL_Path_Linear
				calcon = "linear"
			Case #SMIL_Path_Paced
				calcon = "paced"
			Case #SMIL_Path_Spline
				calcon = "spline"
		EndSelect
		
		Select \type
			Case #SVG_Animator_Attribute
				tagname.s = "animate"
				xm = "<"+tagname+" attributeName='"+\attr_name+"' begin='"+\begin+"' from='"+\from_val+"' to='"+\to_val+"' dur='"+\duration+"' type='' repeatCount='"+repcon+"' />"
			Case #SVG_Animator_Motion
				tagname.s = "animateMotion"
				;coordstring.s = "M 0 0 L 100 100";
				coordstring.s = SVGInternal_CalibratePath(s,@\path)
				
				keypointpart.s = ""
				keytimepart.s = ""
				If ListSize(\key())
					keypointpart = " keyPoints='"
					keytimepart = " keyTimes='"
					ForEach \key()
						keypointpart + StrF(\key()\keypoint)+";"
						keytimepart + \key()\keytime+";"
					Next
					EnsureThisNotEnd(keypointpart,";")
					EnsureThisNotEnd(keytimepart,";")
					keypointpart + "'"
					keytimepart + "'"
				EndIf
				xm = "<"+tagname+" path='"+coordstring+"' begin='"+\begin+"'dur='"+\duration+"' repeatCount='"+repcon+"' rotate='"+rotcon+"'"+keypointpart+keytimepart+" calcMode='"+calcon+"' />"
			Case #SVG_Animator_Transform
				tagname.s = "animateTransform"
				xm = "<"+tagname+" attributeName='transform' begin='"+\begin+"' dur='"+\duration+"' type='' from='"+\from_val+"' to='"+\to_val+"' repeatCount='"+repcon+"' />"
		EndSelect
		
		
	EndWith
	
	ProcedureReturn xm
	
EndProcedure

Procedure.b SVG_ClearAnimator(a.i=0)
	If Not a
		a = svgdata\a
	EndIf
	s = svgdata\s
	e = svgdata\e
	If s And e And a
		svgmem(s)\element(e)\animator(a)\in_use = #False
	EndIf
EndProcedure

Procedure.b SVG_ClearAllAnimators()
	AnimatorCheck
	For a = 1 To svgmem(s)\element(e)\animators
		svgmem(s)\element(e)\animator(a)\in_use = #False
	Next a
	svgmem(s)\element(e)\animators = 0
EndProcedure

Global NewList svg_animator_clipboard.SVG_Animator_Structure()

Procedure.b SVG_CopyAnimator(a.i=0)
	ClearList(svg_animator_clipboard())
	If Not a
		a = svgdata\a
	EndIf
	s = svgdata\s
	e = svgdata\e
	If s And e And a
		AddElement(svg_animator_clipboard())
		CopyStructure(@svgmem(s)\element(e)\animator(a),@svg_animator_clipboard(),SVG_Animator_Structure)
	EndIf
EndProcedure

Procedure.b SVG_CopyAllAnimators()
	ClearList(svg_animator_clipboard())
	s = svgdata\s
	e = svgdata\e
	If s And e
		For a = 1 To svgmem(s)\element(e)\animators
			AddElement(svg_animator_clipboard())
			CopyStructure(@svgmem(s)\element(e)\animator(a),@svg_animator_clipboard(),SVG_Animator_Structure)
		Next a
	EndIf
EndProcedure

Procedure.b SVG_PasteAnimators()
	s = svgdata\s
	e = svgdata\e
	If s And e
		svgmem(s)\element(e)\animators + ListSize(svg_animator_clipboard())
		If ArraySize(svgmem(s)\element(e)\animator())<svgmem(s)\element(e)\animators
			ReDim svgmem(s)\element(e)\animator(svgmem(s)\element(e)\animators)
		EndIf
		ForEach svg_animator_clipboard()
			CopyStructure(@svg_animator_clipboard(),@svgmem(s)\element(e)\animator(a),SVG_Animator_Structure)
		Next
	EndIf
EndProcedure











;-
;- NAMESPACES

Procedure.b SVG_AddNameSpace(p1.s,p2.s)
	s = svgdata\s
	If s>0
		;AddMapElement(svgmem(s)\namespace(),p1)
		If p2
			svgmem(s)\namespace(p1) = p2
		Else
			DeleteMapElement(svgmem(s)\namespace(),p1)
		EndIf
	EndIf
EndProcedure

Procedure.b SVG_DefaultNameSpaces()
	s = svgdata\s
	If s>0
		ClearMap(svgmem(s)\namespace())
		SVG_AddNameSpace("xmlns","http://www.w3.org/2000/svg")
		SVG_AddNameSpace("xmlns:svg","http://www.w3.org/2000/svg")
		SVG_AddNameSpace("xmlns:xlink","http://www.w3.org/1999/xlink")
	EndIf
EndProcedure










;-

;-

Procedure.b WASTE_CalculateTotalSVGElementBounds(s.i)
	
	With svgmem(s)
		;R("SVG_CalculateBounds: "+Str(s)+c13+"ELEMENTS: "+Str(\elements))
		If \bounds_known : ProcedureReturn #True : EndIf
		
		\bound\x = 0
		\bound\y = 0
		If Not \elements
			\bound\w = 1
			\bound\h = 1
			\bounds_known = #True
			ProcedureReturn #True
		EndIf
		
		x.f = 9999999
		y.f = 9999999
		w.f = -9999999
		h.f = -9999999
		
		If \elements
			For e = 1 To \elements
				If Not \element(e)\in_use : Continue : EndIf
				KnowBounds(s,e)
				;If LCase(\element(e)\id)="polyrect" : ReportRectF(@\element(e)\bound,0,"#"+Str(e)+" "+\element(e)\id) : EndIf
				
				x = Defeat(x,\element(e)\bound\x-\element(e)\strokewidth/2)
				y = Defeat(y,\element(e)\bound\y-\element(e)\strokewidth/2)
				w = Beat(w,\element(e)\bound\x+\element(e)\bound\w+(\element(e)\strokewidth))
				h = Beat(h,\element(e)\bound\y+\element(e)\bound\h+(\element(e)\strokewidth))
			Next e
		EndIf
		
		;R("CalculateTotalSVGElementBounds"+c13+c13+StrF(x,0)+","+StrF(y,0)+c13+StrF(w,0)+","+StrF(h,0))
		;x = 0;Beat(0,x)
		;y = 0;Beat(0,y)
		\bound\w = w; -x
		\bound\h = h; -y
		\bounds_known = #True
	EndWith
	
EndProcedure


Procedure.f PercentageOrDecimalToDecimal(n.f)
	
	If n>1
		ProcedureReturn n/100
	EndIf
	ProcedureReturn n
	
EndProcedure

Procedure.f PercentageOrDecimalTo255(n.s)
	
	If Right(n,1)="%"
		n = RemoveString(n,"%")
		nf = ValF(n)
		ProcedureReturn nf/100*255
	EndIf
	
	nf = ValF(n)
	ProcedureReturn nf*255
	
EndProcedure

Macro StrDecimal(n)
	StrF(n,2)
EndMacro

Procedure.f PercentageOrDecimalToPercentage(t.s)
	
	t = Trim(t)
	If Right(t,1)="%"
		t = RemoveString(t,"%")
		ProcedureReturn ValF(t)
	EndIf
	tv.f = ValF(t)
	If tv>1
		ProcedureReturn tv
	EndIf
	ProcedureReturn tv*100
	
EndProcedure




Procedure.s SVGInternal_GradientToString(s.i,g.i)
	With svgmem(s)\gradient(g)
		Select \type
			Case #SVG_Gradient_Radial
				tag.s = "radialGradient"
				coordxml.s = "cx='"+SF(\x,svgmem(s)\output_zoom)+"' cy='"+SF(\y,svgmem(s)\output_zoom)+"' r='"+SF(\w,svgmem(s)\output_zoom)+"'"
			Case #SVG_Gradient_Linear
				tag.s = "linearGradient"
				coordxml.s = "x1='"+SF(\x,svgmem(s)\output_zoom)+"' y1='"+SF(\y,svgmem(s)\output_zoom)+"' x2='"+SF(\w,svgmem(s)\output_zoom)+"' y2='"+SF(\h,svgmem(s)\output_zoom)+"'"
			Default
				R("Can't deal with gradients of type "+Str(\type))
				ProcedureReturn ""
		EndSelect
		gxml.s = "<"+tag+" id='"+\id+"' "+coordxml+" gradientUnits='userSpaceOnUse' >"
		ForEach \Point()
			gxml + "<stop offset='"+StrDecimal(PercentageOrDecimalToDecimal(\Point()\position))+"' stop-color='"+RGB2Hex(\Point()\color)+"' stop-opacity='"+StrDecimal(PercentageOrDecimalToDecimal(\Point()\opacity))+"' />"
		Next
		gxml + "</"+tag+">"
		ProcedureReturn gxml
	EndWith
EndProcedure




Macro SVGInternal_StringifyRegion(struc,var)
	If struc\x = -200
		var + "x='0%' "
	Else
		var + "x='"+StrF(struc\x)+"' "
	EndIf
	If struc\y = -200
		var + "y='0%' "
	Else
		var + "y='"+StrF(struc\y)+"' "
	EndIf
	If struc\w = -100
		var + "width='100%' "
	Else
		var + "width='"+StrF(struc\w)+"' "
	EndIf
	If struc\h = -100
		var + "height='100%' "
	Else
		var + "height='"+StrF(struc\h)+"' "
	EndIf
EndMacro

Procedure.s SVGInternal_LightToString(*struc.SVG_LightStructure)
	With *struc
		pxml.s = "<fe"+SVGFilterPrimitive(\type)\tag+" "
		Select \type
			Case #SVG_FP_DistantLight
				pxml + "azimuth='"+StrF(\azimuth)+"' elevation='"+StrF(\elevation)+"' "
			Case #SVG_FP_PointLight
				pxml + "x='"+StrF(\source_x)+"' y='"+StrF(\source_y)+"' z='"+StrF(\source_z)+"' "
			Case #SVG_FP_SpotLight
				pxml + "x='"+StrF(\source_x)+"' y='"+StrF(\source_y)+"' z='"+StrF(\source_z)+"' "
				pxml + "pointsAtX='"+StrF(\target_x)+"' pointsAtY='"+StrF(\target_y)+"' pointsAtZ='"+StrF(\target_z)+"' "
				pxml + "specularExponent='"+StrF(\specular_exponent)+"' limitingConeAngle ='"+StrF(\limiting_cone_angle)+"' "
		EndSelect
		pxml + "/>"
		ProcedureReturn pxml
	EndWith
EndProcedure

Procedure.s SVGInternal_FilterToString(s.i,f.i)
	
	With svgmem(s)\filter(f)
		;R("STAGES: "+Str(\stages))
		If ListSize(\primitive())<1
			ProcedureReturn ""
		EndIf
		SVGInternal_StringifyRegion(svgmem(s)\filter(f)\region,region.s)
		fxml.s = "<filter id='"+\id+"' filterUnits='userSpaceOnUse' "+region+">"+c13
	EndWith
	t=0
	ForEach svgmem(s)\filter(f)\primitive()
		t+1
		;R("SVG "+Str(s)+" FILTER "+Str(f)+" PRIMITIVE "+Str(t))
		With svgmem(s)\filter(f)\primitive()
			If Not \type : Continue : EndIf
			pxml.s = "<fe"+SVGFilterPrimitive(\type)\tag+" "
			Select \type
				Case #SVG_FP_Blend
					pxml + "mode='"+\blend_mode+"' "
				Case #SVG_FP_ColorMatrix
					
				Case #SVG_FP_ComponentTransfer
					
				Case #SVG_FP_Composite
					pxml + "operator='"+\operator+"' "
					If \operator=#SVG_Composite_Arithmetic
						For k = 1 To 4
							pxml + "k"+Str(k)+"='"+StrF(\k(k))+"' "
						Next k
					EndIf
				Case #SVG_FP_ConvolveMatrix
					pxml + "order='"+\order+"' kernelMatrix='"+\kernel_matrix+"' divisor='"+StrF(\divisor)+"' bias='"+StrF(\bias)+"' targetX='"+Str(\target_x)+"' targetY='"+Str(\target_y)+"' edgeMode='"+\edge_mode+"' "
					
				Case #SVG_FP_DiffuseLighting
					pxml + "lighting-color='"+RGBAToCSS(\light_color)+"' surfaceScale='"+StrF(\surface_scale)+"' diffuseConstant='"+StrF(\diffuse_constant)+"' kernelUnitLength='"+\kernel_unit_length+"' "
					
				Case #SVG_FP_DisplacementMap
					pxml + "scale='"+StrF(\scale)+"' xChannelSelector='"+\xChannelSelector+"' yChannelSelector='"+\yChannelSelector+"' "
				Case #SVG_FP_DistantLight
					
				Case #SVG_FP_Flood
					pxml + "flood-color='"+RGBAToCSS(\flood_color)+"' flood-opacity='"+StrF(\flood_opacity)+"' "
				Case #SVG_FP_GaussianBlur
					pxml + "stdDeviation='"+StrF(\blur_amount)+"' "
				Case #SVG_FP_Image
					pxml + ""
				Case #SVG_FP_Merge
					
				Case #SVG_FP_Morphology
					pxml + "operator='"+\operator+"' radius='"+StrF(\radius_x)+" "+StrF(\radius_y)+"' "
				Case #SVG_FP_Offset
					pxml + "dx='"+StrF(\dx)+"' dy='"+StrF(\dy)+" "
				Case #SVG_FP_PointLight
					
				Case #SVG_FP_SpecularLighting
					pxml + "lighting-color='"+RGBAToCSS(\light_color)+"' surfaceScale='"+StrF(\surface_scale)+"' kernelUnitLength='"+\kernel_unit_length+"' specularConstant='"+StrF(\specular_constant)+"' specularExponent='"+StrF(\specular_exponent)+" "
				Case #SVG_FP_SpotLight
					
				Case #SVG_FP_Tile
					pxml + ""
				Case #SVG_FP_Turbulence
					pxml + "type='"+\turbulence_type+"' baseFrequency='"+StrF(\baseXFrequency)+" "+StrF(\baseYFrequency)+"' numOctaves='"+Str(\octaves)+"' seed='"+StrF(\seed)+"' stitchTiles='"+\stitch+"' "
				Default
					R("Can't deal with filter primitives of type "+Str(\type))
					Continue
			EndSelect
			EnsureThisEnd(pxml," ")
			
			region.s = "" ; necessary for resetting the variable!
			If SVGFilterPrimitive(\type)\uses_region
				SVGInternal_StringifyRegion(svgmem(s)\filter(f)\primitive()\region,region)
			EndIf
			
			inputt.s = ""
			Select SVGFilterPrimitive(\type)\inputs
				Case 0
				Case 1
					inputt+"in='"+\input1+"' "
				Case 2
					inputt+"in='"+\input1+"' in2='"+\input2+"' "
			EndSelect
			
			closer.s = "/>"
			Select \type
				Case #SVG_FP_DiffuseLighting, #SVG_FP_SpecularLighting
					closer = ">"
					ForEach \light()
						closer + SVGInternal_LightToString(@\light())
					Next
					closer + "</"+SVGFilterPrimitive(\type)\tag+">"
			EndSelect
			
			fxml + pxml+region+inputt+closer+c13
		EndWith
	Next
	fxml + "</filter>"+c13
	
	ProcedureReturn fxml
	
EndProcedure


Procedure.b SVGInternal_PolylineDataFromString(s.i,e.i,pointarr.s,add.b)
	
	With svgmem(s)\element(e)\path
		If add=#False
			\pathpoints = 0
		EndIf
		ps = \pathpoints
		
		pointarr = ReplaceString(Trim(pointarr),",",c32)
		EnsureThisEnd(pointarr,c32)
		spaces = CountString(pointarr,c32)
		For j = 1 To spaces
			xs.s = StringField(pointarr,1,c32)
			ys.s = StringField(pointarr,2,c32)
			If Nots xs Or Nots ys : Break : EndIf
			EnsureThisNotStart(pointarr,xs+c32+ys+c32)
			ps+1
			arrsize=ArraySize(\pathpoint(),1) : If arrsize<ps : ReDim \pathpoint(ps+2) : EndIf
			\pathpoint(ps)\mode = #SVG_Path_Straight
			\pathpoint(ps)\x = ValF(xs)
			\pathpoint(ps)\y = ValF(ys)
		Next j
		\pathpoints = ps
		;R("PATH POINTS: "+Str(\pathpoints))
	EndWith
	
EndProcedure

Procedure.b NUSVGInternal_PolylineDataFromString(*s.SVG_Structure,*e.SVG_Element_Structure,pointarr.s,add.b)
	
	With *e\path
		If add=#False
			\pathpoints = 0
		EndIf
		ps = \pathpoints
		
		pointarr = ReplaceString(Trim(pointarr),",",c32)
		EnsureThisEnd(pointarr,c32)
		spaces = CountString(pointarr,c32)
		For j = 1 To spaces
			xs.s = StringField(pointarr,1,c32)
			ys.s = StringField(pointarr,2,c32)
			If Nots xs Or Nots ys : Break : EndIf
			EnsureThisNotStart(pointarr,xs+c32+ys+c32)
			ps+1
			arrsize=ArraySize(\pathpoint(),1) : If arrsize<ps : ReDim \pathpoint(ps+2) : EndIf
			\pathpoint(ps)\mode = #SVG_Path_Straight
			\pathpoint(ps)\x = ValF(xs)
			\pathpoint(ps)\y = ValF(ys)
		Next j
		\pathpoints = ps
		;R("PATH POINTS: "+Str(\pathpoints))
	EndWith
	
EndProcedure



Procedure.s CursorName(c.i)
	
	n.s
	Select c
		Case #SVG_Cursor_CrossHair
			n = "crosshair"
		Case #SVG_Cursor_Pointer
			n = "pointer"
		Case #SVG_Cursor_Move
			n = "move"
		Case #SVG_Cursor_Help
			n = "help"
		Case #SVG_Cursor_Text
			n = "text"
		Case #SVG_Cursor_Wait
			n = "wait"
		Case #SVG_Cursor_EResize
			n = "e-resize"
		Case #SVG_Cursor_NEResize
			n = "ne-resize"
		Case #SVG_Cursor_NWResize
			n = "nw-resize"
		Case #SVG_Cursor_NResize
			n = "n-resize"
		Case #SVG_Cursor_SEResize
			n = "se-resize"
		Case #SVG_Cursor_SWResize
			n = "sw-resize"
		Case #SVG_Cursor_SResize
			n = "s-resize"
		Case #SVG_Cursor_WResize
			n = "w-resize"
	EndSelect
	
	ProcedureReturn n
	
EndProcedure

Procedure.b IdentifyCursor(n.s)
	Select LCase(n)
		Case "crosshair"
			ProcedureReturn #SVG_Cursor_CrossHair
		Case "pointer"
			ProcedureReturn #SVG_Cursor_Pointer
		Case "move"
			ProcedureReturn #SVG_Cursor_Move
		Case "help"
			ProcedureReturn #SVG_Cursor_Help
		Case "text"
			ProcedureReturn #SVG_Cursor_Text
		Case "wait"
			ProcedureReturn #SVG_Cursor_Wait
		Case "e-resize"
			ProcedureReturn #SVG_Cursor_EResize
		Case "ne-resize"
			ProcedureReturn #SVG_Cursor_NEResize
		Case "nw-resize"
			ProcedureReturn #SVG_Cursor_NWResize
		Case "n-resize"
			ProcedureReturn #SVG_Cursor_NResize
		Case "se-resize"
			ProcedureReturn #SVG_Cursor_SEResize
		Case "sw-resize"
			ProcedureReturn #SVG_Cursor_SWResize
		Case "s-resize"
			ProcedureReturn #SVG_Cursor_SResize
		Case "w-resize"
			ProcedureReturn #SVG_Cursor_WResize
	EndSelect
	
	ProcedureReturn #SVG_Cursor_Default
	
EndProcedure

Procedure.b IdentifyStrokeLinecap(n.s)
	
	Select LCase(n)
		Case "butt"
			ProcedureReturn #SVG_Stroke_Butt
		Case "round"
			ProcedureReturn #SVG_Stroke_Round
		Case "square"
			ProcedureReturn #SVG_Stroke_Square
		Case "inherit"
			ProcedureReturn #SVG_Stroke_Inherit
	EndSelect
	
EndProcedure

Procedure.b IdentifyStrokeLinejoin(n.s)
	
	Select LCase(n)
		Case "miter"
			ProcedureReturn #SVG_Stroke_Miter
		Case "round"
			ProcedureReturn #SVG_Stroke_Round
		Case "bevel"
			ProcedureReturn #SVG_Stroke_Bevel
		Case "inherit"
			ProcedureReturn #SVG_Stroke_Inherit
	EndSelect
	
EndProcedure

Procedure.s StrokeLineJoinName(lj.b)
	
	Select lj
		Case #SVG_Stroke_Miter
			ProcedureReturn "miter"
		Case #SVG_Stroke_Round
			ProcedureReturn "round"
		Case #SVG_Stroke_Bevel
			ProcedureReturn "bevel"
		Case #SVG_Stroke_Inherit
			ProcedureReturn "inherit"
	EndSelect
	
EndProcedure

Procedure.s StrokeLineCapName(lc.b)
	
	Select lc
		Case #SVG_Stroke_Butt
			ProcedureReturn "butt"
		Case #SVG_Stroke_Round
			ProcedureReturn "round"
		Case #SVG_Stroke_Square
			ProcedureReturn "square"
		Case #SVG_Stroke_Inherit
			ProcedureReturn "inherit"
	EndSelect
	
EndProcedure

Procedure.s FormAttribute(name.s,value.s,css.b)
	
	If css
		ProcedureReturn name+":"+value+"; "
	Else
		ProcedureReturn name+"='"+value+"' "
	EndIf
	
EndProcedure

Procedure.s SVGInternal_StyleToString(s.i,e.i,css.b=#False)
	
	With svgmem(s)\element(e)
		
		stylingpart.s = ""
		
		If \styling
			stylingpart + EnsureEnd(\styling,";")
		EndIf
		
		cursorpart.s = ""
		Select \cursor
				;Case #SVG_Cursor_Default
				;    cursorpart = ""
			Case #SVG_Cursor_URI
				cursorpart = "url("+c39+\cursor_url+c39+")"
			Default
				cursorpart = CursorName(\cursor)
		EndSelect
		If cursorpart
			cursorpart + FormAttribute("cursor",cursorpart,css)
		EndIf
		
		fillpart.s = ""
		primaryfillpart.s = ""
		Select \fillmode
			Case #SVG_FillMode_None
				primaryfillpart = "none"
			Case #SVG_FillMode_Color
				primaryfillpart = RGB2Hex(\fillvalue)
			Case #SVG_FillMode_Gradient
				If \fillvalue>0
					g = \fillvalue
					svgmem(s)\gradient(g)\employed = #True
					primaryfillpart = "url(#"+svgmem(s)\gradient(g)\id+")"
				EndIf
			Case #SVG_FillMode_Pattern
				If \fillvalue>0
					p = \fillvalue
					svgmem(s)\pattern(p)\employed = #True
					primaryfillpart = "url(#"+svgmem(s)\pattern(p)\id+")"
				EndIf
				
		EndSelect
		Select \type
			Case #SVG_Element_Polyline, #SVG_Element_Line
				fillpart + FormAttribute("fill","none",css)
			Default
				If primaryfillpart
					fillpart + FormAttribute("fill",primaryfillpart,css)
				EndIf
				If \opacity<>255
					fillpart + FormAttribute("fill-opacity",StrF(\opacity/255),css)
				EndIf
		EndSelect
		If Not \visibility
			visibilitypart.s = FormAttribute("visibility","hidden",css)
		EndIf
		
		
		strokepart.s
		Select \type
			Case #SVG_Element_Line, #SVG_Element_Polyline
				strokepart + FormAttribute("stroke",primaryfillpart,css)
				If \strokewidth
					strokepart + FormAttribute("stroke-width",StrF(\strokewidth),css)
				EndIf
				If \strokelinecap
					strokepart + FormAttribute("stroke-linecap",StrokeLineCapName(\strokelinecap),css)
				EndIf
				If \strokelinejoin
					strokepart + FormAttribute("stroke-linejoin",StrokeLineJoinName(\strokelinejoin),css)
				EndIf
				If \opacity<>255
					strokepart + FormAttribute("stroke-opacity",StrF(\opacity/255),css)
				EndIf
			Default
				If \strokecolor Or \strokewidth Or \strokelinejoin Or \strokelinecap
					If \strokecolor>-1
						strokepart + FormAttribute("stroke",RGB2Hex(\strokecolor),css)
					EndIf
					If \strokewidth
						strokepart + FormAttribute("stroke-width",StrF(\strokewidth),css)
					EndIf
					If \strokeopacity>0 And \strokeopacity<255
						strokepart + FormAttribute("stroke-opacity",StrF(\strokeopacity/255),css)
					EndIf
					If \strokelinejoin
						strokepart + FormAttribute("stroke-linejoin",StrokeLineJoinName(\strokelinejoin),css)
					EndIf
					If \strokelinecap
						strokepart + FormAttribute("stroke-linecap",StrokeLineCapName(\strokelinecap),css)
					EndIf
				EndIf
		EndSelect
		
		
		
		; fonts...
		fontpart.s
		If \content
			If SVGElementTypeIsFontable(\type)
				fontpart + FormAttribute("font-family",\fontname,css)
				fontpart + FormAttribute("font-size",Str(\fontsize),css)
			EndIf
		EndIf
		
		
		total.s
		If css
			total = "style="+c34+stylingpart+strokepart+fillpart+cursorpart+visibilitypart+fontpart+c34
		Else
			If stylingpart
				total = "style="+c34+stylingpart+c34+c32
			EndIf
			total + strokepart+fillpart+cursorpart+visibilitypart+fontpart
		EndIf
		
		ProcedureReturn total
		
	EndWith
EndProcedure

Procedure.s SVGInternal_ElementToString(s.i,e.i,calibrate.b)
	
	Protected.f ex,ey,ew,eh
	
	If calibrate
		CalibrateSVGForOutputting(s)
	EndIf
	
	With svgmem(s)\element(e)
		
		Select \type
			Case #SVG_Element_CustomCode
				ProcedureReturn \content
			Case #SVG_Element_Comment
				ProcedureReturn "<!-- "+\content+" -->"
			Case #SVG_Element_GroupCloser
				ProcedureReturn "</g>"
		EndSelect
		
		If \id
			idpart.s = "id='"+\id+"' "
		Else
			idpart = ""
		EndIf
		If \class
			classpart.s = "class='"+\class+"' "
		Else
			classpart = ""
		EndIf
		
		filterpart.s
		If \filter>0
			svgmem(s)\filter(\filter)\employed = #True
			filterpart = "filter='url(#"+svgmem(s)\filter(\filter)\id+")' "
		EndIf
		
		
		Select \type
			Case #SVG_Element_Path, #SVG_Element_TextPath
				If Not \path\pathpoints
					; path is blank -> nothing to draw
					ProcedureReturn ""
				EndIf
				;KnowBounds(s,e)
				;ex = \bound\x
				;ey = \bound\y
				coordstring.s = SVGInternal_CalibratePath(s,@svgmem(s)\element(e)\path)
			Case #SVG_Element_Polygon, #SVG_Element_Polyline
				KnowBounds(s,e)
				coordstring.s = CalibrateSVGPoly(s,e)
			Case #SVG_Element_Circle
				;ex = (\cx-selviz\x)*svgmem(s)\output_zoom
				;ey = (\cy-selviz\y)*svgmem(s)\output_zoom
				ex = (\cx*svgmem(s)\output_zoom)-svgmem(s)\selviz\x
				ey = (\cy*svgmem(s)\output_zoom)-svgmem(s)\selviz\y
				ew = \rx*svgmem(s)\output_zoom
				eh = \rx*svgmem(s)\output_zoom
			Case #SVG_Element_Ellipse
				ex = (\cx*svgmem(s)\output_zoom)-svgmem(s)\selviz\x
				ey = (\cy*svgmem(s)\output_zoom)-svgmem(s)\selviz\y
				ew = \rx*svgmem(s)\output_zoom
				eh = \ry*svgmem(s)\output_zoom
			Case #SVG_Element_Line
				coordstring.s = CalibrateSVGLine(s,e)
			Default
				ex = (\x*svgmem(s)\output_zoom)-svgmem(s)\selviz\x
				ey = (\y*svgmem(s)\output_zoom)-svgmem(s)\selviz\y
				ew = \w*svgmem(s)\output_zoom
				eh = \h*svgmem(s)\output_zoom
		EndSelect
		
		;R(StrF(\x)+c13+StrF(\y)+c13+StrF(\w)+c13+StrF(\h))
		;R(StrF(ex)+c13+StrF(ey)+c13+StrF(ew)+c13+StrF(eh))
		
		
		stylingpart.s = SVGInternal_StyleToString(s,e,#False)
		
		
		
		oldover.s = \action(#SVG_Action_OnMouseOver)
		oldout.s = \action(#SVG_Action_OnMouseOut)
		tooltiph.s
		If \tooltip And Not forpattern
			ttid.s = Str(e)
			svgmem(s)\tooltiph + "<rect id='inbuilt_tooltipbg_"+ttid+"' x='0' y='0' height='16' fill='#ffffe1' style='stroke:black; stroke-width:1; visibility:hidden; ' />"
			svgmem(s)\tooltiph + "<text id='inbuilt_tooltiptx_"+ttid+"' x='0' y='12' font-family='Arial' font-size='12' fill='black' style='visibility:hidden; '>"+ReplaceString(ReplaceString(\tooltip,c39,"&apos;"),c34,"&quot;")+"</text>"
			
			\action(#SVG_Action_OnMouseOver) + "ShowTooltip(event,"+ttid+"); "
			\action(#SVG_Action_OnMouseOut) + "ClearTooltip("+ttid+"); "
			tooltipsused = #True
		EndIf
		
		actionpart.s = ""
		ForEach \action()
			actionpart+MapKey(\action())+"="+c34+\action()+c34+c32
		Next
		
		\action(#SVG_Action_OnMouseOver) = oldover
		\action(#SVG_Action_OnMouseOut) = oldout
		
		
		
		maskpart.s
		If \mask
			maskpart = "mask='url(#"+svgmem(s)\element(\mask)\id+")' "
		EndIf
		
		attrarr.s = StringMapToXMLAttributes(\attribute())
		
		extras.s = idpart+classpart+actionpart+stylingpart+maskpart+filterpart
		tagname.s = SVGElementTypeTag(\type)
		thisexml.s = ""
		xml_defs.s = ""
		Select \type
			Case #SVG_Element_Box
				thisexml + "<"+tagname+" "+extras+"x='"+SF(ex)+"' y='"+SF(ey)+"' width='"+SF(ew)+"' height='"+SF(eh)+"' "+attrarr+">"
			Case #SVG_Element_Circle
				thisexml + "<"+tagname+" "+extras+"cx='"+SF(ex)+"' cy='"+SF(ey)+"' r='"+SF(ew)+"' "+attrarr+">"
			Case #SVG_Element_Ellipse
				thisexml + "<"+tagname+" "+extras+"cx='"+SF(ex)+"' cy='"+SF(ey)+"' rx='"+SF(ew)+"' ry='"+SF(eh)+"' "+attrarr+">"
			Case #SVG_Element_Text
				thisexml + "<"+tagname+" "+extras+"x='"+SF(ex)+"' y='"+SF(ey)+"' "+attrarr+">"+\content
			Case #SVG_Element_WrappedText
				innerhtml.s = RemoveString(\content,c10)
				innerhtml = ReplaceString(innerhtml,c13,"<BR>")
				innerhtml = ReplaceString(innerhtml,"<BR>","<div></div>")
				thisexml + "<foreignObject x='"+SF(ex)+"' y='"+SF(ey)+"' width='"+SF(ew)+"' height='"+SF(eh)+"'><body xmlns='http://www.w3.org/1999/xhtml' "+stylingpart+" style='font-family:'"+\fontname+"'; font-size:"+Str(\fontsize)+"pt; color:"+RGB2Hex(\fillvalue)+"; '>"+innerhtml+"</body></foreignObject>"
			Case #SVG_Element_ForeignObject
				thisexml + "<foreignObject x='"+SF(ex)+"' y='"+SF(ey)+"' width='"+SF(ew)+"' height='"+SF(eh)+"'>"+\content+"</foreignObject>"
			Case #SVG_Element_Line
				thisexml + "<"+tagname+" "+extras+coordstring+attrarr+">"
			Case #SVG_Element_Polyline
				thisexml + "<"+tagname+" "+extras+"points='"+coordstring+"' "+attrarr+">"
			Case #SVG_Element_Polygon
				thisexml + "<"+tagname+" "+extras+"points='"+coordstring+"' "+attrarr+">"
			Case #SVG_Element_Path
				thisexml + "<"+tagname+" "+extras+"d='"+coordstring+"' "+attrarr+">"
			Case #SVG_Element_TextPath
				textpathid.s = RandomID(s)
				xml_defs + "<path id='"+textpathid+"' d='"+coordstring+"'>"
				thisexml + "<text><textPath "+extras+" xlink:href='#"+textpathid+"'>"+\content+"</textPath>"
			Case #SVG_Element_Image, #SVG_Element_PBImage
				content.s = \content
				If \type=#SVG_Element_PBImage
					If \store_internally
						Select \imagecodec
							Case #PB_ImagePlugin_JPEG
								mime.s = "image/jpeg"
							Case #PB_ImagePlugin_PNG
								mime.s = "image/png"
						EndSelect
						slavefilename.s = "C:\PureSVGSlaveFile.txt"
						SaveImage(\imagenumber,slavefilename,\imagecodec)
						
						f = OpenFile(#PB_Any,slavefilename)
						loaf = Lof(f)
						*mem = AllocateMemory(loaf)
						ReadData(f,*mem,loaf)
						CloseFile(f)
						DeleteFile(slavefilename)
						
						bloaf = loaf*1.5
						*b64 = AllocateMemory(bloaf)
						Base64Encoder(*mem,loaf,*b64,bloaf)
						
						content = "data:"+mime+";base64,"+PeekS(*b64)
						FreeMemory(*mem)
						FreeMemory(*b64)
					EndIf
				EndIf
				thisexml + "<"+tagname+" "+extras+"x='"+SF(ex)+"' y='"+SF(ey)+"' width='"+SF(ew)+"' height='"+SF(eh)+"' "+attrarr+" xlink:href='"+content+"'>"
			Case #SVG_Element_Base64Image
				Select \imagecodec
					Case #PB_ImagePlugin_JPEG
						mime.s = "image/jpeg"
					Case #PB_ImagePlugin_PNG
						mime.s = "image/png"
				EndSelect
				content.s = "data:"+mime+";base64,"+\base64data
				thisexml + "<"+tagname+" "+extras+"x='"+SF(ex)+"' y='"+SF(ey)+"' width='"+SF(ew)+"' height='"+SF(eh)+"' "+attrarr+" xlink:href='"+content+"'>"
			Case #SVG_Element_Group
				;thisexml + "<g "+extras+"x="+c34+SF(ex)+c34+" y="+c34+SF(ey)+c34+" "+\attrs+" >"
				If \ismask
					thisexml = "<mask "+extras+" x='0' y='0' width='100%' height='100%'>"
				Else
					thisexml + "<g "+extras+" "+attrarr+" >"
				EndIf
				SVGInternal_ForEachElementInSVGGroup(s,e,a)
					thisexml + SVGInternal_ElementToString(s,a,#False)+c13
				Next a
				If \ismask
					thisexml + "</mask>"
				Else
					;thisexml + "</g>"
				EndIf
			Default
				R("DON'T KNOW HOW TO HANDLE SVG ELEMENTS OF TYPE "+Str(\type))
	EndSelect
	
	;R("ANIMATORS: "+Str(\animators))
	If \animators
		For a = 1 To \animators
			;R("ANIMATOR #"+Str(a)+" IN USE: "+Str(\animator(a)\in_use))
			If \animator(a)\in_use
				thisexml + SVGInternal_AnimatorToString(s,e,a)
			EndIf
		Next a
	EndIf
	thisexml+"</"+tagname+">"
	
	If \href
		thisexml = "<a xlink:href='"+ReplaceString(\href,"&","&amp;")+"' target='"+HyperlinkTargetName(\href_target)+"'>"+thisexml+"</a>"
	EndIf
	
	ProcedureReturn thisexml
	
EndWith
EndProcedure


;-
Procedure.s SVG_FormCode(s.i,standalone.b,useviewbox.b=#True,forpattern.b=#False)
	
	tooltipsused.b
	
	If Not IsSVG(s) : ProcedureReturn "" : EndIf
	
	
	CalibrateSVGForOutputting(s)
	;R("OZ AFTER CALIBRATION: "+StrF(svgmem(s)\output_zoom))
	
	xml_defs.s = ""
	xml_els.s = ""
	
	With svgmem(s)
		
		; reset def employments...
		For a = 1 To \gradients
			\gradient(a)\employed = #False
		Next a
		For a = 1 To \patterns
			\pattern(a)\employed = #False
		Next a
		For a = 1 To \filters
			\filter(a)\employed = #False
		Next a
		
		;ReportRectF(@selviz)
		
		For e = 1 To \elements
			If Not \element(e)\in_use : Continue : EndIf
			If \element(e)\ismask
				xml_defs + SVGInternal_ElementToString(s,e,#False)+c13
			Else
				xml_els + SVGInternal_ElementToString(s,e,#False)+c13
				;R(xml_els)
			EndIf
			If \element(e)\type = #SVG_Element_Group
				gc = CloserForSVGGroup(s,e)
				;R("GC: "+Str(gc))
				e = gc;+1
			EndIf
		Next e
		
		#SVG_BG_ID = "puresvg_background"
		If \bg_mode
			Select \bg_mode
				Case #SVG_FillMode_Color
					bg.s = "<rect id='"+#SVG_BG_ID+"' x='0' y='0' width='100%' height='100%' fill='"+RGB2Hex(\bg_value)+"' />"
				Case #SVG_FillMode_Gradient
					g = \bg_value
					\gradient(g)\employed = #True
					bg.s = "<rect id='"+#SVG_BG_ID+"' x='0' y='0' width='100%' height='100%' fill='url(#"+\gradient(g)\id+")' />"
				Case #SVG_FillMode_Pattern
					p = \bg_value
					\pattern(p)\employed = #True
					bg.s = "<rect id='"+#SVG_BG_ID+"' x='0' y='0' width='100%' height='100%' fill='url(#"+\pattern(p)\id+")' />"
			EndSelect
			If bg
				xml_els = bg+c13+c13+xml_els
			EndIf
		EndIf
		If xml_els
			xml_els+c13+c13
		EndIf
		
		
		
		If forpattern
			ProcedureReturn xml_els
		EndIf
		
		
		resc.s
		ForEach \external_css_file()
			resc + "<?xml-stylesheet href='"+EnsureEnd(\external_css_file(),".css")+"' type='text/css'?>"
		Next
		ForEach \external_js_file()
			resc+"<script type='text/javascript' xlink:href='"+EnsureEnd(\external_js_file(),".js")+"'></script>"+c13
		Next
		js.s = \javascript
		If tooltipsused
			js + tooltip_js
		EndIf
		If js
			resc+"<script type='text/javascript'>"+c13+js+c13+"</script>"+c13
		EndIf
		If resc : resc+c13 : EndIf
		
		
		
		If \filters
			For f = 1 To \filters
				If Not \filter(f)\employed Or Not \filter(f)\in_use : Continue : EndIf
				xml_defs + SVGInternal_FilterToString(s,f)+c13
			Next f
		EndIf
		If \gradients
			For g = 1 To \gradients
				If Not \gradient(g)\employed Or Not \gradient(g)\in_use Or Not ListSize(\gradient(g)\Point()) : Continue : EndIf
				xml_defs + SVGInternal_GradientToString(s,g)+c13;@svgmem(s)\gradient(g))+c13
			Next g
		EndIf
		If \patterns
			For a = 1 To \patterns
				If Not \pattern(a)\employed Or Not \pattern(a)\in_use Or Not \pattern(a)\mode : Continue : EndIf
				pxml.s = ""
				Select \pattern(a)\mode
					Case #SVG_Pattern_DirectCode
						pxml = \pattern(a)\code
					Case #SVG_Pattern_SVG
						If \pattern(a)\svg <> s
							If PatternRecursivityOK(@\pattern(a),s)
								pxml = SVG_FormCode(\pattern(a)\svg,#True,#False)
							EndIf
						EndIf
					Default
						R("Can't deal with patterns of mode "+Str(\pattern(a)\mode))
						Continue
				EndSelect
				If Nots pxml : Continue : EndIf
				xml_defs+c13
				xml_defs + "<pattern id='"+\id+"' x='0' y='0' width='"+Str(\w)+"' height='"+Str(\h)+"' patternUnits='userSpaceOnUse' >"
				xml_defs + c13+pxml+c13
				xml_defs + "</pattern>"
				xml_defs +c13
			Next a
		EndIf
		
		
		;R("OZ AFTER FILGRADPAT: "+StrF(svgmem(s)\output_zoom))
		
		titlestuff.s
		If \md[#SVG_Metadata_Title]
			titlestuff+"<title>"+\md[#SVG_Metadata_Title]+"</title>"
		EndIf
		If \md[#SVG_Metadata_Description]
			titlestuff+"<desc>"+\md[#SVG_Metadata_Description]+"</desc>"
		EndIf
		If titlestuff
			titlestuff+c13
		EndIf
		
		
		metadata.s
		For a = 1 To #TotalSVGMetadataFields
			If Nots \md[a] : Continue : EndIf
			Select a
				Case #SVG_Metadata_Title
					metadata + "<dc:title>"+\md[a]+"</dc:title>"
				Case #SVG_Metadata_Description
					metadata + "<dc:description>"+\md[a]+"</dc:description>"
				Case #SVG_Metadata_Coverage
					metadata + "<dc:coverage>"+\md[a]+"</dc:coverage>"
				Case #SVG_Metadata_Language
					metadata + "<dc:language>"+\md[a]+"</dc:language>"
				Case #SVG_Metadata_Source
					metadata + "<dc:source>"+\md[a]+"</dc:source>"
				Case #SVG_Metadata_Date
					metadata + "<dc:date>"+SVGToCoolDate(Val(\md[a]))+"</dc:date>"
				Case #SVG_Metadata_Publisher
					metadata + "<dc:publisher><cc:Agent><dc:title>"+\md[a]+"</dc:title></cc:Agent></dc:publisher>"
				Case #SVG_Metadata_Creator
					metadata + "<dc:creator><cc:Agent><dc:title>"+\md[a]+"</dc:title></cc:Agent></dc:creator>"
				Case #SVG_Metadata_Rightsholder
					metadata + "<dc:rights><cc:Agent><dc:title>"+\md[a]+"</dc:title></cc:Agent></dc:rights>"
				Case #SVG_Metadata_Contributors
					metadata + "<dc:contributor><cc:Agent><dc:title>"+\md[a]+"</dc:title></cc:Agent></dc:contributor>"
				Case #SVG_Metadata_Keywords
					items = CountString(\md[a],"|")
					kw.s = "<dc:subject><rdf:Bag>"
					For b = 1 To items
						kw + "<rdf:li>"+StringField(\md[a],b,"|")+"</rdf:li>"
					Next b
					kw + "</rdf:Bag></dc:subject>"
					metadata + kw
			EndSelect
			metadata+c13
		Next a
		If metadata
			metadata = "<metadata id='dublin_core_metadata'><rdf:RDF><cc:Work rdf:about=''>" +c13+ "<dc:format>image/svg+xml</dc:format>" +c13+ "<dc:type rdf:resource='http://purl.org/dc/dcmitype/StillImage' />" +c13+ metadata + "</cc:Work></rdf:RDF></metadata>"
			metadata+c13
		EndIf
		;metadata=""
		
		
		If \classinfo
			xml_defs + "<style type='text/css'>"+c13+\classinfo+"</style>"+c13
		EndIf
		
	
	
	;  R("ZOOM: "+StrF(svgmem(s)\output_zoom)+c13+c13+"OUTPUTW: "+StrF(outputw)+c13+"OUTPUTH: "+StrF(outputh))
	
	
	If standalone
		If metadata
			SVG_AddNameSpace("xmlns:dc","http://purl.org/dc/elements/1.1/")
			SVG_AddNameSpace("xmlns:cc","http://creativecommons.org/ns#")
			SVG_AddNameSpace("xmlns:rdf","http://www.w3.org/1999/02/22-rdf-syntax-ns#")
		EndIf
	EndIf
	namespacearr.s = ""
	ForEach \namespace()
		namespacearr + MapKey(\namespace())+"='"+\namespace()+"' "
	Next
	
	attrarr.s = StringMapToXMLAttributes(\attribute())
	
	If \tooltiph
		\tooltiph = ReplaceString(\tooltiph,"^",c34)
	EndIf
	
	genbit.s = "<!-- Created with PureSVG (http://www.purebasic.fr/english/viewtopic.php?f=12&t=40540) -->"
	
	
	mainidpart.s=""
	If \id
		mainidpart = " id='"+\id+"'"
	EndIf
	
	If xml_defs
		xml_defs = "<defs>"+c13+xml_defs+c13+"</defs>"+c13+c13
	EndIf
	
	xml.s = ""
	If standalone
		xml = "<?xml version='1.0'?>"+c13+#SVG_DocType+c13
	EndIf
		;xml + genbit+c13+"<svg "+namespacearr+"xmlns:svg="+c34+"http://www.w3.org/2000/svg"+c34+" xmlns="+c34+"http://www.w3.org/2000/svg"+c34+" version="+c34+"1.0"+c34+mainidpart+" width="+c34+SF(\output_w)+c34+" height="+c34+SF(\output_h)+c34+">"+c13+titlestuff+c13+metadata+c13+resc+\xml_defs+\xml_els+\tooltiph+"</svg>"
		If useviewbox
			xml + genbit+c13+"<svg "+namespacearr+attrarr+"version='1.0' "+mainidpart+" viewBox='0 0 "+SF(\output_w)+" "+SF(\output_h)+"' preserveAspectRatio='xMinYMin'>"+c13+titlestuff+c13+metadata+c13+resc+xml_defs+xml_els+\tooltiph+"</svg>"
		Else
			xml + genbit+c13+"<svg "+namespacearr+attrarr+"version='1.0' "+mainidpart+" width='"+SF(\output_w)+"' height='"+SF(\output_h)+"'>"+c13+titlestuff+c13+metadata+c13+resc+xml_defs+xml_els+\tooltiph+"</svg>"
		EndIf
	EndWith
	
	;xml = RemoveString(xml,LSet(".",#SVG_FloatAccuracy+1,"0")) ; this reduces filesize. i don't know if it causes problems - if it does, delete it
	
	ProcedureReturn xml
	
EndProcedure
;-






Procedure.b EnsureSVGPeripherals(s.i,path.s)
	
	EnsureThisEnd(path,"\")
	
	With svgmem(s)
		For e = 1 To \elements
			If \element(e)\type = #SVG_Element_PBImage
				If \element(e)\store_internally : Continue : EndIf
				If \element(e)\content : Continue : EndIf
				imgfilename.s = path+\element(e)\filename+"."
				Select \element(e)\imagecodec
					Case #PB_ImagePlugin_PNG
						imgfilename+"png"
					Case #PB_ImagePlugin_JPEG
						imgfilename+"jpg"
				EndSelect
				
				SaveImage(\element(e)\imagenumber,imgfilename,\element(e)\imagecodec)
				\element(e)\content = imgfilename
			EndIf
		Next e
	EndWith
	
EndProcedure

Procedure.b ReportPath(*p.SVG_Path_Structure,id.s="",prepend.s="")
	With *p
		tx.s = "ID: "+id+c13+c13
		If prepend
			tx + prepend+c13+c13
		EndIf
		For p = 1 To *p\pathpoints
			tx + UCase(PathMovementModeName(*p\pathpoint(p)\mode))+" TO "+Str(*p\pathpoint(p)\x)+","+Str(*p\pathpoint(p)\y)+c13
		Next p
		R(tx)
	EndWith
EndProcedure

Procedure.b ConvertElementToPath(s.i,e.i)
	
	With svgmem(s)\element(e)
		;R("CONVERTING "+\id+" TO PATH")
		pathdone.b = #False
		Select \type
			Case #SVG_Element_Path, #SVG_Element_TextPath, #SVG_Element_Polygon, #SVG_Element_Polyline
				ProcedureReturn #True
			Case #SVG_Element_Group
				ProcedureReturn #False
			Case #SVG_Element_Box
				\path\pathpoints = 5
				EnsureEnoughPathPoints(svgmem(s)\element(e)\path,5)
				;R("NEW PATH POINTS: "+Str(ArraySize(svgmem(s)\element(e)\pathpoint())))
				;R("s"+Str(s)+"  e"+Str(e)+"  id: "+\id+c13+Str(\x)+","+Str(\y)+","+Str(\w)+","+Str(\h))
				
				\path\pathpoint(1)\mode = #SVG_Path_SegmentStart
				\path\pathpoint(1)\x = \x
				\path\pathpoint(1)\y = \y
				\path\pathpoint(2)\mode = #SVG_Path_Straight
				\path\pathpoint(2)\x = \x+\w
				\path\pathpoint(2)\y = \y
				\path\pathpoint(3)\mode = #SVG_Path_Straight
				\path\pathpoint(3)\x = \x+\w
				\path\pathpoint(3)\y = \y+\h
				\path\pathpoint(4)\mode = #SVG_Path_Straight
				\path\pathpoint(4)\x = \x
				\path\pathpoint(4)\y = \y+\h
				
				\path\pathpoint(5)\mode = #SVG_Path_SegmentEnd
				
				;ReportPath(s,e)
				pathdone = #True
				
			Case #SVG_Element_Circle, #SVG_Element_Ellipse
				\path\pathpoints = 4
				EnsureEnoughPathPoints(svgmem(s)\element(e)\path,4)
				;topx.f = \cx
				topy.f = \cy-\ry
				;botx.f = \cx
				boty.f = \cy+\ry
				
				\path\pathpoint(1)\mode = #SVG_Path_SegmentStart
				\path\pathpoint(1)\x = \cx;topx
				\path\pathpoint(1)\y = topy
				
				\path\pathpoint(2)\mode = #SVG_Path_EllipticalArc
				\path\pathpoint(2)\x = \cx;botx
				\path\pathpoint(2)\y = boty
				\path\pathpoint(2)\rx = \rx
				\path\pathpoint(2)\ry = \ry
				\path\pathpoint(2)\large_arc = #False
				\path\pathpoint(2)\sweep = #True
				\path\pathpoint(2)\x_rotation = 0
				
				\path\pathpoint(3)\mode = #SVG_Path_EllipticalArc
				\path\pathpoint(3)\x = \cx;topx
				\path\pathpoint(3)\y = topy
				\path\pathpoint(3)\rx = \rx
				\path\pathpoint(3)\ry = \ry
				\path\pathpoint(3)\large_arc = #False
				\path\pathpoint(3)\sweep = #True
				\path\pathpoint(3)\x_rotation = 0
				
				\path\pathpoint(4)\mode = #SVG_Path_SegmentEnd
				pathdone = #True
				
			Case #SVG_Element_Line
				\path\pathpoints = 2
				EnsureEnoughPathPoints(svgmem(s)\element(e)\path,2)
				\path\pathpoint(1)\mode = #SVG_Path_SegmentStart
				\path\pathpoint(1)\x = \w
				\path\pathpoint(1)\y = \y
				
				\path\pathpoint(2)\mode = #SVG_Path_Straight
				\path\pathpoint(2)\x = \w
				\path\pathpoint(2)\y = \h
				
				;\pathpoint(3)\mode = #SVG_Path_SegmentEnd
				pathdone = #True
				
				
		EndSelect
		
		;//
		;//
		
		If pathdone
			\type = #SVG_Element_Path
			;ReportPath(s,e)
			ProcedureReturn #True
		EndIf
	EndWith
	
EndProcedure




Procedure.b SVGInternal_ShiftElement(s.i,e.i,xshift.f,yshift.f)
	
	If xshift=0 And yshift=0
		ProcedureReturn #True
	EndIf
	
	If Not s Or Not e
		R("SVG_ShiftElement fail: there is no current SVG or there is no current element.")
		ProcedureReturn #False
	EndIf
	
	With svgmem(s)\element(e)
		
		Select \type
			Case #SVG_Element_Group
				SVGInternal_ForEachElementInSVGGroup(s,e,a)
				SVGInternal_ShiftElement(s,a,xshift,yshift)
			Next a
		Case #SVG_Element_GroupCloser, #SVG_Element_CustomCode, #SVG_Element_Comment
			ProcedureReturn #True
		Case #SVG_Element_Circle, #SVG_Element_Ellipse
			\cx + xshift
			\cy + yshift
		Case #SVG_Element_Path, #SVG_Element_Polygon, #SVG_Element_Polyline, #SVG_Element_Line
			If \path\pathpoints
				For p = 1 To \path\pathpoints
					\path\pathpoint(p)\x + xshift
					\path\pathpoint(p)\y + yshift
					
					If \path\pathpoint(p)\intermediary_points
						For i = 1 To \path\pathpoint(p)\intermediary_points
							\path\pathpoint(p)\intermediary_point(i)\x + xshift
							\path\pathpoint(p)\intermediary_point(i)\y + yshift
						Next i
					EndIf
				Next p
			EndIf
		Default
			;R(\id+" OLD: "+StrF(\x,rnd)+","+StrF(\y,rnd))
			\x + xshift
			\y + yshift
			;R(\id+" NEW: "+StrF(\x,rnd)+","+StrF(\y,rnd))
	EndSelect
	
	If \bounds_known ; otherwise there's no point adjusting it
		\bound\x + xshift
		\bound\y + yshift
	EndIf
	\bounds_known = #False
EndWith

ProcedureReturn #True

EndProcedure

Procedure.b SVGInternal_ShiftGroupElement(s.i,e.i,xshift.f,yshift.f)
	
	If Not s Or Not e
		ProcedureReturn #False
	EndIf
	
	With svgmem(s)\element(e)
		
		If \type<>#SVG_Element_Group : ProcedureReturn #True : EndIf
		
		ga = e
		gb = CloserForSVGGroup(s,ga)
		If Not gb
			R("ShiftGroupElement. ERROR: this group has not been closed!")
			ProcedureReturn #False
		EndIf
		
		For e = ga+1 To gb-1
			Select svgmem(s)\element(e)\type
				Case #SVG_Element_GroupCloser
					Continue
				Case #SVG_Element_Group
					SVGInternal_ShiftGroupElement(s,e,xshift,yshift)
					Continue
			EndSelect
			
			SVGInternal_ShiftElement(s,e,xshift,yshift)
		Next e
		
		ProcedureReturn #True
		
	EndWith
	
EndProcedure







Procedure.b SVGInternal_RotateElement(s.i,e.i,deg.f,cx.f=#PB_Ignore,cy.f=#PB_Ignore)
	
	If Not s Or Not e
		ProcedureReturn #False
	EndIf
	
	If Not deg : ProcedureReturn #True : EndIf
	
	With svgmem(s)\element(e)
		
		Select \type
			Case #SVG_Element_Path, #SVG_Element_Polygon, #SVG_Element_Polyline, #SVG_Element_Line
				If \path\pathpoints<1 : ProcedureReturn #True : EndIf
			Case #SVG_Element_GroupCloser, #SVG_Element_Comment, #SVG_Element_CustomCode
				ProcedureReturn #True
			Case #SVG_Element_Group 
				ProcedureReturn SVGInternal_RotateGroupElement(s,e,deg,cx,cy)
		EndSelect
		
		If cx=#PB_Ignore Or cy=#PB_Ignore
			GetSVGElementCenter(s,e,@ec.PointF)
			;ReportRectF(@ec,0,"CENTER")
			If cx = #PB_Ignore
				cx = ec\x
			EndIf
			If cy = #PB_Ignore
				cy = ec\y
			EndIf
		EndIf
		
		rc.PointF : rc\x = cx : rc\y = cy
		
		Select \type
			Case #SVG_Element_Circle
				If rc\x=\cx And rc\y=\cy
					;R("CIRCLE ROTATION AROUND CENTRE: POINTLESS")
					ProcedureReturn #True
				EndIf
		EndSelect
		
		;R("type: "+Str(\type))
		;If FindString(\id,"red",1) : R(\id+" is a "+SVGElementTypeName(\type)) : EndIf
		Select \type
			Case #SVG_Element_Path, #SVG_Element_Polygon, #SVG_Element_Polyline, #SVG_Element_Line
				;R(\id+" is a path")
			Default
				;R(\id+" is NOT a path, but a "+SVGElementTypeName(\type))
				If Not ConvertElementToPath(s,e)
					R("COULD NOT CONVERT TO PATH")
					ProcedureReturn #False
				EndIf
				;ReportPath(s,e)
		EndSelect
		\bounds_known = #False
		;ProcedureReturn #True ;FUCK
		
		
		;R("PATH POINTS: "+Str(\pathpoints))
		If \path\pathpoints<1 : R("NO PATHPOINTS") : ProcedureReturn #False : EndIf
		For p = 1 To \path\pathpoints
			If \path\pathpoint(p)\mode = #SVG_Path_SegmentEnd : Continue : EndIf
			
			cur.PointF : cur\x = \path\pathpoint(p)\x : cur\y = \path\pathpoint(p)\y
			
			DegreeRotatePointAroundPoint(@cur,@rc,deg,@np.PointF)
			;Report2PointF(cur,np)
			\path\pathpoint(p)\x = np\x : \path\pathpoint(p)\y = np\y
			
			Select \path\pathpoint(p)\mode
				Case #SVG_Path_CubicBezier, #SVG_Path_SmoothCubicBezier, #SVG_Path_QuadraticBezier, #SVG_Path_SmoothQuadraticBezier
					; intermediary points
					If \path\pathpoint(p)\intermediary_points
						For i = 1 To \path\pathpoint(p)\intermediary_points
							DegreeRotatePointAroundPoint(@\path\pathpoint(p)\intermediary_point(i),@rc,deg,@np.PointF)
							\path\pathpoint(p)\intermediary_point(i)\x = np\x : \path\pathpoint(p)\intermediary_point(i)\y = np\y
						Next i
					EndIf
				Case #SVG_Path_EllipticalArc
					\path\pathpoint(p)\x_rotation + deg
			EndSelect
		Next p
		
		ProcedureReturn #True
		
	EndWith
	
EndProcedure

Procedure.b SVGInternal_RotateGroupElement(s.i,g.i,deg.f,cx.f=#PB_Ignore,cy.f=#PB_Ignore)
	
	If Not s Or Not g
		ProcedureReturn #False
	EndIf
	
	If Not deg : ProcedureReturn #True : EndIf
	
	With svgmem(s)
		
		If \element(g)\type<>#SVG_Element_Group : ProcedureReturn #True : EndIf
		
		GetSVGElementBoundingBox(s,g,@gb.RectF)
		;DebugRectF(@gb) : Debug ""
		If cx=#PB_Ignore Or cy=#PB_Ignore
			GetSVGElementCenter(s,g,@ec.PointF)
			If cx = #PB_Ignore
				cx = ec\x
			EndIf
			If cy = #PB_Ignore
				cy = ec\y
			EndIf
		EndIf
		rc.PointF
		rc\x = cx
		rc\y = cy
		
		\element(g)\bounds_known = #False
		
		;ReportPointF(rc)
		
		;SVGInternal_ForEachElementInSVGGroup(s,g,e)
		gc = CloserForSVGGroup(s,g)
		For e = g+1 To gc-1
			
			;R("ROTATING "+\element(e)\id+", A "+SVGElementTypeName(\element(e)\type))
			Select \element(e)\type
				Case #SVG_Element_Group
					;R("ROTATING A GROUP")
					SVGInternal_RotateGroupElement(s,e,deg,cx,cy)
				Case #SVG_Element_GroupCloser, #SVG_Element_CustomCode, #SVG_Element_Comment
					Continue
				;Case #SVG_Element_Path, #SVG_Element_TextPath
					;R("PATH")
				;	SVGInternal_RotateElement(s,e,deg,cx,cy)
				Default
					;ConvertElementToPath(s,e)
					;ReportPath(s,e)
					;SVGInternal_ShiftElement(s,e,Random(100),Random(100))
					SVGInternal_RotateElement(s,e,deg,cx,cy)
			EndSelect
		Next e
		
	EndWith
	
	ProcedureReturn #True

EndProcedure




Enumeration 0
	#SVG_Orient_Center
	#SVG_Orient_TopLeft
	#SVG_Orient_TopRight
	#SVG_Orient_BottomRight
	#SVG_Orient_BottomLeft
EndEnumeration

Declare.b SVGInternal_ScaleElement(s.i,e.i,wpc.f,hpc.f,maintain.b=#SVG_Orient_Center)
Declare.b SVGInternal_SetElementPosition(s.i,e.i,x.f,y.f)

Procedure.b CalculateShifts(*o.RectF,w.f,h.f,maintain.b,*wpc,*hpc,*xshift,*yshift)
	
	xpc.f : ypc.f
	xshift.f = 0
	yshift.f = 0
	
	If w = *o\w
		;R("WIDTH MATCHES ORIGINAL")
		xpc = 1
		xshift = 0
	Else
		xpc = 1 / *o\w * w
		;R("\bound\w = "+StrF(\bound\w,0)+c13+"o\w = "+StrF(o\w,0)+c13+"dest w = "+StrF(w,0)+c13+"xPC: "+StrF(xpc))
		Select maintain
			Case #SVG_Orient_Center
				xshift = (*o\w-w)/2
			Case #SVG_Orient_TopLeft+100, #SVG_Orient_BottomLeft
				newx.f = *o\x*xpc
				xshift = (*o\x-newx)
			Case #SVG_Orient_TopRight, #SVG_Orient_BottomRight
				xshift = *o\w-w
		EndSelect
	EndIf
	If h = *o\h
		;R("HEIGHT MATCHES ORIGINAL")
		ypc = 1
		yshift = 0
	Else
		ypc = 1 / *o\h * h
		;R("\bound\h = "+StrF(\bound\h,0)+c13+"o\h = "+StrF(o\h,0)+c13+"dest h = "+StrF(h,0)+c13+"yPC: "+StrF(ypc))
		Select maintain
			Case #SVG_Orient_Center
				yshift = (*o\h-h)/2
			Case #SVG_Orient_TopLeft+100, #SVG_Orient_TopRight
				newy.f = *o\y*ypc
				yshift = (*o\y-newy)
			Case #SVG_Orient_BottomRight, #SVG_Orient_BottomLeft
				yshift = *o\h-h
		EndSelect
	EndIf
	
	;Debug xpc
	;Debug ypc
	;Debug xshift
	;Debug yshift
	
	PokeF(*wpc,xpc)
	PokeF(*hpc,ypc)
	PokeF(*xshift,xshift)
	PokeF(*yshift,yshift)
	
EndProcedure

Procedure.b SVGInternal_SetElementDimensions(s.i,e.i,w.f,h.f,maintain.b=#SVG_Orient_Center)
	
	If w=#PB_Ignore And h=#PB_Ignore
		ProcedureReturn #True
	EndIf
	
	With svgmem(s)\element(e)
		KnowBounds(s,e)
		
		If w = #PB_Ignore
			w = \bound\w
		EndIf
		If h = #PB_Ignore
			h = \bound\h
		EndIf
		If w=\bound\w And h=\bound\h ; no resizing to be done
			ProcedureReturn #True
		EndIf
		
		
		CopyRectF(\bound,@o.RectF)
		
		
		CalculateShifts(@o,w,h,maintain,@xpc.f,@ypc.f,@xshift.f,@yshift.f)
		;R("TYPE: "+SVGElementTypeName(\type)+c13+"ID: "+\id+c13+"SCALE: "+StrF(xpc,1)+","+StrF(ypc,1)+c13+"SHIFT: "+StrF(xshift,0)+","+StrF(yshift,0))
		
		
		
		
		Select \type
				
			Case #SVG_Element_Box, #SVG_Element_Image, #SVG_Element_PBImage, #SVG_Element_Base64Image
				;ReportRectF(@o)
				\w = w
				\h = h
				Select maintain
					Case #SVG_Orient_Center
						\x + xshift
						\y + yshift
					Case #SVG_Orient_TopLeft
					Case #SVG_Orient_TopRight
						\x + (o\w-w)
					Case #SVG_Orient_BottomRight
						\x + (o\w-w)
						\y + (o\h-h)
					Case #SVG_Orient_BottomLeft
						\y + (o\h-h)
				EndSelect
				
			Case #SVG_Element_Circle, #SVG_Element_Ellipse
				\rx = w/2
				\ry = h/2
				If \rx<>\ry
					\type = #SVG_Element_Ellipse
				EndIf
				Select maintain
					Case #SVG_Orient_TopLeft
						\cx - ((o\w-w)/2)
						\cy - ((o\h-h)/2)
					Case #SVG_Orient_TopRight
						\cx + ((o\w-w)/2)
						\cy - ((o\h-h)/2)
					Case #SVG_Orient_BottomRight
						\cx + ((o\w-w)/2)
						\cy + ((o\h-h)/2)
					Case #SVG_Orient_BottomLeft
						\cx - ((o\w-w)/2)
						\cy + ((o\h-h)/2)
				EndSelect
				
				
			Case #SVG_Element_Polygon, #SVG_Element_Polyline, #SVG_Element_Line, #SVG_Element_Path, #SVG_Element_TextPath ;`
				If \path\pathpoints>0
					;ReportElementBounds(s,e)
					nw.f = xpc*o\w
					;R("OLD WIDTH: "+StrF(o\w)+c13+"NEW WIDTH: "+StrF(nw))
					Select maintain
						Case #SVG_Orient_TopRight
							yshift=0
						Case #SVG_Orient_BottomLeft
							xshift = 0
					EndSelect
					;R("XPC: "+StrF(xpc)+c13+"YPC: "+StrF(ypc)+c13+c13+"SHIFT: "+StrF(xshift,0)+","+StrF(yshift,0))
					For p = 1 To \path\pathpoints
						oldx.f = \path\pathpoint(p)\x : oldy.f = \path\pathpoint(p)\y
						
						\path\pathpoint(p)\x = o\x+(xpc*(\path\pathpoint(p)\x-o\x)) + xshift
						\path\pathpoint(p)\y = o\y+(ypc*(\path\pathpoint(p)\y-o\y)) + yshift
						If \path\pathpoint(p)\intermediary_points
							For i = 1 To \path\pathpoint(p)\intermediary_points
								\path\pathpoint(p)\intermediary_point(i)\x = o\x+(xpc*(\path\pathpoint(p)\intermediary_point(i)\x-o\x)) + xshift
								\path\pathpoint(p)\intermediary_point(i)\y = o\y+(ypc*(\path\pathpoint(p)\intermediary_point(i)\y-o\y)) + yshift
							Next i
						EndIf
						
						; elliptical arc stuff...
						If \path\pathpoint(p)\mode = #SVG_Path_EllipticalArc
							
							; get previous x/y
							x0.f = \path\pathpoint(p-1)\x
							y0.f = \path\pathpoint(p-1)\y
							
							If \path\pathpoint(p)\x_rotation = 0 And xpc=ypc And \rx=\ry
								;R("equal dimensions, no rotation -> no need for special calculations")
								\rx*xpc
								\ry*ypc
							Else
								
							EndIf
						EndIf
					Next p
				Else
					R("NO PATHPOINTS")
				EndIf
				
			Case #SVG_Element_Text ;`
				
				
			Case #SVG_Element_WrappedText ;'
				
				
			Case #SVG_Element_Group ;`
				;R(\id+c13+"SetSVGElementDimensions"+c13+c13+"WPC: "+StrF(xpc,0)+c13+"HPC: "+StrF(ypc,0))
				;ReportRectF(@o)
				
				SVGInternal_ForEachElementInSVGGroup(s,e,a)
				
				; get current bounds...
				KnowBounds(s,a)
				CopyRectF(@svgmem(s)\element(a)\bound,@bb.RectF)
				ix.f = (bb\x - o\x)
				iy.f = (bb\y - o\y)
				
				; resize...
				SVGInternal_SetElementDimensions(s,a,xpc*bb\w,ypc*bb\h,#SVG_Orient_TopLeft)
				
				; restore position relative to group (x,y)...
				ix*xpc
				iy*ypc
				SVGInternal_ShiftElement(s,a,o\x+ix+xshift-bb\x,o\y+iy+yshift-bb\y)
			Next a
			
	EndSelect
	\bounds_known = #False
	
EndWith

ProcedureReturn #True

EndProcedure


Procedure.b SVGInternal_ScaleElement(s.i,e.i,wpc.f,hpc.f,maintain.b=#SVG_Orient_Center)
	
	With svgmem(s)\element(e)
		KnowBounds(s,e)
		;R(\id+c13+"ScaleSVGElement"+c13+c13+"WPC: "+StrF(wpc,0)+c13+"HPC: "+StrF(hpc,0))
		w.f = \bound\w /100 *wpc
		h.f = \bound\h /100 *hpc
		;R("ScaleSVGElement"+c13+c13+StrF(\bound\w,0)+","+StrF(\bound\h,0)+c13+c13+StrF(w,0)+","+StrF(h,0))
		;ReportRectF(@\bound)
		;PolyRect(@\bound,#Yellow)
		ProcedureReturn SVGInternal_SetElementDimensions(s,e,w,h,maintain)
	EndWith
	
EndProcedure




Procedure.b SVGInternal_SetElementPosition(s.i,e.i,x.f,y.f)
	
	With svgmem(s)\element(e)
		Select \type
			Case #SVG_Element_Group ;`
				KnowBounds(s,e)
				SVGInternal_ForEachElementInSVGGroup(s,e,a)
				ofx.f = svgmem(s)\element(a)\bound\x - svgmem(s)\element(e)\bound\x
				ofy.f = svgmem(s)\element(a)\bound\y - svgmem(s)\element(e)\bound\y
				SVGInternal_SetElementPosition(s,a,x+ofx,y+ofy)
			Next a
			\bounds_known = #False
		Case #SVG_Element_Circle, #SVG_Element_Ellipse
			\cx - (\cx-\rx-x)
			\cy - (\cy-\ry-y)
			\bounds_known = #False
		Case #SVG_Element_Path, #SVG_Element_Line, #SVG_Element_Polygon, #SVG_Element_Polyline
			;R(\element(e)\content)
			KnowBounds(s,e)
			tx.f = \bound\x
			ty.f = \bound\y
			ox.f = -tx+x
			oy.f = -ty+y
			
			;R("CURRENT LX: "+StrF(tx)+c13+"DESIRED LX: "+StrF(x)+c13+c13+"SHIFT: "+StrF(ox))
			
			SVGInternal_ShiftElement(s,e,-tx+x,-ty+y)
			\bounds_known = #False
		Default
			\x = x
			\y = y
			\bounds_known = #False
	EndSelect
EndWith

ProcedureReturn #True

EndProcedure



Procedure.b SVGInternal_ApplyTransforming(s.i,e.i,transformdata.s)
	; transform="translate(0,239) scale(0.016963,-0.016963)"
	items = CountString(transformdata,")")
	If items
		For a = 1 To items
			item.s = StringField(transformdata,a,")")+")"
			item = Trim(item)
			;translate(0,239)
			mode.s = StringField(item,1,"(") ;translate
			EnsureThisNotStart(item,mode)
			EnsureThisNotEnd(item,")")
			;0,239
			item+#cma+#cma+#cma+#cma+#cma+#cma+#cma+#cma
			Select LCase(mode)
				Case "translate"
					x.f = ValF(StringField(item,1,#cma))
					y.f = ValF(StringField(item,2,#cma))
					SVGInternal_SetElementPosition(s,e,x,y)
				Case "scale"
					wpc.f = ValF(StringField(item+#cma,1,#cma))
					hpcs.s = StringField(item+#cma,2,#cma)
					If hpcs
						hpc.f = ValF(hpcs)
					Else
						hpc = wpc
					EndIf
					SVGInternal_ScaleElement(s,e,wpc,hpc,#SVG_Orient_TopLeft)
				Case "matrix"
					;`
				Case "rotate"
					item+","
					angle.f = ValF(StringField(item,1,#cma))
					cxs.s = StringField(item,2,#cma)
					cx.f = #PB_Ignore : If cxs : cx=ValF(cxs) : EndIf
					cys.s = StringField(item,3,#cma)
					cy.f = #PB_Ignore : If cys : cy=ValF(cys) : EndIf
					SVGInternal_RotateElement(s,e,angle,cx,cy)
				Case "skewx"
					angle.f = ValF(StringField(item,1,#cma))
					;`
				Case "skewy"
					angle.f = ValF(StringField(item,1,#cma))
					;`
			EndSelect
		Next a
	EndIf
EndProcedure



;-

Procedure.b DecipherFillInstruction(info.s,*s.SVG_Structure,*mode,*value)
	
	linfo.s = LCase(info)
	
	Select linfo
		Case "red"
			PokeB(*mode,#SVG_FillMode_Color)
			PokeI(*value,#Red)
			ProcedureReturn #True
		Case "green"
			PokeB(*mode,#SVG_FillMode_Color)
			PokeI(*value,#Green)
			ProcedureReturn #True
		Case "blue"
			PokeB(*mode,#SVG_FillMode_Color)
			PokeI(*value,#Blue)
			ProcedureReturn #True
		Case "white"
			PokeB(*mode,#SVG_FillMode_Color)
			PokeI(*value,#White)
			ProcedureReturn #True
		Case "black"
			PokeB(*mode,#SVG_FillMode_Color)
			PokeI(*value,#Black)
			ProcedureReturn #True
	EndSelect
	
	If Left(linfo,1) = "#"
		PokeI(*value,Hex2RGB(linfo))
		PokeB(*mode,#SVG_FillMode_Color)
		ProcedureReturn #True
	EndIf
	
	If Left(linfo,4) = "rgb("
		linfo = RemoveString(linfo,"rgb(")
		linfo = RemoveString(linfo,")")+#cma+#cma+#cma
		r = Val(StringField(linfo,1,#cma))
		g = Val(StringField(linfo,2,#cma))
		b = Val(StringField(linfo,3,#cma))
		PokeI(*value,RGB(r,g,b))
		PokeB(*mode,#SVG_FillMode_Color)
		ProcedureReturn #True
	EndIf
	
	If Left(linfo,5) = "url(#"
		n.s = EnsureNotStart(linfo,"url(#")
		EnsureThisNotEnd(n,")")
		;R(n)
		For g = 1 To *s\gradients
			If *s\gradient(g)\id = n
				PokeB(*mode,#SVG_FillMode_Gradient)
				PokeI(*value,g)
				ProcedureReturn #True
			EndIf
		Next g
		For p = 1 To *s\patterns
			If *s\pattern(p)\id = n
				PokeB(*mode,#SVG_FillMode_Pattern)
				PokeI(*value,p)
				ProcedureReturn #True
			EndIf
		Next p
	EndIf
	
	PokeB(*mode,#SVG_FillMode_None)
	PokeI(*value,0)
	ProcedureReturn #True
	
EndProcedure



Procedure.b SVGInternal_StyleFromString(*svg.SVG_Structure,*e.SVG_Element_Structure,t.s)
	
	t = RemoveString(t,"{")
	t = RemoveString(t,"}")
	t = Trim(t)
	If Nots t
		ProcedureReturn #True
	EndIf
	
	dlm.s = ";"
	EnsureThisEnd(t,dlm)
	
	With *e
		For a = 1 To CountString(t,dlm)
			item.s = StringField(t,a,dlm)
			itemname.s = Trim(StringField(item,1,":"))
			itemvalue.s = Trim(StringField(item+":",2,":"))
			If Nots itemname Or Nots itemvalue : Continue : EndIf
			
			Select LCase(itemname)
				Case "font-name"
					\fontname = itemvalue
				Case "font-family"
					\fontname = StringField(itemvalue+",",1,",")
				Case "font-size"
					\fontsize = ValF(itemvalue)
				Case "fill"
					DecipherFillInstruction(itemvalue,*svg,@\fillmode,@\fillvalue)
				Case "fill-opacity"
					\opacity = PercentageOrDecimalTo255(itemvalue)
				Case "opacity"
					\opacity = PercentageOrDecimalTo255(itemvalue)
					\strokeopacity = \opacity
				Case "stroke-color", "stroke"
					DecipherFillInstruction(itemvalue,*svg,@mode.b,@\strokecolor)
				Case "stroke-width"
					\strokewidth = ValF(itemvalue)
					;R(StrF(\strokewidth))
				Case "stroke-opacity"
					\strokeopacity = PercentageOrDecimalTo255(itemvalue)
				Case "stroke-linecap"
					\strokelinecap = IdentifyStrokeLinecap(itemvalue)
				Case "stroke-linejoin"
					\strokelinejoin = IdentifyStrokeLinejoin(itemvalue)
				Case "cursor"
					If Left(itemvalue,4) = "url("
						\cursor = #SVG_Cursor_URI
						\cursor_url = EnsureNotStart(EnsureNotEnd(itemvalue,")"),"url(")
					Else
						\cursor = IdentifyCursor(itemvalue)
					EndIf
				Case "visibility"
					If itemvalue = "hidden"
						\visibility = #False
					EndIf
				Default
					styleattr.s+itemname+":"+itemvalue+"; "
			EndSelect
			
		Next a
		
		If styleattr
			*e\styling + styleattr
		EndIf
		
		ProcedureReturn #True
	EndWith
	
EndProcedure











;-
;------------------------------------------------------------------------------
;- LIBRARY FUNCTIONS ---------------------------------------
;------------------------------------------------------------------------------


Procedure.b GetSVGFileDimensions(filename.s,*w,*h)
	
	Protected.f w, h
	
	f = ReadFile(#PB_Any,filename)
	If f
		While Not Eof(f)
			txt.s = ReadString(f)
			detect = FindString(txt,"<svg ",1)
			If detect
				While Not Eof(f)
					new.s = ReadString(f)
					txt+new+c32
					If FindString(new,">",1)
						Break
					EndIf
				Wend
			Else
				Continue
			EndIf
			enddetect = FindString(txt,">",detect+4)
			If detect And enddetect
				txt = Mid(txt,detect,enddetect-detect)+" "
				txt = LCase(txt)
				txt = ReplaceString(txt,c39,c34)
				
				vbdetect = FindString(txt,"viewbox=",1)
				If vbdetect
					; using the viewbox attribute...
					endvbdetect = FindString(txt,c34,vbdetect+9)
					If vbdetect And endvbdetect
						;R("NOW: "+Mid(txt,vbdetect+9,Len(txt)))
						ws.s = Mid(txt,vbdetect+9,endvbdetect-vbdetect-8)
						;R("RAW WS: *"+ws+"*")
						ws = Trim(ws,Chr(32))
						ws = Trim(ws,Chr(34))
						ws = Trim(ws,Chr(32))
						ws = Trim(ws,Chr(34))
						;R("TRIMMED WS: *"+ws+"*")
						ws+c32
						bx.f = ValF(StringField(ws,1,c32))
						by.f = ValF(StringField(ws,2,c32))
						bw.f = ValF(StringField(ws,3,c32))
						bh.f = ValF(StringField(ws,4,c32))
						w = bw-bx
						h = bh-by
						;R("BX: "+StrF(bx)+c13+"BY: "+StrF(by)+c13+"BW: "+StrF(bw)+c13+"BH: "+StrF(bh)+c13+c13+"W: "+StrF(w)+c13+"H: "+StrF(h))
						;w = ValF(ws)
					EndIf
				Else
					detect = FindString(txt,"width=",1)
					enddetect = FindString(txt," ",detect+5)
					If detect And enddetect
						ws.s = Mid(txt,detect+6,enddetect-detect-6)
						ws = Trim(ws,Chr(32))
						ws = Trim(ws,Chr(34))
						ws = Trim(ws,Chr(32))
						ws = Trim(ws,Chr(34))
						;R("*"+ws+"*")
						w = ValF(ws)
					EndIf
					detect = FindString(txt,"height=",1)
					enddetect = FindString(txt," ",detect+6)
					If detect And enddetect
						hs.s = Mid(txt,detect+7,enddetect-detect-7)
						hs = Trim(hs,Chr(32))
						hs = Trim(hs,Chr(34))
						hs = Trim(hs,Chr(32))
						hs = Trim(hs,Chr(34))
						;R("*"+hs+"*")
						h = ValF(hs)
					EndIf
				EndIf
				Break
			EndIf
		Wend
		CloseFile(f)
	EndIf
	
	PokeF(*w,w)
	PokeF(*h,h)
	
EndProcedure




Macro FreeSVG(s)
	If IsSVG(s)
		ClearStructure(@svgmem(s),SVG_Structure)
	EndIf
EndMacro


Procedure.b IsSVG(s.i)
	
	If s>0 And s<=#MaxSVGs
		If svgmem(s)\in_use
			ProcedureReturn #True
		EndIf
	EndIf
	
	ProcedureReturn #False
	
EndProcedure


Procedure.i CreateSVG(s.i=#PB_Any,w.f=0,h.f=0)
	
	If s = #PB_Any
		s = FirstAvailableSVGSlot()
	EndIf
	If Not ValidSVGNumber(s)
		For a = 1 To #maxsvgs
			test.s+LSet(Str(a),3,Chr(32))+" "
			If svgmem(a)\in_use
				test+"USING"
			EndIf
			test+c13
		Next a
		R("CAN'T CREATE NEW SVG."+c13+Str(#MaxSVGs)+c13+c13+test)
		ProcedureReturn 0
	EndIf
	
	InitiateSVG(s)
	svgmem(s)\setw = w
	svgmem(s)\seth = h
	
	ProcedureReturn s
	
EndProcedure


Procedure.i CopySVG(s1.i,s2.i=#PB_Any,*namechanger.SVGPassString=0)
	
	If Not IsSVG(s1) : ProcedureReturn 0 : EndIf
	If s2=#PB_Any
		s2 = FirstAvailableSVGSlot()
	EndIf
	If Not ValidSVGNumber(s2)
		ProcedureReturn 0
	EndIf
	
	CopyStructure(@svgmem(s1),@svgmem(s1),SVG_Structure)
	svgmem(s2)\in_use = #True
	
	ProcedureReturn s2
	
EndProcedure


Procedure.i SVGSize(s.i)
	
	With svgmem(s)
		If \elements
			For a = 1 To \elements
				If \element(a)\in_use
					If \element(a)\type<>#SVG_Element_GroupCloser
						num+1
					EndIf
				EndIf
			Next a
			ProcedureReturn num
		EndIf
	EndWith
	
	ProcedureReturn 0
	
EndProcedure


Procedure.i SVGIndex()
	If svgdata\s
		If svgdata\e
			ProcedureReturn svgdata\e
		EndIf
	EndIf
	ProcedureReturn 0
EndProcedure






Procedure.b SVG_SetVisiblePortion(nx,ny,nw,nh)
	s=svgdata\s
	If Not s : ProcedureReturn #False : EndIf
	svgmem(s)\visible\x = nx
	svgmem(s)\visible\y = ny
	svgmem(s)\visible\w = nw
	svgmem(s)\visible\h = nh
	svgmem(s)\setw = nw-nx
	svgmem(s)\seth = nh-ny
	ProcedureReturn #True
EndProcedure

Procedure.b SVG_EnsureAllVisible()
	
	s = svgdata\s
	If Not s : ProcedureReturn #False : EndIf
	With svgmem(s)
		If \elements<1 : ProcedureReturn #True : EndIf
		;CalculateTotalSVGElementBounds(s)
		SafeMasterForAllElements
		ProcedureReturn SVG_SetVisiblePortion(master\x,master\y,master\w,master\h)
	EndWith
	
EndProcedure




Procedure.b SVG_StartDrawing(s)
	
	If svgdata\s
		MessageRequester("Error.","SVG_StopDrawing() needs to be called before calling SVG_StartDrawing() again!",0)
		ProcedureReturn #False
	EndIf
	If IsSVG(s)
		svgdata\s = s
		svgdata\FrontColor = -1
		svgdata\BackColor = -1
		
		svgdata\FontName = "Helvetica"
		svgdata\FontSize = 16
		ProcedureReturn #True
	Else
		ProcedureReturn #False
	EndIf
	
EndProcedure

Macro SVG_StopDrawing()
	svgdata\s = 0
	svgdata\e = 0
	svgdata\g = 0
	svgdata\p = 0
	svgdata\f = 0
EndMacro




Macro SVG_SetTitle(t)
	svgmem(svgdata\s)\md[#SVG_Metadata_Title] = t
EndMacro




Macro SVG_FrontColor(clr)
	svgdata\FrontColor=clr
EndMacro

Macro SVG_BackColor(clr)
	svgdata\BackColor=clr
EndMacro

Macro SVG_DrawingFont(fname,fsize)
	If fname<>""
		svgdata\FontName = fname
	EndIf
	If fsize>0
		svgdata\FontSize = fsize
	EndIf
EndMacro

Procedure.b SVG_SetArrowDimensions(stalklength.f,stalkwidth.f,headlength.f,headwidth.f)
	svgdata\arrow_stalkl = Beat(0.1,stalklength)
	svgdata\arrow_stalkw = Beat(0.1,stalkwidth)
	svgdata\arrow_headl = Beat(0.1,headlength)
	svgdata\arrow_headw = Beat(0.1,headwidth)
EndProcedure




Macro SVG_SetBackground(mode,value=0)
	svgmem(svgdata\s)\bg_mode = mode
	svgmem(svgdata\s)\bg_value = value
EndMacro




Macro ForEachSVGElement(svg_ge)
	For svg_ge = 1 To svgmem(svgdata\s)\elements
		If svgmem(svgdata\s)\element(svg_ge)\type=#SVG_Element_GroupCloser : Continue : EndIf
		If Not svgmem(svgdata\s)\element(svg_ge)\in_use : Continue : EndIf
		SVG_SelectElement(svg_ge)
	EndMacro
	
	
	Macro ForEachElementInSVGGroup(g,svg_ge)
		svg_lastge = CloserForSVGGroup(svgdata\s,g)
		;R("S: "+Str(s)+c13+"G: "+Str(g)+c13+"SVG_LASTGE: "+Str(svg_lastge))
		For svg_ge = g+1 To svg_lastge-1
			If svgmem(svgdata\s)\element(svg_ge)\type=#SVG_Element_GroupCloser : Continue : EndIf
			If Not svgmem(svgdata\s)\element(svg_ge)\in_use : R("NOT IN USE") : Continue : EndIf
			SVG_SelectElement(svg_ge)
		EndMacro
		
		
		Macro ForEachSVGElementOfType(tp,svg_ge)
			For svg_ge = 1 To svgmem(svgdata\s)\elements
				If svgmem(svgdata\s)\element(svg_ge)\type=#SVG_Element_GroupCloser : Continue : EndIf
				If svgmem(svgdata\s)\element(svg_ge)\type<>tp : Continue : EndIf
				If Not svgmem(svgdata\s)\element(svg_ge)\in_use : Continue : EndIf
				SVG_SelectElement(svg_ge)
			EndMacro
			
			
			
			
			
			
			Enumeration 1
				#SVG_Cycle_Definitions
				#SVG_Cycle_Elements
			EndEnumeration
			
			Procedure.b SVGInternal_DecipherStroke(s.i,e.i,info.s)
				
				With svgmem(s)\element(e)
					DecipherFillInstruction(info,@svgmem(s),@mode.b,@svgmem(s)\element(e)\strokecolor)
				EndWith
				
			EndProcedure
			Procedure.b NUSVGInternal_DecipherStroke(*svg.SVG_Structure,*e.SVG_Element_Structure,info.s)
				
				DecipherFillInstruction(info,*svg,@mode.b,*e\strokecolor)
				
			EndProcedure
			
			Macro ParseSVGElementAttributes1(s,e)
				If ExamineXMLAttributes(*nd)
					While NextXMLAttribute(*nd)
						anam.s = XMLAttributeName(*nd)
						aval.s = XMLAttributeValue(*nd)
						If Nots anam Or Nots aval : Continue : EndIf
						Select anam
							Case "id"
								svgmem(s)\element(e)\id = aval
							Case "class"
								svgmem(s)\element(e)\class = aval
							Case "d"
								AddToSVGPath(aval,@svgmem(s)\element(e)\path)
							Case "points"
								SVGInternal_PolylineDataFromString(s,e,aval,#False)
							Case "x", "x1"
								svgmem(s)\element(e)\x = ValF(aval)
							Case "y", "y1"
								svgmem(s)\element(e)\y = ValF(aval)
							Case "x2"
								svgmem(s)\element(e)\w = ValF(aval)-svgmem(s)\element(e)\x
							Case "y2"
								svgmem(s)\element(e)\h = ValF(aval)-svgmem(s)\element(e)\y
							Case "width"
								svgmem(s)\element(e)\w = ValF(aval)
							Case "height"
								svgmem(s)\element(e)\h = ValF(aval)
							Case "cx"
								svgmem(s)\element(e)\cx = ValF(aval)
							Case "cy"
								svgmem(s)\element(e)\cy = ValF(aval)
							Case "r"
								svgmem(s)\element(e)\rx = ValF(aval)
								svgmem(s)\element(e)\ry = ValF(aval)
							Case "rx"
								svgmem(s)\element(e)\rx = ValF(aval)
							Case "ry"
								svgmem(s)\element(e)\ry = ValF(aval)
							Case "visibility"
								svgmem(s)\element(e)\visibility = #True
								If aval="hidden"
									svgmem(s)\element(e)\visibility = #False
								EndIf
							Case "fill"
								DecipherFillInstruction(aval,@svgmem(s),@svgmem(s)\element(e)\fillmode,@svgmem(s)\element(e)\fillvalue)
							Case "fill-opacity", "opacity"
								svgmem(s)\element(e)\opacity = PercentageOrDecimalTo255(aval)
							Case "stroke"
								SVGInternal_DecipherStroke(s,e,aval)
							Case "stroke-color"
								DecipherFillInstruction(aval,@svgmem(s),@dummybyte.b,@svgmem(s)\element(e)\strokecolor)
							Case "stroke-opacity"
								svgmem(s)\element(e)\strokeopacity = PercentageOrDecimalTo255(aval)
							Case "stroke-width"
								svgmem(s)\element(e)\strokewidth = Val(aval)
							Case "stroke-linecap"
								svgmem(s)\element(e)\strokelinecap = IdentifyStrokeLinecap(aval)
							Case "stroke-linejoin"
								svgmem(s)\element(e)\strokelinejoin = IdentifyStrokeLinejoin(aval)
							Case "font-name"
								svgmem(s)\element(e)\fontname = aval
							Case "font-family"
								svgmem(s)\element(e)\fontname = StringField(aval+#cma,1,#cma)
							Case "font-size"
								svgmem(s)\element(e)\fontsize = Val(aval)
								
							Case "mask"
								aval = RemoveString(aval,"url(#")
								EnsureThisNotEnd(aval,")")
								For m = 1 To svgmem(s)\elements
									If svgmem(s)\element(m)\ismask
										If svgmem(s)\element(m)\id = aval
											svgmem(s)\element(e)\mask = m
											Break 1
										EndIf
									EndIf
								Next m
							Case "style"
								SVGInternal_StyleFromString(@svgmem(s),@svgmem(s)\element(e),aval)
							Case "src", "xlink:href"
								svgmem(s)\element(e)\filename = aval
								
								
							Case #SVG_Action_OnActivate, #SVG_Action_OnClick, #SVG_Action_OnFocusIn, #SVG_Action_OnFocusOut, #SVG_Action_OnLoad, #SVG_Action_OnMouseDown, #SVG_Action_OnMouseMove, #SVG_Action_OnMouseOut, #SVG_Action_OnMouseOver, #SVG_Action_OnMouseUp
								svgmem(s)\element(e)\action(anam) = aval
								
								
							Case "transform"
								transformdata.s = aval
								
								
							Case "xlink:href"
								anchor\href = aval
							Case "target", "xlink:target"
								anchor\href_target = HyperlinkTargetNumber(aval)
								
							Default
								svgmem(s)\element(e)\attribute(anam) = aval
						EndSelect
					Wend
				EndIf
			EndMacro
			
			Macro ParseSVGElementAttributes2(s,e)
				
				If transformdata
					SVGInternal_ApplyTransforming(s,e,transformdata)
				EndIf
				
			EndMacro
			
			
Macro NUParseSVGElementAttributes1(svgname,elname)
	If ExamineXMLAttributes(*nd)
		While NextXMLAttribute(*nd)
			anam.s = XMLAttributeName(*nd)
			aval.s = XMLAttributeValue(*nd)
			If Nots anam Or Nots aval : Continue : EndIf
			Select anam
				Case "id"
					elname\id = aval
				Case "class"
					elname\class = aval
				Case "d"
					AddToSVGPath(aval,@elname)
				Case "points"
					NUSVGInternal_PolylineDataFromString(@svgname,@elname,aval,#False)
				Case "x", "x1"
					elname\x = ValF(aval)
				Case "y", "y1"
					elname\y = ValF(aval)
				Case "x2"
					elname\w = ValF(aval)-elname\x
				Case "y2"
					elname\h = ValF(aval)-elname\y
				Case "width"
					elname\w = ValF(aval)
				Case "height"
					elname\h = ValF(aval)
				Case "cx"
					elname\cx = ValF(aval)
				Case "cy"
					elname\cy = ValF(aval)
				Case "r"
					elname\rx = ValF(aval)
					elname\ry = ValF(aval)
				Case "rx"
					elname\rx = ValF(aval)
				Case "ry"
					elname\ry = ValF(aval)
				Case "visibility"
					elname\visibility = #True
					If aval="hidden"
						elname\visibility = #False
					EndIf
				Case "fill"
					DecipherFillInstruction(aval,@svgname,@elname\fillmode,@elname\fillvalue)
				Case "fill-opacity", "opacity"
					elname\opacity = PercentageOrDecimalTo255(aval)
				Case "stroke"
					NUSVGInternal_DecipherStroke(@svgname,@elname,aval)
				Case "stroke-color"
					DecipherFillInstruction(aval,@svgname,@dummybyte.b,@elname\strokecolor)
				Case "stroke-opacity"
					elname\strokeopacity = PercentageOrDecimalTo255(aval)
				Case "stroke-width"
					elname\strokewidth = Val(aval)
				Case "stroke-linecap"
					elname\strokelinecap = IdentifyStrokeLinecap(aval)
				Case "stroke-linejoin"
					elname\strokelinejoin = IdentifyStrokeLinejoin(aval)
				Case "font-name"
					elname\fontname = aval
				Case "font-family"
					elname\fontname = StringField(aval+#cma,1,#cma)
				Case "font-size"
					elname\fontsize = Val(aval)
					
				Case "mask"
					aval = RemoveString(aval,"url(#")
					EnsureThisNotEnd(aval,")")
					For m = 1 To svgname\elements
						If svgname\element(m)\ismask
							If svgname\element(m)\id = aval
								elname\mask = m
								Break 1
							EndIf
						EndIf
					Next m
				Case "style"
					SVGInternal_StyleFromString(@svgmem(s),@elname,aval)
				Case "src", "xlink:href"
					elname\filename = aval
					
					
				Case #SVG_Action_OnActivate, #SVG_Action_OnClick, #SVG_Action_OnFocusIn, #SVG_Action_OnFocusOut, #SVG_Action_OnLoad, #SVG_Action_OnMouseDown, #SVG_Action_OnMouseMove, #SVG_Action_OnMouseOut, #SVG_Action_OnMouseOver, #SVG_Action_OnMouseUp
					elname\action(anam) = aval
					
					
				Case "transform"
					transformdata.s = aval
					
					
				Case "xlink:href"
					elname\href = ReplaceString(aval,"&amp;","&")
				Case "target", "xlink:target"
					elname\href_target = HyperlinkTargetNumber(aval)
					
				Default
					;elname\attrs + anam+"="+c34+aval+c34+c32
					elname\attribute(anam) = aval
			EndSelect
		Wend
	EndIf
EndMacro

			
			Procedure.i DateFromString(t.s)
				
				CompilerIf Defined(d1,#PB_Constant)=0
					#d1 = "|"
				CompilerEndIf
				
				t = Trim(t)
				t = ReplaceString(t,"/",#d1)
				t = ReplaceString(t,"-",#d1)
				t = ReplaceString(t,":",#d1)
				t = ReplaceString(t,".",#d1)
				If Right(t,1)<>#d1 : t + #d1 : EndIf
				
				
				d = Val(StringField(t,1,#d1))
				y = Val(StringField(t,3,#d1))
				
				ms.s = LCase(StringField(t,2,#d1))
				For mn = 1 To 12
					If ms = Left(LCase(MonthName(mn)),3)
						m = mn
						Break 1
					EndIf
				Next mn
				
				ProcedureReturn Date(y,m,d,0,0,0)
				
			EndProcedure
			
			Procedure.b SVGInternal_ExtractDublinCoreMetadata(s.i,*nd,CurrentSubLevel.i)
				
				;R("DUbLIN CORE")
				
				;*cnd = ChildXMLNode(*nd)
				;If Not *cnd
				;    ProcedureReturn #False
				;EndIf
				
				found.b = #False
				*dnd = *nd
				For a = 1 To 10
					*dnd = ChildXMLNode(*dnd)
					If GetXMLNodeName(*dnd)="cc:Work"
						found=#True
						Break
					EndIf
					SetXMLNodeName(*dnd,"null")
				Next
				If Not found
					R("CAN'T FIND METADATA")
					ProcedureReturn #False
				EndIf
				SetXMLNodeName(*dnd,"null")
				
				
				; <dc:contributor><cc:Agent><dc:title>contributors</dc:title></cc:Agent></dc:contributor>
				; <dc:language>english</dc:language>
				; <dc:subject><rdf:Bag></rdf:Bag></dc:subject>
				; <dc:subject><rdf:Bag><rdf:li>keyword1</rdf:li><rdf:li>keyword2</rdf:li></rdf:Bag></dc:subject>
				
				*cnd = ChildXMLNode(*dnd)
				While *cnd
					
					anam.s = GetXMLNodeName(*cnd)
					kw.b=#False : If FindString(anam,"subject",1) : kw=#True : EndIf
					
					*fnd = *cnd
					While ChildXMLNode(*fnd)
						*fnd = ChildXMLNode(*fnd)
					Wend
					If kw
						;*fnd = ParentXMLNode(*fnd)
					EndIf
					
					aval.s = GetXMLNodeText(*fnd)
					Select anam
						Case "dc:title"
							svgmem(s)\md[#SVG_Metadata_Title] = aval
						Case "dc:description"
							svgmem(s)\md[#SVG_Metadata_Description] = aval
						Case "dc:contributor"
							svgmem(s)\md[#SVG_Metadata_Contributors] = aval
						Case "dc:coverage"
							svgmem(s)\md[#SVG_Metadata_Coverage] = aval
						Case "dc:creator"
							svgmem(s)\md[#SVG_Metadata_Creator] = aval
						Case "dc:language"
							svgmem(s)\md[#SVG_Metadata_Language] = aval
						Case "dc:publisher"
							svgmem(s)\md[#SVG_Metadata_Publisher] = aval
						Case "dc:rights"
							svgmem(s)\md[#SVG_Metadata_Rightsholder] = aval
						Case "dc:source"
							svgmem(s)\md[#SVG_Metadata_Source] = aval
						Case "dc:subject"
							;R("FIRST:"+c13+c13+GetXMLNodeName(*fnd)+c13+GetXMLNodeText(*fnd))
							While *fnd
								;R(GetXMLNodeName(*fnd)+c13+GetXMLNodeText(*fnd))
								svgmem(s)\md[#SVG_Metadata_Keywords] + GetXMLNodeText(*fnd) + #d1
								*fnd = NextXMLNode(*fnd)
							Wend
						Case "dc:date"
							svgmem(s)\md[#SVG_Metadata_Date] = Str(DateFromString(aval))
					EndSelect
					
					*cnd = NextXMLNode(*cnd)
				Wend
				
			EndProcedure
			
			Procedure ParseSVGXML(s.i,*nd, CurrentSublevel.i,cycle.b,href.s="",target.b=0)
				
				nodename.s = GetXMLNodeName(*nd)
				If nodename="null" : ProcedureReturn : EndIf
				
				Select XMLNodeType(*nd)
					Case #PB_XML_Comment
						AddSVGElement(s,#SVG_Element_Comment,"",0,0,0,0,0,GetXMLNodeText(*nd))
					Case #PB_XML_Normal, #PB_XML_Instruction
					Default
						ProcedureReturn
				EndSelect
				
				slavesvg.SVG_Structure
				anchor.SVG_Element_Structure
				
				Select cycle
					Case #SVG_Cycle_Definitions
					;{
					Select LCase(nodename)
						Case "title"
							If CurrentSubLevel=1
								svgmem(s)\md[#SVG_Metadata_Title] = GetXMLNodeText(*nd)
							EndIf
						Case "desc"
							If CurrentSubLevel=1
								svgmem(s)\md[#SVG_Metadata_Description] = GetXMLNodeText(*nd)
							EndIf
						Case "defs"
							;Internal_DefinitionsFromXML(s,*nd)
						Case "mask"
							ms = svgmem(s)\elements
							SetXMLNodeName(*nd,"null")
							g = AddSVGElement(s,#SVG_Element_Group,"",0,0,0,0,0)
							ParseSVGElementAttributes1(s,g)
							*cnd = ChildXMLNode(*nd)
							While *cnd
								ParseSVGXML(s,*cnd,CurrentSublevel+1,#SVG_Cycle_Elements)
								SetXMLNodeName(*cnd,"null")
								*cnd = NextXMLNode(*cnd)
							Wend
							svgmem(s)\element(g)\ismask = #True
							AddSVGElement(s,#SVG_Element_GroupCloser,"",0,0,0,0,0)
						Case "lineargradient"
							SetXMLNodeName(*nd,"null")
							If ExamineXMLAttributes(*nd)
								While NextXMLAttribute(*nd)
									anam.s = XMLAttributeName(*nd)
									aval.s = XMLAttributeValue(*nd)
									Select anam
										Case "id"
											id.s = aval
										Case "x1"
											x.f = ValF(aval)
										Case "y1"
											y.f = ValF(aval)
										Case "x2"
											w.f = ValF(aval)
										Case "y2"
											h.f = ValF(aval)
									EndSelect
								Wend
							EndIf
							g = AddSVGGradient(s,#SVG_Gradient_Linear,id,x,y,w,h)
							*cnd = ChildXMLNode(*nd)
							While *cnd
								AddElement(svgmem(s)\gradient(g)\Point())
								If ExamineXMLAttributes(*cnd)
									While NextXMLAttribute(*cnd)
										anam.s = XMLAttributeName(*cnd)
										aval.s = XMLAttributeValue(*cnd)
										Select anam
											Case "offset"
												;R(aval+c13+c13+StrF(PercentageOrDecimalToDecimal(ValF(aval))))
												svgmem(s)\gradient(g)\Point()\position = PercentageOrDecimalToDecimal(ValF(aval))
											Case "stop-color"
												svgmem(s)\gradient(g)\Point()\color = Hex2RGB(aval)
											Case "stop-opacity"
												svgmem(s)\gradient(g)\Point()\opacity = ValF(aval)
										EndSelect
									Wend
								EndIf
								SetXMLNodeName(*cnd,"null")
								*cnd = NextXMLNode(*cnd)
							Wend
						Case "radialgradient"
							SetXMLNodeName(*nd,"null")
							If ExamineXMLAttributes(*nd)
								While NextXMLAttribute(*nd)
									anam.s = XMLAttributeName(*nd)
									aval.s = XMLAttributeValue(*nd)
									Select anam
										Case "id"
											id.s = aval
										Case "cx"
											x.f = ValF(aval)
										Case "cy"
											y.f = ValF(aval)
										Case "r"
											w.f = ValF(aval)
									EndSelect
								Wend
							EndIf
							g = AddSVGGradient(s,#SVG_Gradient_Radial,id,x,y,w,w)
							*cnd = ChildXMLNode(*nd)
							While *cnd
								AddElement(svgmem(s)\gradient(g)\Point())
								If ExamineXMLAttributes(*cnd)
									While NextXMLAttribute(*cnd)
										anam.s = XMLAttributeName(*cnd)
										aval.s = XMLAttributeValue(*cnd)
										Select anam
											Case "offset"
												svgmem(s)\gradient(g)\Point()\position = PercentageOrDecimalToDecimal(ValF(aval))
											Case "stop-color"
												svgmem(s)\gradient(g)\Point()\color = Hex2RGB(aval)
											Case "stop-opacity"
												svgmem(s)\gradient(g)\Point()\opacity = ValF(aval)
										EndSelect
									Wend
								EndIf
								SetXMLNodeName(*cnd,"null")
								*cnd = NextXMLNode(*cnd)
							Wend
						Case "filter"
							
						Case "pattern"
							
						Case "style"
							svgmem(s)\classinfo + c13+Trim(RemoveString(GetXMLNodeText(*nd),c10),c13)
						Case "xml-stylesheet"
							;R("STYLESHEET")
							t.s = GetXMLNodeText(*nd)
							seg.s = ExtractXMLSegment(t,"href")
							If Nots seg
								seg = ExtractXMLSegment(t,"xlink:href")
								If Nots seg
									seg = ExtractXMLSegment(t,"src")
								EndIf
							EndIf
							If seg
								AddElement(svgmem(s)\external_css_file())
								svgmem(s)\external_css_file() = seg
							EndIf
						Case "script"
							external.b = #False
							If ExamineXMLAttributes(*nd)
								While NextXMLAttribute(*nd)
									anam.s = XMLAttributeName(*nd)
									aval.s = XMLAttributeValue(*nd)
									If anam="src" Or anam="xlink:href"
										AddElement(svgmem(s)\external_js_file())
										svgmem(s)\external_js_file() = aval
										external=#True
										Break
									EndIf
								Wend
							EndIf
							If Not external
								svgmem(s)\javascript + GetXMLNodeText(*nd)
							EndIf
						Case "metadata"
							SetXMLNodeName(*nd,"null")
							If ExamineXMLAttributes(*nd)
								While NextXMLAttribute(*nd)
									anam.s = XMLAttributeName(*nd)
									aval.s = XMLAttributeValue(*nd)
									Select anam
										Case "id"
											If aval = "dublin_core_metadata"
												SVGInternal_ExtractDublinCoreMetadata(s,*nd,CurrentSubLevel)
											EndIf
									EndSelect
								Wend
							EndIf
						Default
							;R(nodename+c13+c13+GetXMLNodeText(*nd))
					EndSelect
				
				;}
				
				
				Case #SVG_Cycle_Elements
					;{
					Select LCase(nodename)
						Case "a"
							
							NUParseSVGElementAttributes1(slavesvg,anchor)
							
							SetXMLNodeName(*nd,"null")
							*cnd = ChildXMLNode(*nd)
							While *cnd
								ParseSVGXML(s,*cnd,CurrentSublevel+1,#SVG_Cycle_Elements,anchor\href,anchor\href_target)
								SetXMLNodeName(*cnd,"null")
								*cnd = NextXMLNode(*cnd)
							Wend
							;AddSVGElement(s,#SVG_Element_GroupCloser,"",0,0,0,0,0)
							ParseSVGElementAttributes2(s,e)
							
						Case "rect"
							e = AddSVGElement(s,#SVG_Element_Box,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							If svgmem(s)\element(e)\id = #SVG_BG_ID
								;`svgmem(s)\ = SVG_FindElement(id)
								svgmem(s)\bg_mode = svgmem(s)\element(e)\fillmode
								svgmem(s)\bg_value = svgmem(s)\element(e)\fillvalue
								SVGInternal_RemoveElement(s,e)
							EndIf
						Case "circle"
							;R("CIRCLE NODE TEXT: ."+GetXMLNodeText(*nd)+".")
							e = AddSVGElement(s,#SVG_Element_Circle,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							ParseSVGElementAttributes2(s,e)
						Case "ellipse"
							e = AddSVGElement(s,#SVG_Element_Ellipse,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							ParseSVGElementAttributes2(s,e)
						Case "polyline"
							e = AddSVGElement(s,#SVG_Element_Polyline,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							ParseSVGElementAttributes2(s,e)
						Case "polygon"
							e = AddSVGElement(s,#SVG_Element_Polygon,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							ParseSVGElementAttributes2(s,e)
						Case "path"
							e = AddSVGElement(s,#SVG_Element_Path,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							ParseSVGElementAttributes2(s,e)
						Case "textpath"
							e = AddSVGElement(s,#SVG_Element_TextPath,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							svgmem(s)\element(e)\content = GetXMLNodeText(*nd)
							ParseSVGElementAttributes2(s,e)
						Case "text"
							e = AddSVGElement(s,#SVG_Element_Text,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							svgmem(s)\element(e)\content = GetXMLNodeText(*nd)
							ParseSVGElementAttributes2(s,e)
						Case "image"
							e = AddSVGElement(s,#SVG_Element_Image,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							ParseSVGElementAttributes2(s,e)
						Case "line"
							e = AddSVGElement(s,#SVG_Element_Line,"",0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							With svgmem(s)\element(e)
								\path\pathpoints=2
								\path\pathpoint(1)\mode = #SVG_Path_SegmentStart
								\path\pathpoint(1)\x = \x
								\path\pathpoint(1)\y = \y
								\path\pathpoint(1)\intermediary_points = 0
								\path\pathpoint(2)\mode = #SVG_Path_Straight
								\path\pathpoint(2)\x = \x+\w
								\path\pathpoint(2)\y = \y+\h
								\path\pathpoint(2)\intermediary_points = 0
							EndWith
							ParseSVGElementAttributes2(s,e)
						Case "g"
							e = AddSVGElement(s,#SVG_Element_Group,"",0,0,0,0,0)
							ParseSVGElementAttributes1(s,e)
							SetXMLNodeName(*nd,"null")
							*cnd = ChildXMLNode(*nd)
							While *cnd
								ParseSVGXML(s,*cnd,CurrentSublevel+1,#SVG_Cycle_Elements)
								SetXMLNodeName(*cnd,"null")
								*cnd = NextXMLNode(*cnd)
							Wend
							AddSVGElement(s,#SVG_Element_GroupCloser,"",0,0,0,0,0)
							ParseSVGElementAttributes2(s,e)
					EndSelect
					
				EndSelect
				;}
				
				
				
				
				; Loop through all available child nodes and call this procedure again
				*cnd = ChildXMLNode(*nd)
				While *cnd
					ParseSVGXML(s,*cnd,CurrentSublevel+1,cycle)
					;SetXMLNodeName(*cnd,"null")
					*cnd = NextXMLNode(*cnd)
				Wend
				
			EndProcedure
			
			Procedure.i SVGFromXML(s.i,x.i)
				
				If XMLStatus(x) <> #PB_XML_Success
					Message$ = "Error in the XML file:" + Chr(13)
					Message$ + "Message: " + XMLError(x) + Chr(13)
					Message$ + "Line: " + Str(XMLErrorLine(x)) + "   Character: " + Str(XMLErrorPosition(x))
					MessageRequester("Error", Message$)
					ProcedureReturn #False
				EndIf
				
				*nd = MainXMLNode(x)
				If *nd
					InitiateSVG(s)
					svgdata\s = s
					If ExamineXMLAttributes(*nd)
						While NextXMLAttribute(*nd)
							anam.s = XMLAttributeName(*nd)
							aval.s = XMLAttributeValue(*nd)
							;R(anam+":   "+aval)
							lanam.s = LCase(anam)
							Select lanam
								Case "id"
									svgmem(s)\id = aval
								Case "viewbox"
									aval = ReplaceString(aval,#cma,c32)+c32+c32
									svgmem(s)\visible\x = ValF(StringField(aval,1,c32))
									svgmem(s)\visible\y = ValF(StringField(aval,2,c32))
									svgmem(s)\visible\w = ValF(StringField(aval,3,c32))
									svgmem(s)\visible\h = ValF(StringField(aval,4,c32))
								Case "width"
									svgmem(s)\setw = ValF(aval)
								Case "height"
									svgmem(s)\seth = ValF(aval)
								Case "version", "baseprofile" ; ignore these attributes
								Default
									If StartsWith(lanam,"xmlns")
										SVG_AddNameSpace(anam,aval)
										Continue
									EndIf
									svgmem(s)\attribute(anam) = aval
							EndSelect
						Wend
					EndIf
					If Not svgmem(s)\setw
						svgmem(s)\setw = svgmem(s)\visible\w
					EndIf
					If Not svgmem(s)\seth
						svgmem(s)\seth = svgmem(s)\visible\h
					EndIf
					ParseSVGXML(s,*nd,0,#SVG_Cycle_Definitions)
					ParseSVGXML(s,*nd,0,#SVG_Cycle_Elements)
					FreeXML(x)
					
					SVGInternal_ForEachElement(s,e)
					Select svgmem(s)\element(e)\type
						Case #SVG_Element_Line, #SVG_Element_Polyline
							svgmem(s)\element(e)\fillmode = #SVG_FillMode_Color
							svgmem(s)\element(e)\fillvalue = svgmem(s)\element(e)\strokecolor
					EndSelect
				Next e
				
				svgdata\s = 0
				ProcedureReturn s
			Else
				FreeXML(x)
				svgdata\s = 0
				ProcedureReturn 0
			EndIf
			
		EndProcedure
		
		Procedure.i LoadSVG(s.i,filename.s)
			
			If svgdata\s
				R("Can't load SVG: an SVG drawing block is open.")
				ProcedureReturn #False
			EndIf
			
			If Not FileExists(filename)
				R("NO FILE")
				ProcedureReturn #False
			EndIf
			
			If s = #PB_Any
				s = FirstAvailableSVGSlot()
			EndIf
			If Not ValidSVGNumber(s)
				ProcedureReturn 0
			EndIf
			
			x = LoadXML(#PB_Any,filename)
			;R("X: "+Str(x))
			If Not x
				ProcedureReturn #False
			EndIf
			
			ProcedureReturn SVGFromXML(s,x)
			
		EndProcedure
		
		
		Procedure.i CatchSVG(s.i,*mem)
			
			If Not *mem
				R("POINTER INVALID")
				ProcedureReturn 0
			EndIf
			If Not MemorySize(*mem)
				R("MEMORY TOO SMALL.")
				ProcedureReturn 0
			EndIf
			
			If s = #PB_Any
				s = FirstAvailableSVGSlot()
			EndIf
			If Not ValidSVGNumber(s)
				ProcedureReturn 0
			EndIf
			
			x = CatchXML(#PB_Any,*mem,MemorySize(*mem))
			;R("X: "+Str(x))
			If Not x
				ProcedureReturn #False
			EndIf
			
			ProcedureReturn SVGFromXML(s,x)
			
		EndProcedure
		
		Procedure.i LoadSVGFromString(s.i,t.s)
			
			If Nots t
				R("POINTER INVALID")
				ProcedureReturn 0
			EndIf
			
			If s = #PB_Any
				s = FirstAvailableSVGSlot()
			EndIf
			If Not ValidSVGNumber(s)
				ProcedureReturn 0
			EndIf
			
			x = CatchXML(#PB_Any,@t,StringByteLength(t))
			;R("X: "+Str(x))
			If Not x
				ProcedureReturn #False
			EndIf
			
			ProcedureReturn SVGFromXML(s,x)
			
		EndProcedure
		
		
		
		
		
		
		
		
		;-
		;- ELEMENT CREATING
		
		
		Procedure.b SVG_IDIsUsed(id.s,s.i=0)
			If Not s
				s = svgdata\s
			EndIf
			If s
				If svgmem(s)\elements
					For e = 1 To svgmem(s)\elements
						If svgmem(s)\element(e)\id = id
							ProcedureReturn #True
						EndIf
					Next e
				EndIf
				If svgmem(s)\gradients
					For e = 1 To svgmem(s)\gradients
						If svgmem(s)\gradient(e)\id = id
							ProcedureReturn #True
						EndIf
					Next e
				EndIf
				If svgmem(s)\patterns
					For e = 1 To svgmem(s)\patterns
						If svgmem(s)\pattern(e)\id = id
							ProcedureReturn #True
						EndIf
					Next e
				EndIf
			EndIf
			ProcedureReturn #False
		EndProcedure
		
		Procedure.b SVG_RemoveElement(e.i=0)
			
			s = svgdata\s
			If Not s
				R("SVG_RemoveElement fail: there is no current SVG.")
				ProcedureReturn #False
			EndIf
			
			If Not e
				e = svgdata\e
			EndIf
			If Not e
				R("SVG_RemoveElement fail: no element was specified.")
				ProcedureReturn #False
			EndIf
			
			ProcedureReturn SVGInternal_RemoveElement(s,e)
			
		EndProcedure
		
		
		Procedure.i SVG_CopyElement(e1.i=0,*namechanger.SVGPassString=0)
			
			s = svgdata\s
			If Not s
				R("SVG_CopyElement fail: there is no current SVG.")
				ProcedureReturn 0
			EndIf
			
			If Not e1
				e1 = svgdata\e
			EndIf
			If Not e1
				R("SVG_CopyElement fail: no element was specified, and there is no current element.")
				ProcedureReturn 0
			EndIf
			
			;If svgmem(s)\element(e1)\type = #SVG_Element_Group
			;	e2 = SVGInternal_CopyGroupElement(s,e1,*namechanger)
			;Else
				e2 = SVGInternal_CopyElement(s,e1,*namechanger)
			;EndIf
			
			svgdata\e = e2
			ProcedureReturn e2
			
		EndProcedure
		
		
		Procedure.i SVG_Plot(x.f,y.f,surround.f=0,color=-1,id.s="")
			w.f=1
			h.f=1
			If surround
				w+surround+surround
				h+surround+surround
				x-surround
				y-surround
			EndIf
			ProcedureReturn AddSVGElement(svgdata\s,#SVG_Element_Box,id,x,y,w,h,color)
		EndProcedure
		
		Macro SVG_Box(x,y,w,h,color=-1,id="")
			AddSVGElement(svgdata\s,#SVG_Element_Box,id,x,y,w,h,color)
		EndMacro
		
		Macro SVG_Circle(cx,cy,r,color=-1,id="")
			AddSVGElement(svgdata\s,#SVG_Element_Circle,id,cx,cy,r,r,color)
		EndMacro
		
		Macro SVG_Ellipse(cx,cy,rx,ry,color=-1,id="")
			AddSVGElement(svgdata\s,#SVG_Element_Ellipse,id,cx,cy,rx,ry,color)
		EndMacro
		
		Macro SVG_Text(x,y,text,color=-1,id="")
			AddSVGElement(svgdata\s,#SVG_Element_Text,id,x,y,0,0,color,text)
		EndMacro
		
		Macro SVG_MultilineText(x,y,w,h,text,color=-1,id="")
			AddSVGElement(svgdata\s,#SVG_Element_WrappedText,id,x,y,w,h,color,text)
		EndMacro
		
		Macro SVG_Line(x,y,w,h,thickness=1,color=-1,id="")
			AddSVGElement(svgdata\s,#SVG_Element_Line,id,x,y,w,h,color,"",0,0,"",thickness)
		EndMacro
		
		Macro SVG_LineXY(x1,y1,x2,y2,thickness=1,color=-1,id="")
			AddSVGElement(svgdata\s,#SVG_Element_Line,id,x1,y1,x2-x1,y2-y1,color,"",0,0,"",thickness)
		EndMacro
		
		Procedure.i SVG_Polyline(pointarr.s,thickness.f,color=-1,id.s="")
			
			s = svgdata\s
			
			e = AddSVGElement(s,#SVG_Element_Polyline,id,0,0,0,0,color,"",0,0,"",thickness)
			
			If pointarr
				SVGInternal_PolylineDataFromString(s,e,pointarr,#True)
			EndIf
			
			CalculateSVGElementBounds(s,e)
			ProcedureReturn e
			
		EndProcedure
		
		Procedure.i SVG_Polygon(pointarr.s,color=-1,id.s="")
			
			s = svgdata\s
			
			e = AddSVGElement(s,#SVG_Element_Polygon,id,0,0,0,0,color,"")
			
			With svgmem(s)\element(e)
				pointarr = ReplaceString(Trim(pointarr),",",c32)
				EnsureThisEnd(pointarr,c32)
				spaces = CountString(pointarr,c32)
				Dim \path\pathpoint(spaces)
				For j = 1 To spaces
					xs.s = StringField(pointarr,1,c32)
					ys.s = StringField(pointarr,2,c32)
					If Nots xs Or Nots ys : Break : EndIf
					EnsureThisNotStart(pointarr,xs+c32+ys+c32)
					ps+1
					\path\pathpoint(ps)\mode = #SVG_Path_Straight
					\path\pathpoint(ps)\x = ValF(xs)
					\path\pathpoint(ps)\y = ValF(ys)
					;CUNTarrsize=ArraySize(\pathpoint(),1) : If arrsize<items : ReDim \pathpoint(arrsize+10) : EndIf
					;If ps=#MaxSVGPathPoints : Break : EndIf
				Next j
				\path\pathpoints = ps
				;R("PATH POINTS: "+Str(\pathpoints))
			EndWith
			
			CalculateSVGElementBounds(s,e)
			ProcedureReturn e
			
		EndProcedure
		
		Macro SVG_Path(color=-1,id="",validsvgpointarr="")
			AddSVGElement(svgdata\s,#SVG_Element_Path,id,0,0,0,0,color,validsvgpointarr)
		EndMacro
		
		Macro SVG_Image(x,y,w,h,relativefilename,id="")
			AddSVGElement(svgdata\s,#SVG_Element_Image,id,x,y,w,h,-1,relativefilename)
		EndMacro
		
		Procedure.i SVG_PBImage(x,y,w,h,img.i,id.s="",relativefilename.s="",codec=#PB_ImagePlugin_PNG)
			If IsImage(img)
				ProcedureReturn AddSVGElement(svgdata\s,#SVG_Element_PBImage,id,x,y,w,h,-1,"",img,codec,relativefilename)
			EndIf
			ProcedureReturn 0
		EndProcedure
		
		Procedure.i SVG_Base64Image(x,y,w,h,base64data.s,id.s="",codec.i=#PB_ImagePlugin_PNG)
			If w<1 Or h<1 Or Nots base64data : ProcedureReturn 0 : EndIf
			If codec<>#PB_ImagePlugin_JPEG And codec<>#PB_ImagePlugin_PNG : ProcedureReturn 0 : EndIf
			ProcedureReturn AddSVGElement(svgdata\s,#SVG_Element_Base64Image,id,x,y,w,h,-1,base64data,-1,codec)
		EndProcedure
		
		
		Macro SVG_InsertCustomCode(code)
			AddSVGElement(svgdata\s,#SVG_Element_CustomCode,"",0,0,0,0,0,code,0)
		EndMacro
		
		
		Macro SVG_Group(id="")
			AddSVGElement(svgdata\s,#SVG_Element_Group,id,0,0,-1,-1)
			svgdata\grouplevel+1
		EndMacro
		
		Macro SVG_CloseGroup()
			If svgdata\s
				If svgdata\grouplevel>0
					AddSVGElement(svgdata\s,#SVG_Element_GroupCloser,"",0,0,0,0)
					svgdata\grouplevel-1
					SVGInternal_ForEachElement(s,svg_ge)
					If svgmem(s)\element(svg_ge)\type=#SVG_Element_Group
						svgdata\e = svg_ge
					EndIf
				Next svg_ge
			EndIf
		EndIf
	EndMacro
	
	Macro SVG_CloseAllGroups
		While svgdata\grouplevel
			SVG_CloseGroup
		Wend
	EndMacro
	
	
	
	
	;- COMMON POLYGONS
	
	Procedure.i SVG_RegularPolygon(cx.f,cy.f,r.f,sides.i,clr.i=0,id.s="")
		
		If sides<3 : ProcedureReturn 0 : EndIf
		
		e = SVG_Path(clr,id)
		
		base.PointF
		base\x = cx
		base\y = cy
		
		inc.f = 360/sides
		angle.f = -90
		
		DegreeCoordsFromPoint(@base,angle,r,@np.PointF)
		SVG_Path_StartSegment(np\x,np\y)
		
		For a = 2 To sides
			angle+inc
			DegreeCoordsFromPoint(@base,angle,r,@np.PointF)
			SVG_Path_Line(np\x,np\y)
		Next a
		
		;R(StrF(angle+inc))
		
		SVG_Path_EndSegment()
		
		ProcedureReturn e
		
	EndProcedure
	
	Macro SVG_Triangle(cx,cy,r,clr=-1,id="")
		SVG_RegularPolygon(cx,cy,r,3,clr,id)
	EndMacro
	
	Macro SVG_Pentagon(cx,cy,r,clr=-1,id="")
		SVG_RegularPolygon(cx,cy,r,5,clr,id)
	EndMacro
	
	Macro SVG_Hexagon(cx,cy,r,clr=-1,id="")
		SVG_RegularPolygon(cx,cy,r,6,clr,id)
	EndMacro
	
	Macro SVG_Septagon(cx,cy,r,clr=-1,id="")
		SVG_RegularPolygon(cx,cy,r,7,clr,id)
	EndMacro
	
	Macro SVG_Octagon(cx,cy,r,clr=-1,id="")
		SVG_RegularPolygon(cx,cy,r,8,clr,id)
	EndMacro
	
	Macro SVG_Nonagon(cx,cy,r,clr=-1,id="")
		SVG_RegularPolygon(cx,cy,r,9,clr,id)
	EndMacro
	
	Macro SVG_Decagon(cx,cy,r,clr=-1,id="")
		SVG_RegularPolygon(cx,cy,r,10,clr,id)
	EndMacro
	
	
	
	
	Procedure.i SVG_CircleSegmentV1(cx.f,cy.f,radius.f,radians.f,startangle.f=0,color.i=-1,id.s="")
		
		Protected.f x, y
		
		e = SVG_Path(color,id)
		
		; move to centre-point
		SVG_Path_StartSegment(cx,cy)
		
		x = (Cos(startangle)*radius)+cx
		y = (Sin(startangle)*radius)+cy
		SVG_Path_Line(x,y) ; draw outwards from centre to startangle
		
		large=#False
		sweep=#True
		Debug radians
		If radians=>#PI
			large=#True
			;sweep=#False
			;R("")
		EndIf
		radians = startangle+radians
		x = (Cos(radians)*radius)+cx
		y = (Sin(radians)*radius)+cy
		
		SVG_Path_EllipticalArc(x,y,radius,radius,0,large,sweep) ; curve from startangle to endangle
		;SVG_Path_Line(x,y)
		
		; draw inwards from endangle to centre
		;SVG_Path_Line(cx,cy)
		SVG_Path_EndSegment()
		
		ProcedureReturn e
		
	EndProcedure
	
	Procedure.i SVG_CircleSegment(cx.f,cy.f,radius.f,radians.f,startangle.f=0,color.i=-1,id.s="")
		
		Protected.f x, y
		Protected.b large, sweep
		
		e = SVG_Path(color,id)
		
		; move to centre-point
		SVG_Path_StartSegment(cx,cy)
		
		x = (Cos(startangle)*radius)+cx
		y = (Sin(startangle)*radius)+cy
		SVG_Path_Line(x,y) ; draw outwards from centre to startangle
		
		killer_rad.f = Radian(119)
		;R("RADIANS: "+StrF(radians)+c13+" KILLER: "+StrF(killer_rad))
		If radians=>killer_rad
			inc.f = killer_rad/2
			thisradians.f = startangle
			thisx.f
			thisy.f
			Repeat
				
				thisradians+inc
				thisx = (Cos(thisradians)*radius)+cx
				thisy = (Sin(thisradians)*radius)+cy
				SVG_Path_EllipticalArc(thisx,thisy,radius,radius,0,0,1) ; curve from startangle to endangle
				
				If (thisradians+inc)=>radians : Break : EndIf
			ForEver
			; now the remainder...
			finalrad.f = (startangle+radians)-thisradians
			;R("RADIANS: "+StrF(startangle+radians)+c13+"DRAWNRADIANS: "+StrF(thisradians)+c13+"DIF (FINALRAD): "+StrF(finalrad))
			If finalrad>0
				radians+startangle
				thisx = (Cos(radians)*radius)+cx
				thisy = (Sin(radians)*radius)+cy
				;SVG_Path_Line(thisx,thisy)
				SVG_Path_EllipticalArc(thisx,thisy,radius,radius,0,0,1)
			EndIf
		Else
			large=#False
			sweep=#True
			;Debug radians
			If radians=>#PI
				large=#True
				sweep=#False
				;R("")
			EndIf
			radians = startangle+radians
			x = (Cos(radians)*radius)+cx
			y = (Sin(radians)*radius)+cy
			SVG_Path_EllipticalArc(x,y,radius,radius,0,large,sweep) ; curve from startangle to endangle
		EndIf
		
		; draw inwards from endangle to centre
		;SVG_Path_Line(cx,cy)
		SVG_Path_EndSegment()
		
		ProcedureReturn e
		
	EndProcedure
	
	Procedure.i SVG_SharpCircleSegment(cx.f,cy.f,radius.f,radians.f,startangle.f=0,color.i=-1,id.s="")
		
		Protected.f x, y
		Protected.b large, sweep
		
		e = SVG_Path(color,id)
		
		; move to centre-point
		SVG_Path_StartSegment(cx,cy)
		
		x = (Cos(startangle)*radius)+cx
		y = (Sin(startangle)*radius)+cy
		SVG_Path_Line(x,y) ; draw outwards from centre to startangle
		
		radians = startangle+radians
		x = (Cos(radians)*radius)+cx
		y = (Sin(radians)*radius)+cy
		SVG_Path_Line(x,y) ; curve from startangle to endangle
		
		; draw inwards from endangle to centre
		;SVG_Path_Line(cx,cy)
		SVG_Path_EndSegment()
		
		ProcedureReturn e
		
	EndProcedure
	
	
	
	
	
	Enumeration -999999 ; directions
		#Up
		#UpRight
		#Right
		#DownRight
		#Down
		#DownLeft
		#Left
		#UpLeft
	EndEnumeration
	
	Macro ArrowDeg(a,deg)
		deg+a
	EndMacro
	
	Procedure.i SVG_Arrow(px.f,py.f,deg.f,color=-1,id.s="")
		
		;{
		Select deg
			Case #Up
				deg = 180
			Case #UpRight
				deg = 225
			Case #Right
				deg = 270
			Case #DownRight
				deg = 315
			Case #Down
				deg = 0
			Case #DownLeft
				deg = 45
			Case #Left
				deg = 90
			Case #UpLeft
				deg = 135
		EndSelect
		
		deg-90
		
		sw.f = svgdata\arrow_stalkw
		sl.f = svgdata\arrow_stalkl
		hw.f = svgdata\arrow_headw
		hl.f = svgdata\arrow_headl
		
		base.PointF : base\x = px : base\y = py
		
		e = SVG_Path(color,id)
		
		cc.PointF
		totallength.f = sl+hl
		
		arrowdiag.f = Sqr((hw*hw)+(hl*hl))
		
		stalkhalfw.f = sw/2
		
		extraheadwidth.f = (hw/2)-(sw/2)
		
		SVG_Path_StartSegment(base\x,base\y) ; point
		;}
		;Mark(@base,#White)
		
		; from point to north head flank
		DegreeCoordsFromPoint(@base,ArrowDeg(0,deg),hl,@stalkstartcentre.PointF)
		DegreeCoordsFromPoint(@stalkstartcentre,ArrowDeg(90,deg),hw/2,@cc)
		SVG_Path_Line(cc\x,cc\y)
		;Mark(@cc,#Blue)
		
		; from north head flank to stalk start north
		DegreeCoordsFromPoint(@cc,ArrowDeg(-90,deg),extraheadwidth,@cc)
		SVG_Path_Line(cc\x,cc\y)
		;Mark(@cc,#Purple)
		
		; from stalk start north to stalk end north
		DegreeCoordsFromPoint(@cc,ArrowDeg(0,deg),sl,@cc)
		SVG_Path_Line(cc\x,cc\y)
		;Mark(@cc,#Red)
		
		; from stalk end north to stalk end south
		DegreeCoordsFromPoint(@cc,ArrowDeg(-90,deg),sw,@cc)
		SVG_Path_Line(cc\x,cc\y)
		;Mark(@cc,#Orange)
		
		; from stalk end south to stalk start south
		DegreeCoordsFromPoint(@stalkstartcentre,ArrowDeg(-90,deg),sw/2,@cc)
		SVG_Path_Line(cc\x,cc\y)
		;Mark(@cc,#Yellow)
		
		; from stalk start south to south head flank
		DegreeCoordsFromPoint(@cc,ArrowDeg(-90,deg),extraheadwidth,@cc)
		SVG_Path_Line(cc\x,cc\y)
		;Mark(@cc,#Green)
		
		SVG_Path_EndSegment() ; back to point
		
		ProcedureReturn e
		
	EndProcedure
	
	
	
	
	Procedure.i SVG_RectangularFrame(pad.f=0,thickness.f=1,color.i=-1,id.s="")
		
		s = svgdata\s
		e = svgdata\e
		If Not s Or Not e : ProcedureReturn #False : EndIf
		
		GetSVGElementBoundingBox(s,e,@bb.RectF,#True)
		
		With bb
			
			If pad>0
				\x-pad
				\y-pad
				\w+pad+pad
				\h+pad+pad
			EndIf
			
			pd.s
			pd + StrF(\x)+","+StrF(\y)+" "
			pd + StrF(\x+\w)+","+StrF(\y)+" "
			pd + StrF(\x+\w)+","+StrF(\y+\h)+" "
			pd + StrF(\x)+","+StrF(\y+\h)+" "
			pd + StrF(\x)+","+StrF(\y)+" "
		EndWith
		
		ProcedureReturn SVG_Polyline(pd,thickness,color,id)
		
	EndProcedure
	
	Procedure.i SVG_SquareFrame(pad.f=0,thickness.f=1,color.i=-1,id.s="")
		
		s = svgdata\s
		e = svgdata\e
		If Not s Or Not e : ProcedureReturn #False : EndIf
		
		GetSVGElementBoundingBox(s,e,@bb.RectF,#True)
		
		With bb
			
			If \w > \h
				dif.f = (\w-\h)/2
				\y-dif
				\h=\w
			Else
				If \h > \w
					dif.f = (\h-\w)/2
					\x-dif
					\w=\h
				EndIf
			EndIf
			
			
			If pad>0
				\x-pad
				\y-pad
				\w+pad+pad
				\h+pad+pad
			EndIf
			
			pd.s
			pd + StrF(\x)+","+StrF(\y)+" "
			pd + StrF(\x+\w)+","+StrF(\y)+" "
			pd + StrF(\x+\w)+","+StrF(\y+\h)+" "
			pd + StrF(\x)+","+StrF(\y+\h)+" "
			pd + StrF(\x)+","+StrF(\y)+" "
		EndWith
		
		ProcedureReturn SVG_Polyline(pd,thickness,color,id)
		
	EndProcedure
	
	Procedure.i SVG_CircularFrame(pad.f=0,thickness.f=1,color.i=-1,id.s="")
		
		s = svgdata\s
		e = svgdata\e
		If Not s Or Not e : ProcedureReturn #False : EndIf
		
		GetSVGElementBoundingBox(s,e,@bb.RectF,#True)
		
		With bb
			
			cx.f = \x+(\w/2)
			cy.f = \y+(\h/2)
			c.PointF : c\x = cx : c\y = cy
			
			
			w.f = \w/2
			h.f = \h/2
			corner.PointF
			corner\x = \x+\w
			corner\y = \y+\h
			d.f = DistanceBetweenTwoPoints(@c,@corner)
			
			
			If pad>0
				d+pad+pad
			EndIf
			
			
			ne = SVG_Circle(cx,cy,d,-1,id)
			svgmem(s)\element(ne)\fillmode = #SVG_FillMode_None
			svgmem(s)\element(ne)\strokecolor = color
			svgmem(s)\element(ne)\strokewidth = thickness
			ProcedureReturn ne
			
		EndWith
		
	EndProcedure
	
	Procedure.i SVG_EllipticalFrame(pad.f=0,thickness.f=1,color.i=-1,id.s="")
		
		s = svgdata\s
		e = svgdata\e
		If Not s Or Not e : ProcedureReturn #False : EndIf
		
		GetSVGElementBoundingBox(s,e,@bb.RectF,#True)
		
		With bb
			
			cx.f = \x+(\w/2)
			cy.f = \y+(\h/2)
			dx.f = \w*Sqr(2)/2
			dy.f = \h*Sqr(2)/2
			
			If pad>0
				dx+pad+pad
				dy+pad+pad
			EndIf
			
			ne = SVG_Ellipse(cx,cy,dx,dy,-1,id)
			svgmem(s)\element(ne)\fillmode = #SVG_FillMode_None
			svgmem(s)\element(ne)\strokecolor = color
			svgmem(s)\element(ne)\strokewidth = thickness
			ProcedureReturn ne
			
		EndWith
		
	EndProcedure
	
	
	Procedure.b PolyRect(*r.RectF,clr=#Green)
		
		pd.s
		pd + StrF(*r\x)+","+StrF(*r\y)+" "
		pd + StrF(*r\x+*r\w)+","+StrF(*r\y)+" "
		pd + StrF(*r\x+*r\w)+","+StrF(*r\y+*r\h)+" "
		pd + StrF(*r\x)+","+StrF(*r\y+*r\h)+" "
		pd + StrF(*r\x)+","+StrF(*r\y)+" "
		
		ProcedureReturn SVG_Polyline(pd,1,clr,"polyrect")
		
	EndProcedure
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	;-
	;- IMPORTING
	
	Procedure.i SVG_ElementFromString(t.s)
		s = svgdata\s
		If s And t
			t = RemoveString(t,c13)
			t = RemoveString(t,c10)
			t = Trim(t)
			If Nots t : ProcedureReturn #False : EndIf
			If Left(t,4)="<!--"
				t = RemoveString(t,"<!--")
				t = RemoveString(t,"-->")
				ProcedureReturn AddSVGElement(svgdata\s,#SVG_Element_Comment,"",0,0,0,0,-1,Trim(t))
			EndIf
			e = svgmem(s)\elements
			x = CatchXML(#PB_Any,@t,StringByteLength(t))
			*nd = MainXMLNode(x)
			ParseSVGXML(s,*nd,0,#SVG_Cycle_Elements)
			FreeXML(x)
			If svgmem(s)\elements>e
				ne = svgmem(s)\elements
				If svgmem(s)\element(ne)\type=#SVG_Element_GroupCloser
					; find opener for closer
					e = OpenerForSVGGroup(s,ne)
					svgdata\e = e
					ProcedureReturn e
				Else
					svgdata\e = e+1
					ProcedureReturn e+1
				EndIf
			Else
				ProcedureReturn #False
			EndIf
		EndIf
	EndProcedure
	
	
	Procedure.b SVGInternal_GradientFromString(s.i,t.s)
		Protected.f x,y,w,h
		
		;*mem = @t
		g = svgmem(s)\gradients
		x = CatchXML(#PB_Any,@t,StringByteLength(t))
		*nd = MainXMLNode(x)
		ParseSVGXML(s,*nd,0,#SVG_Cycle_Definitions)
		
		If svgmem(s)\gradients>g
			svgdata\g = g+1
			ProcedureReturn g+1
		Else
			ProcedureReturn #False
		EndIf
		
	EndProcedure
	
	Procedure.i SVG_GradientFromString(t.s)
		
		s = svgdata\s
		If s
			;If svgmem(s)\gradients<#MaxSVGGradients
			gr = SVGInternal_GradientFromString(s,t)
			ProcedureReturn gr
			;EndIf
		EndIf
		
	EndProcedure
	
	
	
	
	
	
	
	
	
	
	;-
	;- ELEMENT EDITING
	
	Macro EditElementCheck
		s = svgdata\s
		e = svgdata\e
		If Not s Or Not e : ProcedureReturn : EndIf
		If e>svgmem(s)\elements Or Not svgmem(s)\element(e)\in_use : ProcedureReturn : EndIf
	EndMacro
	
	Procedure.i SVG_FindElement(id.s)
		
		If Not svgdata\s Or Nots id : ProcedureReturn 0 : EndIf
		s = svgdata\s
		
		id = LCase(id)
		With svgmem(s)
			If \elements
				For e = 1 To \elements
					If \element(e)\id = id
						ProcedureReturn e
					EndIf
				Next e
			EndIf
		EndWith
		
		ProcedureReturn 0
		
	EndProcedure
	
	Procedure.b SVG_SelectElement(e.i)
		
		s = svgdata\s
		If s>0 And s<=#MaxSVGs
			If e>0 And e<=svgmem(s)\elements
				If svgmem(s)\element(e)\in_use
					svgdata\e = e
					ProcedureReturn #True
				EndIf
			EndIf
		EndIf
		
		ProcedureReturn #False
		
	EndProcedure
	
	Procedure.b SVG_SetElementFill(mode,value=0)
		EditElementCheck
		svgmem(s)\element(e)\fillmode = mode
		svgmem(s)\element(e)\fillvalue = value
	EndProcedure
	
	Procedure.b SVG_SetElementID(newid.s)
		EditElementCheck
		svgmem(s)\element(e)\id = newid
	EndProcedure
	
	Procedure.b SVG_SetElementAction(actiontype.s,whattodo.s,add.b=#False)
		
		EditElementCheck
		If add
			svgmem(s)\element(e)\action(actiontype) + whattodo
		Else
			svgmem(s)\element(e)\action(actiontype) = whattodo
		EndIf
		
	EndProcedure
	
	Procedure.b SVG_SetElementStroke(clr=-1,wdth=-1,opc.f=-1,lj=-1,lc=-1)
		
		EditElementCheck
		If clr>-1
			svgmem(s)\element(e)\strokecolor = clr
		EndIf
		If wdth>-1
			svgmem(s)\element(e)\strokewidth = wdth
		EndIf
		If opc>-1
			svgmem(s)\element(e)\strokeopacity = opc
		EndIf
		If lj>-1
			svgmem(s)\element(e)\strokelinejoin = lj
		EndIf
		If lc>-1
			svgmem(s)\element(e)\strokelinecap = lc
		EndIf
		
	EndProcedure
	
	
	Procedure.b SVG_SetElementAttribute(attrname.s,attrvalue.s)
		EditElementCheck
		svgmem(s)\element(e)\attribute(attrname) = attrvalue
	EndProcedure
	
	Procedure.b SVG_RemoveElementAttribute(attrname.s)
		EditElementCheck
		DeleteMapElement(svgmem(s)\element(e)\attribute(),attrname)
	EndProcedure
	
	Macro SVGInternal_StripInkScapeAttributes(attrmp)
		ClearMap(del())
		ForEach attrmp
			key.s = MapKey(attrmp)
			If StartsWith(key,"inkscape") Or StartsWith(key,"sodipodi")
				AddMapElement(del(),key)
			EndIf
		Next
		ForEach del()
			DeleteMapElement(attrmp,MapKey(del()))
		Next
	EndMacro
	
	Procedure.b SVG_StripInkScapeData()
		s = svgdata\s
		If s
			NewMap del.b()
			With svgmem(s)
				SVGInternal_StripInkScapeAttributes(\attribute())
				ForEachSVGElement(e)
					SVGInternal_StripInkScapeAttributes(\element(e)\attribute())
				Next
				
				; prune namespaces..
				ClearMap(del())
				ForEach \namespace()
					key.s = MapKey(\namespace())
					If FindString(key,"inkscape") Or FindString(key,"sodipodi")
						AddMapElement(del(),key)
					EndIf
				Next
				ForEach del()
					DeleteMapElement(\namespace(),MapKey(del()))
				Next
			EndWith
		EndIf
	EndProcedure
	
	
	Procedure.b SVG_SetElementClass(info.s)
		EditElementCheck
		svgmem(s)\element(e)\class = info
	EndProcedure
	
	Procedure.b SVG_SetElementFont(fname.s,fsize=0)
		
		EditElementCheck
		If SVGElementTypeIsFontable(svgmem(s)\element(e)\type)
			If fname<>"" And fname<>svgmem(s)\element(e)\fontname
				svgmem(s)\element(e)\fontname = fname
				svgmem(s)\element(e)\bounds_known = #False
			EndIf
			If fsize<>0 And fsize<>svgmem(s)\element(e)\fontsize
				svgmem(s)\element(e)\fontsize = fsize
				svgmem(s)\element(e)\bounds_known = #False
			EndIf
		EndIf
	EndProcedure
	
	Procedure.b SVG_SetElementCursor(c.i,url.s="")
		
		EditElementCheck
		svgmem(s)\element(e)\cursor = c
		svgmem(s)\element(e)\cursor_url = url
		
	EndProcedure
	
	
	Enumeration 0
		#SVG_Opacity_Both
		#SVG_Opacity_Fill
		#SVG_Opacity_Stroke
	EndEnumeration
	
	Procedure.b SVG_SetElementOpacity(newopc.f,ofcurrent.b=#False,mode.b=#SVG_Opacity_Both)
		
		EditElementCheck
		
		current.f
		opc.f
		
		If svgmem(s)\element(e)\type = #SVG_Element_Group
			SVGInternal_ForEachElementInSVGGroup(s,e,a)
			Select mode
				Case #SVG_Opacity_Both
					If ofcurrent
						opc = Beat(0,Defeat(100,svgmem(s)\element(a)\opacity/100*newopc))
					Else
						opc = newopc
					EndIf
					svgmem(s)\element(a)\opacity = opc
					svgmem(s)\element(a)\strokeopacity = opc
				Case #SVG_Opacity_Fill
					If ofcurrent
						opc = Beat(0,Defeat(100,svgmem(s)\element(a)\opacity/100*newopc))
					Else
						opc = newopc
					EndIf
					svgmem(s)\element(a)\opacity = opc
				Case #SVG_Opacity_Stroke
					If ofcurrent
						opc = Beat(0,Defeat(100,svgmem(s)\element(a)\strokeopacity/100*newopc))
					Else
						opc = newopc
					EndIf
					svgmem(s)\element(a)\strokeopacity = opc
			EndSelect
		Next a
	Else
		Select mode
			Case #SVG_Opacity_Both
				If ofcurrent
					opc = Beat(0,Defeat(100,svgmem(s)\element(e)\opacity/100*newopc))
				Else
					opc = newopc
				EndIf
				svgmem(s)\element(e)\opacity = opc
				svgmem(s)\element(e)\strokeopacity = opc
			Case #SVG_Opacity_Fill
				If ofcurrent
					opc = Beat(0,Defeat(100,svgmem(s)\element(e)\opacity/100*newopc))
				Else
					opc = newopc
				EndIf
				svgmem(s)\element(e)\opacity = opc
			Case #SVG_Opacity_Stroke
				If ofcurrent
					opc = Beat(0,Defeat(100,svgmem(s)\element(e)\strokeopacity/100*newopc))
				Else
					opc = newopc
				EndIf
				svgmem(s)\element(e)\strokeopacity = opc
		EndSelect
	EndIf
	
EndProcedure

Procedure.b SVG_SetElementTooltip(t.s)
	EditElementCheck
	svgmem(s)\element(e)\tooltip = t
EndProcedure

Procedure.b SVG_SetElementVisibility(viz)
	EditElementCheck
	svgmem(s)\element(e)\visibility = viz
EndProcedure

Procedure.b SVG_SetElementHREF(u.s,target.b=#SVG_Target_Top)
	EditElementCheck
	svgmem(s)\element(e)\href = u
	svgmem(s)\element(e)\href_target = target
EndProcedure

Procedure.b SVG_SetElementHREFTarget(target.b)
	EditElementCheck
	svgmem(s)\element(e)\href_target = target
EndProcedure

Procedure.b SVG_SetElementText(t.s)
	EditElementCheck
	With svgmem(s)\element(e)
		Select \type
			Case #SVG_Element_Text, #SVG_Element_WrappedText, #SVG_Element_TextPath
				\content = t
				\bounds_known = #False
			Case #SVG_Element_Path
				\type = #SVG_Element_TextPath
				\content = t
				\bounds_known = #False
			Case #SVG_Element_Comment
				\content = t
			Default
				ProcedureReturn #False
		EndSelect
		ProcedureReturn #True
	EndWith
EndProcedure

Procedure.b SVG_SetElementImage(img.i)
	If Not IsImage(img) : ProcedureReturn #False : EndIf
	EditElementCheck
	With svgmem(s)\element(e)
		If \type = #SVG_Element_PBImage
			\imagenumber = img
			\w = ImageWidth(img)
			\h = ImageHeight(img)
		EndIf
	EndWith
EndProcedure

Procedure.b SVG_SetElementMask(m.i)
	EditElementCheck
	With svgmem(s)\element(e)
		If m<1
			\mask = 0
			ProcedureReturn #True
		EndIf
		If m<=svgmem(s)\elements
			If svgmem(s)\element(m)\ismask
				\mask = m
				ProcedureReturn #True
			EndIf
		EndIf
		ProcedureReturn #False
	EndWith
EndProcedure

Procedure.b SVG_SetElementFilter(f.i)
	EditElementCheck
	With svgmem(s)\element(e)
		If f<1
			\filter = 0
			ProcedureReturn #True
		EndIf
		\filter = f
		ProcedureReturn #True
	EndWith
EndProcedure


Procedure.b SVG_ShiftElement(xshift.f,yshift.f)
	
	EditElementCheck
	
	If xshift=0 And yshift=0
		ProcedureReturn #True
	EndIf
	
	Select svgmem(s)\element(e)\type
		Case #SVG_Element_GroupCloser
			ProcedureReturn #True
		Case #SVG_Element_Group
			ProcedureReturn SVGInternal_ShiftGroupElement(s,e,xshift,yshift)
		Default
			ProcedureReturn SVGInternal_ShiftElement(s,e,xshift,yshift)
	EndSelect
	
EndProcedure


Procedure.b SVG_SetElementPosition(x.f,y.f)
	
	;If x<0 Or y<0
	;    R("SVG_SetElementPosition fail: destination coordinates not valid.")
	;    ProcedureReturn #False
	;EndIf
	
	EditElementCheck
	
	ProcedureReturn SVGInternal_SetElementPosition(s,e,x,y)
	
EndProcedure




Procedure.b SVG_GetElementBoundingBox(b,includestroke.b=#False)
	If svgdata\s And svgdata\e
		GetSVGElementBoundingBox(svgdata\s,svgdata\e,b,includestroke)
	EndIf
EndProcedure

Macro SVG_GetElementCenter(c)
	If svgdata\s And svgdata\e
		GetElementCenter(svgdata\s,svgdata\e,c)
	EndIf
EndMacro

Procedure.b SVG_CentraliseElementTo(cx.f,cy.f)
	
	EditElementCheck
	
	d.PointF
	d\x = cx
	d\y = cy
	
	KnowBounds(s,e)
	GetSVGElementCenter(s,e,@c.PointF)
	
	xshift.f
	Select d\x
		Case c\x
			xshift = 0
		Case #PB_Ignore
			xshift = 0
		Default
			xshift = Difference(c\x,d\x)
			If d\x < c\x ; we're moving it leftwards
				xshift = 0-xshift
			EndIf
	EndSelect
	
	yshift.f
	Select d\y
		Case c\y
			yshift = 0
		Case #PB_Ignore
			yshift = 0
		Default
			yshift = Difference(c\y,d\y)
			If d\y < c\y ; we're moving it upwards
				yshift = 0-yshift
			EndIf
	EndSelect
	
	;R("SHIFT: "+StrF(xshift,0)+","+StrF(yshift,0))
	
	SVGInternal_ShiftElement(s,e,xshift,yshift)
	
EndProcedure


Procedure.b SVG_SetElementDimensions(w.f,h.f,maintain.b=#SVG_Orient_Center)
	
	EditElementCheck
	
	ProcedureReturn SVGInternal_SetElementDimensions(s,e,w,h,maintain)
	
EndProcedure


Procedure.b SVG_ScaleElement(xpc.f,ypc.f,maintain.b=#SVG_Orient_Center)
	
	If xpc<=0 Or ypc<=0 : R("SCALING NOT VALID") : ProcedureReturn #False : EndIf
	EditElementCheck
	
	ProcedureReturn SVGInternal_ScaleElement(s,e,xpc,ypc,maintain)
	
EndProcedure


Procedure.b SVG_RotateElement(deg.f,cx.f=#PB_Ignore,cy.f=#PB_Ignore)
	
	s = svgdata\s
	e = svgdata\e
	If Not s Or Not e
		ProcedureReturn #False
	EndIf
	
	If Not deg : ProcedureReturn #True : EndIf
	;If svgmem(s)\element(e)\type=#SVG_Element_Group
;		ProcedureReturn SVGInternal_RotateGroupElement(s,e,deg,cx,cy)
;	Else
		ProcedureReturn SVGInternal_RotateElement(s,e,deg,cx,cy)
;	EndIf
	
EndProcedure




Procedure.b SVG_Ungroup(e.i=0)
	
	EditElementCheck
	If svgmem(s)\element(e)\type<>#SVG_Element_Group : ProcedureReturn #False : EndIf
	
	gc = CloserForSVGGroup(s,e)
	
	ClearStructure(@svgmem(s)\element(e),SVG_Element_Structure)
	If gc
		ClearStructure(@svgmem(s)\element(gc),SVG_Element_Structure)
	EndIf
	
EndProcedure

Procedure.b SVG_EmptyGroup(e.i=0)
	
	EditElementCheck
	If svgmem(s)\element(e)\type<>#SVG_Element_Group : ProcedureReturn #False : EndIf
	
	NewList del.i()
	SVGInternal_ForEachElementInSVGGroup(s,g,e)
	AddElement(del())
	del() = e
	Next
	
	ForEach del()
		SVGInternal_RemoveElement(s,del())
	Next
	
EndProcedure


Procedure.b SVG_StyleFromString(t.s)
	s = svgdata\s
	e = svgdata\e
	If s And e
		SVGInternal_StyleFromString(@svgmem(s),@svgmem(s)\element(e),t)
	EndIf
EndProcedure


Procedure.b SVG_PolylineDataFromString(pointarr.s,add.b=#False)
	
	EditElementCheck
	Select svgmem(s)\element(e)\type
		Case #SVG_Element_Polygon, #SVG_Element_Polyline
			SVGInternal_PolylineDataFromString(s,e,pointarr,add)
			ProcedureReturn #True
		Default
			ProcedureReturn #False
	EndSelect
	
EndProcedure









;-
;- EDITING ALL ELEMENTS TOGETHER


Procedure.b SVG_ShiftAllElements(xshift.f,yshift.f)
	
	s = svgdata\s
	
	ForEachSVGElement(e)
	Select svgmem(s)\element(e)\type
		Case #SVG_Element_Group, #SVG_Element_GroupCloser
			Continue
	EndSelect
	SVGInternal_ShiftElement(s,e,xshift,yshift)
Next e

EndProcedure

Procedure.b SVG_RemoveAllElements()
	
	s = svgdata\s
	
	With svgmem(s)
		If \elements
			For e = 1 To \elements
				ClearStructure(@\element(e),SVG_Element_Structure)
			Next e
		EndIf
		\elements = 0
	EndWith
	
EndProcedure

Procedure.b SVG_AlignAllElementsTo(tx.f,ty.f)
	
	s = svgdata\s
	
	lx.f = 4000000
	ly.f = 4000000
	
	;elements = svgmem(s)\elements
	;Dim bb.RectF(elements)
	
	SVGInternal_ForEachElement(s,e)
	Select svgmem(s)\element(e)\type
		Case #SVG_Element_Group, #SVG_Element_GroupCloser
			Continue
	EndSelect
	
	GetSVGElementBoundingBox(s,e,@bb.RectF)
	lx = Defeat(lx,bb\x)
	ly = Defeat(ly,bb\y)
Next e

;R("LOWEST X: "+StrF(lx)+c13+"LOWEST Y: "+StrF(ly))
xshift.f = tx-lx
yshift.f = ty-ly

;SVG_ShiftAllElements(xshift,yshift)
;ProcedureReturn

SVGInternal_ForEachElement(s,e)
Select svgmem(s)\element(e)\type
	Case #SVG_Element_Group, #SVG_Element_GroupCloser
		Continue
EndSelect
SVGInternal_ShiftElement(s,e,xshift,yshift)
Next e

EndProcedure

Procedure.b SVG_CentraliseAllElementsTo(dx.f,dy.f)
	
	s = svgdata\s
	
	MasterForAllElements
	
	xhalf.f = master\w / 2
	yhalf.f = master\h / 2
	dx-xhalf
	dy-yhalf
	;SVG_AlignAllElementsTo(dx,dy)
	SVG_ShiftAllElements(dx-master\x,dy-master\y)
	
	;ReportRectF(@master,0,"MASTER")
	;PolyRect(@master,#Red)
	
EndProcedure

Procedure.b SVG_ResizeAllElements(w.f,h.f,maintain.b=#SVG_Orient_Center)
	
	s = svgdata\s
	If Not s : ProcedureReturn #False : EndIf
	
	With svgmem(s)
		
		MasterForAllElements
		
		If w = #PB_Ignore Or w=0
			w = master\w
		EndIf
		If h = #PB_Ignore Or h=0
			h = master\h
		EndIf
		If w=master\w And h=master\h ; no resizing to be done
			;R("NO RESIZING")
			ProcedureReturn #True
		EndIf
		
		;ReportRectF(@master,0,"SVG_ResizeForAllElements (BEFORE RESIZING)"+c13+"RESIZE TO: "+StrF(w)+","+StrF(h))
		
		CalculateShifts(@master,w,h,maintain,@wpc.f,@hpc.f,@xshift.f,@yshift.f)
		;Debug wpc : Debug hpc
		;Debug xshift : Debug yshift
		SVGInternal_ForEachElement(s,a)
		Select \element(a)\type
			Case #SVG_Element_Group, #SVG_Element_GroupCloser, #SVG_Element_CustomCode, #SVG_Element_Comment
				\bounds_known=#False
				Continue
		EndSelect
		
		; get current bounds...
		;KnowBounds(s,a)
		CopyRectF(@\element(a)\bound,@bb.RectF)
		ix.f = (bb\x - master\x)
		iy.f = (bb\y - master\y)
		
		; resize...
		SVGInternal_SetElementDimensions(s,a,wpc*bb\w,hpc*bb\h,#SVG_Orient_TopLeft)
		
		; restore position relative to group (x,y)...
		ix*wpc
		iy*hpc
		SVGInternal_ShiftElement(s,a,master\x+ix+xshift-bb\x,master\y+iy+yshift-bb\y)
		
		\bounds_known = #False
	Next a
	
	\bounds_known = #False
	;ClearStructure(@master,RectF) : MasterForAllElements : ReportRectF(@master,0,"AFTER RESIZING")
	ProcedureReturn #True
EndWith
EndProcedure

Procedure.b SVG_ScaleAllElements(wpc.f,hpc.f,maintain.b=#SVG_Orient_Center)
	
	s = svgdata\s
	If Not s : ProcedureReturn #False : EndIf
	
	;CalculateTotalSVGElementBounds(s)
	MasterForAllElements
	
	w.f : h.f
	With svgmem(s)
		If wpc=#PB_Ignore Or wpc=100
			w = master\w
		Else
			w = master\w*wpc/100
		EndIf
		If hpc=#PB_Ignore Or hpc=100
			h = master\h
		Else
			h = master\h*hpc/100
		EndIf
		If w=master\w And h=master\h
			;R("NO RESIZING")
			ProcedureReturn #True
		EndIf
		
		;R("W: "+StrF(w)+c13+"H: "+StrF(h))
		SVG_ResizeAllElements(w,h,maintain)
		
	EndWith
	
EndProcedure

Procedure.b SVG_RotateAllElements(deg.f,cx.f=#PB_Ignore,cy.f=#PB_Ignore)
	
	s = svgdata\s
	If Not s
		ProcedureReturn #False
	EndIf
	
	If Not deg : ProcedureReturn #True : EndIf
	
	With svgmem(s)
		
		If \elements<1 : ProcedureReturn #True : EndIf
		
		MasterForAllElements
		If cx = #PB_Ignore
			cx = master\x+(master\w/2)
		EndIf
		If cy = #PB_Ignore
			cy = master\y+(master\h/2)
		EndIf
		
		;R("c : "+Str(cx)+","+Str(cy))
		;R("DEG: "+StrF(deg))
		\bounds_known = #False
		
		
		SVGInternal_ForEachElement(s,e)    
		Select \element(e)\type
			Case #SVG_Element_Group, #SVG_Element_GroupCloser, #SVG_Element_CustomCode, #SVG_Element_Comment
				Continue
		EndSelect
		
		;GetSVGElementBoundingBox(s,e,@eb.RectF)
		;Debug "" : DebugRectF(@eb)
		;xshift.f = master\x-master\x
		;yshift.f = master\y-master\y
		;Debug StrF(xshift,2)+","+StrF(yshift,2)
		;Debug deg
		SVGInternal_RotateElement(s,e,deg,cx,cy)
		;ShiftSVGElement(s,e,xshift,yshift)
	Next e
	
	ProcedureReturn #True
	
EndWith

EndProcedure










;-
;- COMMENTING

Macro SVG_Comment(t)
	AddSVGElement(svgdata\s,#SVG_Element_Comment,"",0,0,0,0,-1,t)
EndMacro

Procedure.b SVG_StripComments()
	
	s = svgdata\s
	SVGInternal_ForEachElement(s,e)
	If svgmem(s)\element(e)\type = #SVG_Element_Comment
		ClearStructure(@svgmem(s)\element(e),SVG_Element_Structure)
	EndIf
Next e

EndProcedure

Procedure.i SVG_ExtractComments(List t.s())
	
	ClearList(t())
	
	s = svgdata\s
	If Not s : ProcedureReturn 0 : EndIf
	If Not svgmem(s)\elements : ProcedureReturn 0 : EndIf
	
	SVGInternal_ForEachElement(s,e)
	If svgmem(s)\element(e)\type = #SVG_Element_Comment
		AddElement(t())
		t() = svgmem(s)\element(e)\content
	EndIf
Next e
ProcedureReturn ListSize(t())

EndProcedure










;-
;- GETTING ELEMENT ATTRIBUTES

Macro SVG_GetElementType()
	svgmem(svgdata\s)\element(svgdata\e)\type
EndMacro

Macro SVG_GetElementID()
	svgmem(svgdata\s)\element(svgdata\e)\id
EndMacro

Macro SVG_GetElementX()
	svgmem(svgdata\s)\element(svgdata\e)\x
EndMacro
Macro SVG_GetElementY()
	svgmem(svgdata\s)\element(svgdata\e)\y
EndMacro

Macro SVG_GetElementHREF()
	svgmem(svgdata\s)\element(svgdata\e)\href
EndMacro

Macro SVG_GetElementHREFTarget()
	svgmem(svgdata\s)\element(svgdata\e)\href_target
EndMacro

Macro SVG_GetElementClass()
	svgmem(svgdata\s)\element(svgdata\e)\class
EndMacro

Macro SVG_GetElementFillMode()
	svgmem(svgdata\s)\element(svgdata\e)\fillmode
EndMacro

Procedure.i SVG_GetElementColor()
	With  svgmem(svgdata\s)\element(svgdata\e)
		If \fillmode=#SVG_FillMode_Color
			ProcedureReturn \fillvalue
		EndIf
	EndWith
	ProcedureReturn -1
EndProcedure
Procedure.i SVG_GetElementGradient()
	With  svgmem(svgdata\s)\element(svgdata\e)
		If \fillmode=#SVG_FillMode_Gradient
			ProcedureReturn \fillvalue
		EndIf
	EndWith
	ProcedureReturn -1
EndProcedure
Procedure.i SVG_GetElementPattern()
	With  svgmem(svgdata\s)\element(svgdata\e)
		If \fillmode=#SVG_FillMode_Pattern
			ProcedureReturn \fillvalue
		EndIf
	EndWith
	ProcedureReturn -1
EndProcedure

Macro SVG_GetElementVisibility()
	svgmem(svgdata\s)\element(svgdata\e)\visibility
EndMacro

Macro SVG_GetElementTooltip()
	svgmem(svgdata\s)\element(svgdata\e)\tooltip
EndMacro

Macro SVG_GetElementAction(a)
	svgmem(svgdata\s)\element(svgdata\e)\action[a]
EndMacro


Procedure.s SVG_GetElementText()
	EditElementCheck
	Select svgmem(s)\element(e)\type
		Case #SVG_Element_Text, #SVG_Element_WrappedText, #SVG_Element_TextPath, #SVG_Element_Comment
			ProcedureReturn svgmem(s)\element(e)\content
		Default
			ProcedureReturn ""
	EndSelect
EndProcedure

Procedure.i SVG_GetElementImage()
	EditElementCheck
	If svgmem(s)\element(e)\type = #SVG_Element_PBImage
		ProcedureReturn svgmem(s)\element(e)\imagenumber
	EndIf
	ProcedureReturn 0
EndProcedure

Procedure.b SVG_GetElementFont(*fname,*fsize)
	EditElementCheck
	Select svgmem(s)\element(e)\type
		Case #SVG_Element_Text, #SVG_Element_WrappedText, #SVG_Element_Group, #SVG_Element_TextPath
			PokeS(*fname,svgmem(s)\element(e)\fontname)
			PokeI(*fsize,svgmem(s)\element(e)\fontsize)
			ProcedureReturn #True
	EndSelect
	ProcedureReturn #False
EndProcedure

Procedure.i SVG_GetElementMask()
	EditElementCheck
	ProcedureReturn svgmem(s)\element(e)\mask
EndProcedure










;-
;- PATHS

Procedure.i SVG_Path_StartSegment(x.f,y.f)
	
	If Not svgdata\s Or Not svgdata\e : ProcedureReturn #False : EndIf
	
	With svgmem(svgdata\s)\element(svgdata\e)\path
		;If \pathpoints=#MaxSVGPathPoints : ProcedureReturn #False : EndIf
		\pathpoints+1
		arrsize=ArraySize(\pathpoint(),1) : If arrsize<\pathpoints : ReDim \pathpoint(arrsize+#SVG_ArrayJump) : EndIf
		\pathpoint(\pathpoints)\x = x
		\pathpoint(\pathpoints)\y = y
		\pathpoint(\pathpoints)\mode = #SVG_Path_SegmentStart
		ProcedureReturn \pathpoints
	EndWith
	
EndProcedure


Procedure.i SVG_Path_Line(dx.f,dy.f,mode.b=#SVG_Path_Straight,ix1.f=-#PB_Ignore,iy1.f=#PB_Ignore,ix2.f=#PB_Ignore,iy2.f=#PB_Ignore,ix3.f=#PB_Ignore,iy3.f=#PB_Ignore,ix4.f=#PB_Ignore,iy4.f=#PB_Ignore)
	
	s = svgdata\s
	e = svgdata\e
	If Not s Or Not e : ProcedureReturn #False : EndIf
	
	With svgmem(s)\element(e)
		;If \pathpoints=#MaxSVGPathPoints : ProcedureReturn #False : EndIf
		\path\pathpoints+1
		ps = \path\pathpoints
		arrsize=ArraySize(\path\pathpoint(),1) : If arrsize<ps : ReDim \path\pathpoint(arrsize+10) : EndIf
		\bounds_known = #False
	EndWith
	
	With svgmem(s)\element(e)\path\pathpoint(ps)
		\mode = mode
		\x = dx
		\y = dy
		\intermediary_points = 0
		
		Select \mode
			Case #SVG_Path_CubicBezier, #SVG_Path_SmoothCubicBezier, #SVG_Path_QuadraticBezier, #SVG_Path_SmoothQuadraticBezier
				If ix1<>#PB_Ignore And iy1<>#PB_Ignore
					\intermediary_points+1
					\intermediary_point(\intermediary_points)\x = ix1
					\intermediary_point(\intermediary_points)\y = iy1
				EndIf
				If ix2<>#PB_Ignore And iy2<>#PB_Ignore
					\intermediary_points+1
					\intermediary_point(\intermediary_points)\x = ix2
					\intermediary_point(\intermediary_points)\y = iy2
				EndIf
				If ix3<>#PB_Ignore And iy3<>#PB_Ignore
					\intermediary_points+1
					\intermediary_point(\intermediary_points)\x = ix3
					\intermediary_point(\intermediary_points)\y = iy3
				EndIf
				If ix4<>#PB_Ignore And iy4<>#PB_Ignore
					\intermediary_points+1
					\intermediary_point(\intermediary_points)\x = ix4
					\intermediary_point(\intermediary_points)\y = iy4
				EndIf
		EndSelect
	EndWith
	
	ProcedureReturn ps
	
EndProcedure


Procedure.i SVG_Path_LineV2(dx.f,dy.f,mode.b,Array intermediary_point.PointF(1),intermediary_points.i=0)
	
	s = svgdata\s
	e = svgdata\e
	If Not s Or Not e : ProcedureReturn #False : EndIf
	
	With svgmem(s)\element(e)
		;If \pathpoints=#MaxSVGPathPoints : ProcedureReturn #False : EndIf
		\path\pathpoints+1
		ps = \path\pathpoints
		arrsize=ArraySize(\path\pathpoint(),1) : If arrsize<ps : ReDim \path\pathpoint(arrsize+10) : EndIf
		\bounds_known = #False
	EndWith
	
	With svgmem(s)\element(e)\path\pathpoint(ps)
		\mode = mode
		\x = dx
		\y = dy
		If intermediary_points<1 : intermediary_points = ArraySize(intermediary_point(),1) : EndIf
		\intermediary_points = intermediary_points
		Dim \intermediary_point(\intermediary_points)
		Select mode
			Case #SVG_Path_CubicBezier, #SVG_Path_SmoothCubicBezier, #SVG_Path_QuadraticBezier, #SVG_Path_SmoothQuadraticBezier
				If \intermediary_points
					For a = 1 To \intermediary_points
						\intermediary_point(a)\x = intermediary_point(a)\x
						\intermediary_point(a)\y = intermediary_point(a)\y
					Next a
				EndIf
		EndSelect
		
		ProcedureReturn ps
	EndWith
	
EndProcedure


Procedure.i SVG_Path_EllipticalArc(dx.f,dy.f,rx.f,ry.f,xrot.f,large_arc.b,sweep.b) ; elliptical arc
	
	If Not svgdata\s Or Not svgdata\e : ProcedureReturn #False : EndIf
	
	With svgmem(svgdata\s)\element(svgdata\e)
		;If \pathpoints=#MaxSVGPathPoints : ProcedureReturn #False : EndIf
		\path\pathpoints+1
		ps = \path\pathpoints
		arrsize=ArraySize(\path\pathpoint(),1) : If arrsize<ps : ReDim \path\pathpoint(arrsize+10) : EndIf
		\bounds_known = #False
	EndWith
	
	With svgmem(svgdata\s)\element(svgdata\e)\path\pathpoint(ps)
		\mode = #SVG_Path_EllipticalArc
		\x = dx
		\y = dy
		\rx = rx
		\ry = ry
		\x_rotation = xrot
		\large_arc = large_arc
		\sweep = sweep
		\intermediary_points = 0
	EndWith
	
	ProcedureReturn ps
	
EndProcedure


Procedure.i SVG_Path_EndSegment()
	
	If Not svgdata\s Or Not svgdata\e : ProcedureReturn #False : EndIf
	
	With svgmem(svgdata\s)\element(svgdata\e)\path
		;If \pathpoints=#MaxSVGPathPoints : ProcedureReturn #False : EndIf
		\pathpoints+1
		arrsize=ArraySize(\pathpoint(),1) : If arrsize<\pathpoints : ReDim \pathpoint(arrsize+10) : EndIf
		\pathpoint(\pathpoints)\mode = #SVG_Path_SegmentEnd
		ProcedureReturn \pathpoints
	EndWith
	
EndProcedure


Macro SVG_ClearPath()
	If svgdata\s And svgdata\e
		With svgmem(svgdata\s)\element(svgdata\e)
			\pathpoints = 0
			ResetElementBounds(svgdata\s)
			\bounds_known = #True
		EndWith
	EndIf
EndMacro


Procedure.b SVG_PathDataFromString(pathdata$,add.b=#False)
	
	EditElementCheck
	With svgmem(s)\element(e)
		Select \type
			Case #SVG_Element_Path, #SVG_Element_TextPath
				If Not add
					If \path\pathpoints
						For p = 1 To \path\pathpoints
							ClearStructure(@\path\pathpoint(p),SVG_PathPoint_Structure)
						Next p
						\path\pathpoints = 0
					EndIf
				EndIf
				AddToSVGPath(pathdata$,@svgmem(s)\element(e)\path)
				ProcedureReturn #True
			Default
				ProcedureReturn #False
		EndSelect
	EndWith
	
EndProcedure

Procedure.s SVG_PathDataToString()
	
	EditElementCheck
	With svgmem(s)\element(e)
		Select \type
			Case #SVG_Element_Path, #SVG_Element_TextPath
				CalibrateSVGForOutputting(s)
				ProcedureReturn SVGInternal_CalibratePath(s,@svgmem(s)\element(e)\path)
			Default
				ProcedureReturn ""
		EndSelect
	EndWith
	
EndProcedure

Procedure.b SVG_SimplifyPath(mindif.f=#PB_Ignore)
	
	s = svgdata\s
	If Not s : ProcedureReturn #False : EndIf
	e = svgdata\e
	If e<1 : ProcedureReturn #False : EndIf
	If e>svgmem(s)\elements : ProcedureReturn #False : EndIf
	
	
	If mindif=#PB_Ignore
		;CalculateTotalSVGElementBounds(s)
		;dime.f = Defeat(svgmem(s)\bound\w,svgmem(s)\bound\h)
		;mindif = dime/200
		;R("MINDIF: "+StrF(mindif))
		mindif = 0.01
	EndIf
	mindif = Beat(0.01,mindif)
	
	
	With svgmem(s)\element(e)\path
		
		Select svgmem(s)\element(e)\type
			Case #SVG_Element_Path, #SVG_Element_TextPath
			Default
				ProcedureReturn #False
		EndSelect
		
		If \pathpoints<4 : ProcedureReturn #False : EndIf
		
		Dim blank.b(\pathpoints)
		
		
		p=1
		Repeat
			; find next set of three consecutive LineTo's
			np=0
			For a = p To \pathpoints-2
				If \pathpoint(a)\mode=#SVG_Path_Straight
					If \pathpoint(a+1)\mode=#SVG_Path_Straight
						If \pathpoint(a+2)\mode=#SVG_Path_Straight
							np=a
							Break
						EndIf
					EndIf
				EndIf
			Next a
			If Not np : Break : EndIf
			;If np>(\pathpoints-2) : Break : EndIf
			p=np
			
			p1.PointF
			p2.PointF
			p3.PointF
			p1\x = \pathpoint(p)\x
			p1\y = \pathpoint(p)\y
			p2\x = \pathpoint(p+1)\x
			p2\y = \pathpoint(p+1)\y
			p3\x = \pathpoint(p+2)\x
			p3\y = \pathpoint(p+2)\y
			
			ang.f = RadianAngleBetweenTwoPoints(@p1,@p3)
			lng.f = DistanceBetweenTwoPoints(@p1,@p2)
			RadianCoordsFromPoint(@p1,ang,lng,@mid.PointF)
			
			dif.f = DistanceBetweenTwoPoints(@mid,@p2)
			;Debug dif
			If dif<mindif
				blank(p+1) = #True
				;Debug "BLANK"
				altered=#True
				p+3
			Else
				p+1
			EndIf
		ForEver
		
		
		If Not altered
			ProcedureReturn #True
		EndIf
		
		
		; get rid of nulled points...
		For a = 1 To \pathpoints
			If Not blank(a)
				tracked=a
				Break
			EndIf
		Next a
		tracked=1
		For z = 1 To \pathpoints
			;Debug tracked
			realpoints+1
			\pathpoint(realpoints)\x = \pathpoint(tracked)\x
			\pathpoint(realpoints)\y = \pathpoint(tracked)\y
			\pathpoint(realpoints)\mode = \pathpoint(tracked)\mode
			\pathpoint(realpoints)\intermediary_points = \pathpoint(tracked)\intermediary_points
			CopyArray(\pathpoint(tracked)\intermediary_point(),\pathpoint(realpoints)\intermediary_point())
			\pathpoint(realpoints)\sweep = \pathpoint(tracked)\sweep
			\pathpoint(realpoints)\large_arc = \pathpoint(tracked)\large_arc
			\pathpoint(realpoints)\rx = \pathpoint(tracked)\rx
			\pathpoint(realpoints)\ry = \pathpoint(tracked)\ry
			\pathpoint(realpoints)\x_rotation = \pathpoint(tracked)\x_rotation
			
			; next point...
			If tracked=\pathpoints : Break : EndIf
			For a = tracked+1 To \pathpoints
				If Not blank(a)
					newtracked=a
					Break
				EndIf
			Next a
			If Not newtracked : Break : EndIf
			tracked=newtracked
		Next z
		\pathpoints = realpoints
		
	EndWith
	
	ProcedureReturn #True
	
EndProcedure


Procedure.s SVG_PolylineDataToString()
	
	EditElementCheck
	With svgmem(s)\element(e)
		Select \type
			Case #SVG_Element_Polyline, #SVG_Element_Polygon
				CalibrateSVGForOutputting(s)
				ProcedureReturn CalibrateSVGPoly(s,e)
			Default
				ProcedureReturn ""
		EndSelect
	EndWith
	
EndProcedure










;-
;- GRADIENTS

Macro SVG_SelectGradient(g)
	svgdata\g=g
EndMacro

Macro SVG_RadialGradient(cx,cy,rad,id="")
	AddSVGGradient(svgdata\s,#SVG_Gradient_Radial,id,cx,cy,rad,rad)
EndMacro

Macro SVG_LinearGradient(x,y,w,h,id="")
	AddSVGGradient(svgdata\s,#SVG_Gradient_Linear,id,x,y,w,h)
EndMacro

Procedure.b SVG_GradientColor(position.f,clr.i,opc=100)
	
	If Not svgdata\g : ProcedureReturn #False : EndIf
	With svgmem(svgdata\s)\gradient(svgdata\g)
		AddElement(\Point())
		\Point()\position = position
		\Point()\color = clr
		\Point()\opacity = opc
		ProcedureReturn #True
	EndWith
	
EndProcedure

Macro SVG_ResetGradientColors()
	If svgdata\s And svgdata\g
		ResetList(svgmem(svgdata\s)\gradient[svgdata\g]\Point())
	EndIf
EndMacro

Procedure.i SVG_CopyGradient(g1,copyid.s="")
	
	s = svgdata\s
	If Not s
		ProcedureReturn #False
	EndIf
	
	g2 = FirstAvailableGradientSlot(s)
	If Not g2
		ProcedureReturn 0
	EndIf
	
	CopyStructure(@svgmem(s)\gradient(g1),@svgmem(s)\gradient(g2),SVG_Gradient_Structure)
	With svgmem(s)\gradient(g2)
		\id = copyid
		If Nots \id
			\id = RandomID(s)
		EndIf
	EndWith
	
	svgdata\g = g2
	ProcedureReturn g2
	
EndProcedure


Procedure.b SVG_FreeGradient(g)
	
	If Not svgdata\s : ProcedureReturn #False : EndIf
	If g<1 Or g>svgmem(svgdata\s)\gradients : ProcedureReturn #False : EndIf
	
	ClearStructure(@svgmem(svgdata\s)\gradient(g),SVG_Gradient_Structure)
	
	If g=svgdata\g
		svgdata\g = 0
	EndIf
	
	With svgmem(svgdata\s)
		If \bg_mode = #SVG_FillMode_Gradient
			If \bg_value = g
				\bg_mode = #SVG_FillMode_None
			EndIf
		EndIf
		If \elements
			For e = 1 To \elements
				If \element(e)\fillmode = #SVG_FillMode_Gradient
					If \element(e)\fillvalue = g
						\element(e)\fillmode = #SVG_FillMode_Color
						\element(e)\fillvalue = #White
					EndIf
				EndIf
			Next e
		EndIf
	EndWith
	
	ProcedureReturn #True
	
EndProcedure


Procedure.s SVG_GradientToString()
	s = svgdata\s
	g = svgdata\g
	If s And g
		ProcedureReturn SVGInternal_GradientToString(s,g);@svgmem(s)\gradient(g))
	EndIf
EndProcedure











;-
;- PATTERNS

Procedure.i SVG_NewPatternFromCode(w,h,code.s,id.s="")
	
	s = svgdata\s
	If code
		pat = AddSVGPattern(s,w,h,#SVG_Pattern_DirectCode,id)
		svgmem(s)\pattern(pat)\code = code
		ProcedureReturn pat
	Else
		ProcedureReturn 0
	EndIf
	
EndProcedure

Procedure.i SVG_NewPatternFromSVG(s.i,w.f,h.f,id.s="")
	
	If s<>svgdata\s
		If IsSVG(s)
			pat = AddSVGPattern(svgdata\s,w,h,#SVG_Pattern_SVG,id)
			svgmem(svgdata\s)\pattern(pat)\svg = s
			ProcedureReturn pat
		EndIf
	EndIf
	
	ProcedureReturn 0
	
EndProcedure


Procedure.i SVG_CopyPattern(g1,copyid.s="")
	
	s = svgdata\s
	If Not s
		ProcedureReturn #False
	EndIf
	
	g2 = FirstAvailablePatternSlot(s)
	If Not g2
		ProcedureReturn 0
	EndIf
	
	CopyStructure(@svgmem(s)\pattern(g1),@svgmem(s)\pattern(g2),SVG_Pattern_Structure)
	With svgmem(s)\pattern(g2)
		\id = copyid
		If Nots \id
			\id = RandomID(s)
		EndIf
	EndWith
	
	svgdata\p = g2
	ProcedureReturn g2
	
EndProcedure


Procedure.b SVG_FreePattern(p)
	
	If Not svgdata\s : ProcedureReturn #False : EndIf
	If p<1 Or p>svgmem(svgdata\s)\patterns : ProcedureReturn #False : EndIf
	
	ClearStructure(@svgmem(svgdata\s)\pattern(p),SVG_Pattern_Structure)
	
	If p=svgdata\p
		svgdata\p = 0
	EndIf
	
	With svgmem(svgdata\s)
		If \bg_mode = #SVG_FillMode_Pattern
			If \bg_value = p
				\bg_mode = #SVG_FillMode_None
			EndIf
		EndIf
		If \elements
			For e = 1 To \elements
				If \element(e)\fillmode = #SVG_FillMode_Pattern
					If \element(e)\fillvalue = p
						\element(e)\fillmode = #SVG_FillMode_Color
						\element(e)\fillvalue = #White
					EndIf
				EndIf
			Next e
		EndIf
	EndWith
	
	ProcedureReturn #True
	
EndProcedure












;-
;- FILTERS

Procedure.i SVGInternal_FirstAvailableFilterSlot(s)
	
	With svgmem(s)
	
	If \filters
		For a = 1 To \filters
			If Not \filter(a)\in_use
				ProcedureReturn a
			EndIf
		Next a
	EndIf
	
	arrsize = ArraySize(\filter(),1)
	If arrsize<(\filters+1)
		;R("CURRENT ARRSIZE: "+Str(arrsize))
		If \filters<1 Or arrsize<1
			Dim \filter(arrsize+#SVG_ArrayJump)
		Else
			ReDim \filter(arrsize+#SVG_ArrayJump)
		EndIf
		If ArraySize(\filter(),1) = -1
			ProcedureReturn 0
		EndIf
	EndIf
	
	\filters+1
	ProcedureReturn \filters
	
	EndWith
	
EndProcedure

Procedure.s SVGInternal_NewFilterID(s,id.s)
	
	Repeat
		found.b = #False
		For f = 1 To svgmem(s)\filters
			If svgmem(s)\filter(f)\id = id
				found = #True
				Break
			EndIf
		Next f
		If Not found
			ProcedureReturn id
		EndIf
		
		id = "fil"
		For a = 1 To 10
			id+Str(Random(9))
		Next a
	ForEver
	
EndProcedure

Macro SVGInternal_SetRegionEntire(name)
	name\region\x = -200
	name\region\y = -200
	name\region\w = -100
	name\region\h = -100
EndMacro

Procedure.i SVG_CreateFilter(id.s="")
	
	s = svgdata\s
	If Not s : ProcedureReturn 0 : EndIf
	
	f = SVGInternal_FirstAvailableFilterSlot(s)
	
	If Not f : ProcedureReturn #False : EndIf
	
	id = SVGInternal_NewFilterID(s,id)
	
	If nf>svgmem(s)\filters : svgmem(s)\filters+1 : EndIf
	svgmem(s)\filter(f)\in_use = #True
	svgmem(s)\filter(f)\id = id
	SVGInternal_SetRegionEntire(svgmem(s)\filter(f))
	svgdata\f = f
	ProcedureReturn f
	
EndProcedure

Procedure.i SVG_CopyFilter(f.i=0)
	
	s = svgdata\s
	If Not s
		ProcedureReturn #False
	EndIf
	
	If Not f
		f = svgdata\f
	EndIf
	If Not f
		ProcedureReturn #False
	EndIf
	
	nf = SVGInternal_FirstAvailableFilterSlot(s)
	If Not nf : ProcedureReturn #False : EndIf
	
	id.s = ""
	id = SVGInternal_NewFilterID(s,id)
	
	If nf>svgmem(s)\filters : svgmem(s)\filters+1 : EndIf
	CopyStructure(@svgmem(s)\filter(f),@svgmem(s)\filter(nf),SVG_Filter_Structure)
	svgmem(s)\filter(nf)\in_use = #True
	svgmem(s)\filter(nf)\id = id
	svgdata\f = nf
	ProcedureReturn nf
	
EndProcedure

Procedure.b SVG_RemoveFilter(f.i=0)
	
	s = svgdata\s
	If Not s
		ProcedureReturn #False
	EndIf
	
	If Not f
		f = svgdata\f
	EndIf
	If Not f
		ProcedureReturn #False
	EndIf
	
	ClearStructure(@svgmem(s)\filter(f),SVG_Filter_Structure)
	
	If svgdata\f = f
		svgdata\f = 0
	EndIf
	
	For e = 1 To svgmem(s)\elements
		If svgmem(s)\element(e)\filter = f
			svgmem(s)\element(e)\filter = 0
		EndIf
	Next e
	
	ProcedureReturn #True
	
EndProcedure

Procedure.b SVG_SetFilterRegion(x.f=#PB_Ignore,y.f=#PB_Ignore,w.f=#PB_Ignore,h.f=#PB_Ignore)
	
	s = svgdata\s
	f = svgdata\f
	If s And f
		With svgmem(s)\filter(f)
			If x<>#PB_Ignore : \region\x = x : EndIf
			If y<>#PB_Ignore : \region\y = y : EndIf
			If w<>#PB_Ignore : \region\w = w : EndIf
			If h<>#PB_Ignore : \region\h = h : EndIf
			ProcedureReturn #True
		EndWith
	EndIf
	
	ProcedureReturn #False
	
EndProcedure

Procedure.b SVG_SetFilterPrimitiveRegion(x.f,y.f,w.f,h.f)
	
	s = svgdata\s
	f = svgdata\f
	If s And f
		With svgmem(s)\filter(f)
			If ListSize(\primitive())
				LastElement(\primitive())
				\primitive()\region\x = x
				\primitive()\region\y = y
				\primitive()\region\w = w
				\primitive()\region\h = h
				ProcedureReturn #True
			EndIf
		EndWith
	EndIf
	
	ProcedureReturn #False
	
EndProcedure

Procedure.b SVG_SetFilterPrimitiveID(id.s)
	
	s = svgdata\s
	f = svgdata\f
	If s And f
		With svgmem(s)\filter(f)
			If ListSize(\primitive())
				LastElement(\primitive())
				\primitive()\id = id
				ProcedureReturn #True
			EndIf
		EndWith
	EndIf
	
	ProcedureReturn #False
	
EndProcedure











;-
;- MASKING

Procedure.i SVG_Mask(id.s="")
	s = svgdata\s
	If Not s : ProcedureReturn 0 : EndIf
	If Nots id
		id = RandomID(s)
	EndIf
	g = SVG_Group(id)
	svgmem(s)\element(g)\ismask = #True
	ProcedureReturn g
EndProcedure

Procedure.b SVG_EndMask()
	
	SVG_CloseGroup()
	
EndProcedure


Procedure.b SVG_SetGroupMaskStatus(status.b)
	EditElementCheck
	With svgmem(s)\element(e)
		If \type<>#SVG_Element_Group
			ProcedureReturn #False
		EndIf
		If status=#False
			If \ismask
				; reset references
				SVGInternal_ForEachElement(s,a)
				If svgmem(s)\element(a)\mask = e
					svgmem(s)\element(a)\mask = 0
				EndIf
			Next a
		EndIf
		\ismask = #False
		ProcedureReturn #True
	EndIf
	If status=#True
		\ismask = #True
		If Nots \id
			\id = RandomID(svgdata\s)
		EndIf
		ProcedureReturn #True
	EndIf
EndWith
EndProcedure

Macro SVG_GetGroupMaskStatus()
	svgmem(svgdata\s)\element(svgdata\e)\ismask
EndMacro









;-
;- RESOURCES

Macro SVG_AddExternalJavascriptFile(relativefilename)
	AddElement(svgmem(svgdata\s)\external_js_file())
	svgmem(svgdata\s)\external_js_file() = relativefilename
EndMacro

Macro SVG_AddInternalJavascript(newjs)
	svgmem(svgdata\s)\javascript+newjs
EndMacro

Macro SVG_AddExternalStylesheet(relativefilename)
	AddElement(svgmem(svgdata\s)\external_css_file())
	svgmem(svgdata\s)\external_css_file() = relativefilename
EndMacro

Macro SVG_AddClass(classname,classdata)
	If classdata<>"" And svgdata\s>0
		svgmem(svgdata\s)\classinfo+EnsureStart(classname,".")+" { "+Trim(classdata)+" }"+c13
	EndIf
EndMacro

Procedure.b SVG_AbsorbExternalStylesheet(absolutefilename.s)
	
	css.s = FileToString(absolutefilename)
	If css
		svgmem(svgdata\s)\classinfo+c13+css+c13
		ProcedureReturn #True
	EndIf
	ProcedureReturn #False
	
EndProcedure










;-
;- METADATA

Macro SVG_SetMetadataString(f,t)
	svgmem(svgdata\s)\md[f] = t
EndMacro

Macro SVG_SetMetadataInteger(f,n)
	svgmem(svgdata\s)\md[f] = Str(n)
EndMacro

Procedure SVG_ClearMetadata(title.b=#False)
	For a = 1 To #TotalSVGMetadataFields
		If a=#SVG_Metadata_Title And Not title
			Continue
		EndIf
		svgmem(svgdata\s)\md[a] = ""
	Next a
EndProcedure











;-
;- EXPORTING

Procedure.s SVG_ElementToString()
	
	EditElementCheck
	ProcedureReturn SVGInternal_ElementToString(s,e,#True)
	
EndProcedure


Procedure.s SVG_StyleToString()
	
	EditElementCheck
	ProcedureReturn SVGInternal_StyleToString(s,e)
	
EndProcedure










;-
;- OUTPUTTING

Procedure.b SVG_SnapToBounds(pad.f=0)
	
	s = svgdata\s
	If Not s : ProcedureReturn #False : EndIf
	;pad=0
	
	With svgmem(s)
		If \elements<1 : ProcedureReturn #True : EndIf
		MasterForAllElements
		;ReportRectF(@master,0,"MASTER")
		SVGInternal_ForEachElement(s,e)
		KnowBounds(s,e)
		nx.f = \element(e)\bound\x-master\x+pad
		ny.f = \element(e)\bound\y-master\y+pad
		sx.f = nx-\element(e)\bound\x
		sy.f = ny-\element(e)\bound\y
		;ReportRectF(@\element(e)\bound,0,"n "+StrF(nx,0)+","+StrF(ny,0))
		SVGInternal_ShiftElement(s,e,sx,sy)
	Next e
	\setw = master\w+pad+pad
	\seth = master\h+pad+pad
EndWith

ProcedureReturn #True

EndProcedure

Procedure.f SVG_GetHorizontalBound()
	
	s = svgdata\s
	If Not s
		ProcedureReturn -1
	EndIf
	
	With svgmem(s)
		MasterForAllElements
		ProcedureReturn master\w
	EndWith
	
EndProcedure

Procedure.f SVG_GetVerticalBound()
	
	s=svgdata\s
	If Not s
		ProcedureReturn -1
	EndIf
	
	With svgmem(s)
		MasterForAllElements
		ProcedureReturn master\h
	EndWith
	
EndProcedure


Procedure.b SVG_SetOutputZoom(s.i,z.f)
	svgmem(s)\output_zoom = z
EndProcedure


Procedure.s LineNumbering(t.s)
	
	reb.s
	lines = CountString(t,c13)
	w = Len(Str(lines))
	For a = 1 To lines
		reb+RSet(Str(a),w,c32)+c32+c32+c32+c32+StringField(t,a,c13)+c13
	Next a
	
	ProcedureReturn reb
	
EndProcedure


; Procedure.b SaveSVGAsSWF(s.i,filename.s)
;   svgfilename.s = previewfol+"temp.svg"
;   DeleteFile(svgfilename)
;   DeleteFile(filename)
;    If SaveSVG(s,svgfilename)
;       param$ = "--dsize " + StrF(svgmem(s)\w,0) + "x" + StrF(svgmem(s)\h,0) + " " +svgfilename + " " + filename
;       RunProgram("svg2swf.exe",param$,"",#PB_Program_Wait | #PB_Program_Hide )
;       DeleteFile(svgfilename)
;       If FileSize(filename) > -1
;           ProcedureReturn #True
;       EndIf
;   EndIf
;   ProcedureReturn #False
; EndProcedure

Procedure.b ConvertSVGToSWF(svgfilename.s,swffilename.s,w.f,h.f)
	DeleteFile(swffilename)
	param$ = "--dsize "+StrF(w,0)+"x"+StrF(h,0) +" "+ svgfilename +" "+ swffilename +" ";+ "--default-font"
	;RunProgram(svg2swf_exe,param$,"",#PB_Program_Wait | #PB_Program_Hide )
	DeleteFile(svgfilename)
	If FileExists(swffilename)
		ProcedureReturn #True
	EndIf
EndProcedure

Procedure.b RasterizeSVG(svgfilename.s,outputfilename.s,w.f,h.f)
	
	Select LCase(GetExtensionPart(outputfilename))
		Case "png"
			mimetype.s = "image/png"
		Case "jpg"
			mimetype = "image/jpeg"
		Case "tif"
			mimetype = "image/tiff"
		Case "pdf"
			mimetype = "application/pdf"
	EndSelect
	If Nots mimetype : ProcedureReturn #False : EndIf
	
	param.s = "-jar ^"+batik_rasterizer+"^ ^"+svgfilename+"^ -m "+mimetype+" -d ^"+outputfilename+"^"
	starttime = ElapsedMilliseconds()
	;RunProgram(java_runtime_exe,ReplaceString(param,"^",c34),"",#PB_Program_Wait|#PB_Program_Hide)
	R("TIME TO RASTERISE: "+Str(ElapsedMilliseconds()-starttime)+"ms.")
	If FileSize(outputfilename)>0
		ProcedureReturn #True
	EndIf
	ProcedureReturn #False
	
EndProcedure

Procedure.b RasterizeSVG_Chrome(svgfilename.s,outputfilename.s,w.f,h.f)
	
	Select LCase(GetExtensionPart(outputfilename))
		Case "png"
			mimetype.s = "image/png"
		Case "jpg"
			mimetype = "image/jpeg"
		Case "tif"
			mimetype = "image/tiff"
		Case "pdf"
			mimetype = "application/pdf"
	EndSelect
	If Nots mimetype : ProcedureReturn #False : EndIf
	
	param.s = "-jar ^"+batik_rasterizer+"^ ^"+svgfilename+"^ -m "+mimetype+" -d ^"+outputfilename+"^"
	starttime = ElapsedMilliseconds()
	;RunProgram(java_runtime_exe,ReplaceString(param,"^",c34),"",#PB_Program_Wait|#PB_Program_Hide)
	R("TIME TO RASTERISE: "+Str(ElapsedMilliseconds()-starttime)+"ms.")
	If FileSize(outputfilename)>0
		ProcedureReturn #True
	EndIf
	ProcedureReturn #False
	
EndProcedure

Procedure.b SaveSVG(s.i,filename.s,useviewbox.b=#True,showcode.b=#False)
	
	If Not IsSVG(s) : ProcedureReturn #False : EndIf
	
	EnsureSVGPeripherals(s,GetPathPart(filename))
	
	nativefilename.s = filename
	Select LCase(GetExtensionPart(filename))
		Case "svg"
		Case "png", "jpg", "tif", "pdf"
			israster.b = #True
			filename = previewfol+"temp.svg"
		Case "swf"
			isswf.b = #True
			filename = previewfol+"temp.svg"
	EndSelect
	
	svgcode.s = SVG_FormCode(s,#True,useviewbox)
	If svgcode
		If showcode : R("SVG CODE [saving]"+c13+"-----------------------"+c13+c13+LineNumbering(svgcode)) : EndIf
		;SetClipboardText(ReplaceString(svgcode,c13,c13+c10))
		saved = FileFromString(filename,svgcode)
		If Not saved
			ProcedureReturn #False
		EndIf
		If isswf
			status = ConvertSVGToSWF(filename,nativefilename,svgmem(s)\w,svgmem(s)\h)
			DeleteFile(filename)
			ProcedureReturn status
		EndIf
		If israster
			status = RasterizeSVG(filename,nativefilename,svgmem(s)\w,svgmem(s)\h)
			DeleteFile(filename)
			ProcedureReturn status
		EndIf
		ProcedureReturn #True
	Else
		ProcedureReturn #False
	EndIf
	
EndProcedure




Macro CGI_SendReply(response)
	Written=WriteConsoleData(response,MemoryStringLength(response))
EndMacro
Procedure CGI_ReplySVG(s.i)
	
	HttpAnswer.s = "Content-type: image/svg+xml"  + #CRLF$ + #CRLF$ + SVG_FormCode(s,#False)
	
	CGI_SendReply(@HttpAnswer)
	
EndProcedure









;-
;- PREVIEWING




CompilerIf Defined(ReportSVGFile,#PB_Procedure)=0
	Procedure ReportSVGFile(svgfile.s,w.i,h.i,title.s="")
		
		svgfile = ReplaceString(svgfile,"\","/")
		
		htmlcode.s = "<!DOCTYPE HTML PUBLIC '-//W3C//DTD HTML 4.0 Transitional//EN'>"+c13
		htmlcode + "<html><head></head>"+c13
		htmlcode + "<body bgcolor=#aaaaaa>"+c13
		
		If Not w
			w = 800
		EndIf
		If Not h
			h = 600
		EndIf
		
		htmlcode +   "<DIV style='border:0px solid #000; width:"+Str(w)+"px; height:"+Str(h)+"px; '>" + c13
		htmlcode +           "<OBJECT data='"+svgfile+"' type='image/svg+xml' width='100%' height='100%'>" + c13
		htmlcode +               "<EMBED src='"+svgfile+"' type='image/svg+xml' width='100%' height='100%' />" + c13
		htmlcode +           "</OBJECT>"
		htmlcode +       "</DIV>"
		
		htmlcode + "</body></html>"
		
		htmlfilename.s = previewfol+Str(ElapsedMilliseconds())+RandomLetters(8)+".html"
		FileFromString(htmlfilename,htmlcode)
		
		w+16+20
		h+30
		If Nots title
			title = "Report SVG"
		EndIf
		win=OpenWindow(#PB_Any,0,0,w,h,title,#PB_Window_ScreenCentered)
		wg=WebGadget(#PB_Any,0,0,w,h,htmlfilename)
		;SetGadgetItemText(wg,#PB_Web_HtmlCode,htmlcode)
		
		
		escapekey = 1
		spacekey = 2
		returnkey = 3
		AddKeyboardShortcut(win,#PB_Shortcut_Escape,escapekey)
		AddKeyboardShortcut(win,#PB_Shortcut_Space,spacekey)
		AddKeyboardShortcut(win,#PB_Shortcut_Return,returnkey)
		
		
		CompilerIf #PB_Compiler_OS = #PB_OS_Windows
			If WindowID(win)
				SetWindowPos_(WindowID(win),#HWND_TOPMOST,0,0,0,0,#SWP_NOMOVE|#SWP_NOSIZE)    ; gets it on top to stay 
				;SetWindowPos_(WindowID(win),#HWND_NOTOPMOST,0,0,0,0,#SWP_NOMOVE|#SWP_NOSIZE)  ; now it is on top, allow it to go behind
			EndIf
		CompilerEndIf
		
		
		Repeat
			we = WindowEvent()
			If we
				If we=#PB_Event_Menu
					Break
				EndIf
			Else
				Delay(10)
			EndIf
		ForEver
		
		CloseWindow(win)
		
		DeleteFile(htmlfilename)
		
	EndProcedure
	
	Procedure ReportSVG(s.i,showcode.b=#False)
		
		If Not IsSVG(s)
			R("SVG IS NOT VALID.")
			ProcedureReturn
		EndIf
		
		EnsureFolderPath(previewfol)
		
		EnsureSVGPeripherals(s,previewfol)
		svgcode.s = SVG_FormCode(s,#True)
		If showcode : R("SVG CODE [previewing]"+c13+"-----------------------------"+c13+c13+LineNumbering(svgcode)) : EndIf
		previewsvgfile.s = previewfol+Str(ElapsedMilliseconds())+RandomLetters(8)+".svg"
		FileFromString(previewsvgfile,svgcode)
		
		ReportSVGFile(previewsvgfile,svgmem(s)\bound\w*svgmem(s)\output_zoom,svgmem(s)\bound\h*svgmem(s)\output_zoom,svgmem(s)\md[#SVG_Metadata_Title])
		
		DeleteFile(previewsvgfile)
		
	EndProcedure
CompilerEndIf

Enumeration 1
	#Browser_Chrome
	#Browser_FireFox
	#Browser_IE
	#Browser_Opera
EndEnumeration

Procedure.b ShowSVGInBrowser(s.i,br.i=#Browser_Chrome,standalone.b=#True,showcode.b=#False)
	
	If Not IsSVG(s)
		R("SVG #"+Str(s)+" IS NOT VALID.")
		ProcedureReturn
	EndIf
	
	EnsureFolderPath(previewfol)
	
	EnsureSVGPeripherals(s,previewfol)
	;R("FORM CODE")
	svgcode.s = SVG_FormCode(s,1-inhtml,#False)
	;R("OZ AFTER SVG_FormCode: "+StrF(svgmem(s)\output_zoom))
	;R("END FORM CODE")
	If showcode : R("SVG CODE [previewing]"+c13+"-----------------------------"+c13+c13+LineNumbering(svgcode)) : EndIf
	relativesvgfilename.s = Str(ElapsedMilliseconds())+RandomLetters(8)+".svg"
	absolutesvgfilename.s = previewfol+relativesvgfilename
	FileFromString(absolutesvgfilename,svgcode)
	
	
	If Not standalone
		
		htmlcode.s = "<!DOCTYPE HTML PUBLIC '-//W3C//DTD HTML 4.0 Transitional//EN'>"+c13
		htmlcode + "<html><head></head>"+c13
		htmlcode + "<body bgcolor=#777777 style='margin:0px; padding:0px; '>"+c13
		
		w = svgmem(s)\w * svgmem(s)\output_zoom
		h = svgmem(s)\h * svgmem(s)\output_zoom
		If br=#Browser_Opera Or br=#Browser_FireFox
			w+20
			h+20
		EndIf
		
		htmlcode +   "<div style='padding:0px; margin:0px; width:"+Str(w)+"px; height:"+Str(h)+"px; '>"
		htmlcode +       "<object data='"+relativesvgfilename+"' type='image/svg+xml' width='"+Str(w)+"px' height='"+Str(h)+"px'>"
		htmlcode +           "<embed src='"+relativesvgfilename+"' type='image/svg+xml' />"
		htmlcode +       "</object>"
		htmlcode +   "</div>"
		
		htmlcode + "</body></html>"
		
		htmlfilename.s = previewfol+Str(ElapsedMilliseconds())+RandomLetters(8)+".html"
		FileFromString(htmlfilename,htmlcode)
		
		filename.s = htmlfilename
	Else
		filename.s = absolutesvgfilename
	EndIf
	
	Select br
		Case #Browser_Chrome
			RunProgram("chrome.exe",c34+filename+c34,c34+previewfol+c34)
		Case #Browser_FireFox
			RunProgram("firefox.exe",c34+filename+c34,c34+previewfol+c34)
		Case #Browser_IE
			RunProgram("iexplore.exe",c34+filename+c34,c34+previewfol+c34)
		Case #Browser_Opera
			RunProgram("C:\Program Files\Opera\Opera.exe","newpage "+c34+filename+c34,c34+previewfol+c34)
	EndSelect
	
EndProcedure




Global puresvg_serverfolder.s = "P:\website\upload\pollcity\"
Global puresvg_localhostfolder.s = "http://slushprodukt/pollcity/"
Procedure.b SVGWeb(s.i,br=#Browser_Chrome,showcode.b=#False)
	
	EnsureThisEnd(puresvg_serverfolder,"\")
	EnsureThisEnd(puresvg_localhostfolder,"/")
	
	If Not FileExists(puresvg_serverfolder+"svg.swf")
		R("The required file svg.swf is not in the server folder.")
		ProcedureReturn #False
	EndIf
	If Not FileExists(puresvg_serverfolder+"svg.js")
		R("The required file svg.js is not in the server folder.")
		ProcedureReturn #False
	EndIf
	If Not FileExists(puresvg_serverfolder+"svg.htc")
		R("The required file svg.htc is not in the server folder.")
		ProcedureReturn #False
	EndIf
	
	
	
	
	EnsureSVGPeripherals(s,puresvg_serverfolder)
	svgcode.s = SVG_FormCode(s,#True,#False)
	If showcode : R("SVG CODE [previewing]"+c13+"-----------------------------"+c13+c13+LineNumbering(svgcode)) : EndIf
	relativesvgfilename.s = "test_puresvg.svg"
	absolutesvgfilename.s = puresvg_serverfolder+relativesvgfilename
	FileFromString(absolutesvgfilename,svgcode)
	
	
	
	htmlcode.s = "<!DOCTYPE HTML PUBLIC '-//W3C//DTD HTML 4.0 Transitional//EN'>"+c13
	htmlcode + "<html><head><meta name='svg.render.forceflash' content='true'><script type='text/javascript' src='svg.js'></script></head>"+c13
	htmlcode + "<body bgcolor=#777777 style='margin:0px; padding:0px; '>"+c13
	
	w = svgmem(s)\visible\w * svgmem(s)\output_zoom
	h = svgmem(s)\visible\h * svgmem(s)\output_zoom
	If br=#Browser_Opera Or br=#Browser_FireFox
		w+20
		h+20
	EndIf
	
	htmlcode +   "<!--[if IE]>"+c13
	htmlcode + "<object src='"+relativesvgfilename+"' classid='image/svg+xml'"+c13
	htmlcode + "width='"+Str(w)+"' height='"+Str(h)+"'"+c13
	htmlcode + "id='mySVGObject'>"+c13
	htmlcode + "<![endif]-->"+c13
	htmlcode + "<!--[if !IE]>-->"+c13
	htmlcode +           "<OBJECT data='"+relativesvgfilename+"' type='image/svg+xml' width='"+Str(w)+"px' height='"+Str(h)+"px'>"
	htmlcode +               "<EMBED src='"+relativesvgfilename+"' type='image/svg+xml' />"
	htmlcode +           "</OBJECT>"
	
	htmlcode + "</body></html>"
	
	relativehtmlfilename.s = "test_puresvg.html"
	absolutehtmlfilename.s = puresvg_serverfolder+relativehtmlfilename
	FileFromString(absolutehtmlfilename,htmlcode)
	
	
	
	Select br
		Case #Browser_Chrome
			RunProgram("chrome.exe",c34+puresvg_localhostfolder+relativehtmlfilename+c34,c34+puresvg_localhostfolder+c34)
		Case #Browser_FireFox
			RunProgram("firefox.exe",c34+puresvg_localhostfolder+relativehtmlfilename+c34,c34+puresvg_localhostfolder+c34)
		Case #Browser_IE
			RunProgram("iexplore.exe",c34+puresvg_localhostfolder+relativehtmlfilename+c34,c34+puresvg_localhostfolder+c34)
		Case #Browser_Opera
			RunProgram("C:\Program Files\Opera\Opera.exe","newpage "+c34+puresvg_localhostfolder+relativehtmlfilename+c34,c34+puresvg_localhostfolder+c34)
	EndSelect
	
EndProcedure









; IDE Options = PureBasic 5.22 LTS (Windows - x86)
; CursorPosition = 4425
; FirstLine = 1424
; Folding = wACgAgADIForgLhUoI9-vrCAokig5KAu-ftrFnaNSAAA1--fDHAIADAIAEAHEACYURTBAA9oPAx2Q1KISBy-
; EnablePurifier